/* ===================== revxml_clnt.c   ======================================
Purpose:	  To demonstrate the use of the external function revxml.c
Parameters:       
  input:
		- on the commandline:
				1st parameter: filename_a = filename of old document
				2nd parameter: filename_b = filename of new document
		- to be referenced from revxml_client (see the documentation):
			- filename_def = a file with switches, by default called "revxml.def"
			- filename_emp = a file with specifications for empty elements and discriminating attributes,
												by default called "revxml.emp"
  output:
		- on the commandline:
				3rd parameter: filename_c = filename of resulting document
		- to be referenced from revxml_client:
			- filename_log = a file with messages resulting from the execution, by default called "revxml.log"
		- "revxml.trc": only created when the switch "trace" is set in the def-file
Algorithm:
		- on the Windows platform revxml.c is delivered as a dll, on other platforms as an object file
		- revxml can be called in two modes (indicated by the compiler switch "#define operate_on_files"):
			1. operating on external documents, with a call to revxml_on_files
			2. operating on internal buffers which contain the documents in internal memory, with a call to revxml_on_buffers
		- ad 1.
			- revxml is immediately called as "revxml_on_files"
		- ad 2.
			- for demonstration purposes the files A and B are first read into internal buffers; care is taken
				to remember the positions of new lines; a new line is, if appropriate, replaced by a blank
			- revxml is called as "revxml_on_buffers"
			- for demonstration purposes the file C is written, with new lines on the same positions as in file B

====================================================================*/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <setjmp.h>
                                 
#if (defined(WINDOWS) || defined(WIN32))
   #include <malloc.h>
   #include <io.h>
   #include "make_win.h"
#endif
                                 
/* return codes of revxml which will be returned to the calling program */
#define not_enough_memory_on_heap (1)
#define not_enough_memory_for_d (2)
#define no_corresponding_starttag (3)
#define not_positioned_at_equalsign (4)
#define hashed_text_too_long (5)
#define name_too_long (6)
#define unequal_lengths_of_rev_markers (7)
#define file_not_opened (8)
#define combtags_and_printa_both_true (9)
#define file_buffer_too_small	(10)
#define cr_data_buffer_too_small (11)
#define switches_not_correct (12)
#define too_many_names (13)
#define heapsize_too_large (14)
#define license_expired (15)
#define compared_with_buffer_too_small (16)
#define increase_pagesize (17)
/* end of return codes */

#define inputline_size (20000)
#define max_outputline_size (20000)
#define true (1)
#define false (0)

typedef short my_int; 
typedef unsigned char boolean; 

#define USAGE "arg 1: first xmlfile, arg 2: second xmlfile, arg 3: resulting file"


typedef char string256[255];
static string256	listing_name,
									argument,
									str, 
									return_message,
									buffer_sprintf,/* buffer for printing messages on the display
																		 and in the sgdiff.log file */
									filedir,					/* name of directory that contains the docs */
									filename_a, 			/* filename of old doc */
									filename_b, 			/* filename of new doc */
									filename_c;				/* filename of doc with rev markers */

static char *data_a,   /* text of old data */
            *data_b,   /* text of new data */
            *data_c,   /* text of revised data */
            eol_char;  /* the end of line character that has to be
                          written in the new text with difference
                          markers */
static long len_a,   /* length of the text in data_a */
            len_b,   /* length of the text in data_b */
            cr_data_len,   /* length of the array cr_data */
            *cr_data,   /* array which contains the positions of the
                           cr's in data_b */
            len_c,  /* length of the text in data_c, to be returned in the parameter return_len_c */
						return_code,
						eol_proc;		/* boolean: true if end-of-lines in file B should be kept in file C */

static FILE *infile, *outfile;
#define BLANK	' ' 
static size_t textsize, crsize;
static jmp_buf exception_status; /* ANSI C construction:
                          exception_status will be defined
                          later as a program address to which is jumped
                          if an exceptional status occurs */

/* Exporting functions is done very differently in Windows and Windows NT! */
#ifdef WIN32
#define DllExport(p) 	__declspec ( dllexport ) p
#elif defined(WINDOWS)
#define DllExport(p)	p FAR PASCAL _export 
#else
/* On other platforms, no export is done at all */
#define DllExport(p)	p
#endif
                                 
typedef void (*revxml_on_files_type)(char*, char*, char*, char*, char*, char*, char*, char*, long, long*);
typedef void (*revxml_on_buffers_type)(long, long, long, long, long, long*, long*, long, long*,
                                       char*, char*, char*, char*, char*, long, long*);
static revxml_on_files_type revxml_on_files_function;
static revxml_on_buffers_type revxml_on_buffers_function;

DllExport(void) revxml_on_files(char *filename_a,			/* filename of old doc */
				char *filename_b,			/* filename of new doc */
				char *filename_c,			/* filename of doc with rev markers */
				char *filename_def_param, 		/* filename of file with settings, by default "revxml.def" */
				char *filename_emp_param, 		/* filename of file with empty elements and discriminative 
																	       attributes, by default "revxml.emp" */
				char *filename_log_param, 		/* filename of logfile */
				char *filename_lic_param, 		/* filename of license file */
				char *return_message,	                /* string with message after abnormal termination */
				long eol_proc_param,			/* boolean: true if end-of-lines in file B should be kept in file C */
				long *return_error_code		        /* error code after abnormal termination */
                               );

DllExport(void) revxml_on_buffers(long in_data_a,			/* pointer to old doc */
				  long in_len_a,			/* length of old doc */
				  long in_data_b,			/* pointer to new doc */
				  long in_len_b,			/* length of new doc */
				  long in_data_c,			/* pointer to doc with revision markers */
				  long *return_len_c,			/* length of doc with revision markers */

				  long *in_cr_data,			/* array which contains the positions of the cr's in data_b */
				  long in_cr_data_len,			/* length of the array cr_data */
				  long *eol_char_param,			/* end of line character on current platform */
				  char *filename_def_param, 		/* filename of file with settings, by default "revxml.def" */
				  char *filename_emp_param, 		/* filename of file with empty elements and discriminative 
																		 attributes, by default "revxml.emp" */
				  char *filename_log_param, 		/* filename of logfile */
				  char *filename_lic_param, 		/* filename of license file */
				  char *return_message,		        /* string with message after abnormal termination */
				  long eol_proc_param,			/* boolean: true if end-of-lines in file B should be kept in file C */
				  long	*return_error_code	        /* error code after abnormal termination */
                                 );

#if defined(WINDOWS) || defined(WIN32)

/* ===================== get_args_commandline   ======================================
Purpose:	  generalized function to read the commandline arguments, also
						when called from a dll
Parameters:       
  input:
		- lpCmdLine: de commandline itself          
  output:
		- argc: the number of arguments
		- argv: the arguments        
  input/output:   
Global variables: 
Algorithm:
====================================================================*/
void get_args_commandline(lpCmdLine, argc, argv)
LPSTR lpCmdLine;
int *argc;
string256 argv[];
{    
  short i, j;
  short proceed;  
  
  i = 0;         /* counter in lpCmdLine */
  *argc = -1;    /* argv[argc] */
  proceed = TRUE;
  while(proceed)
  {
    while(lpCmdLine[i] == ' ')   /* skip spaces */
    {  
      i++;
    }
    proceed = (lpCmdLine[i] != '\0'); /* already eol ? */
    if (proceed)
    {
      (*argc)++;      /* next argv */
      j = 0;    /* counter in string value of argv[n] */
      while((lpCmdLine[i] != ' ') && (lpCmdLine[i] != '\0'))  /* not yet a space */
      {  
        argv[*argc][j++] = lpCmdLine[i++]; /* copy character */
      }
      argv[*argc][j] = '\0'; /* copy character */
      proceed = (lpCmdLine[i] != '\0');  
    }
  }
  (*argc)++;      /* we counted from 0 */
} /* end of get_args_commandline  */ 
 
#endif

#if defined(WINDOWS) || defined(WIN32)

HINSTANCE revxml_dll;

boolean load_dll()
{
	string256 revxml_dll_loc = "revxml.dll";	
	string256 msg_buf;

	/* identify and load revxml.dll */
	revxml_dll = NULL;
	revxml_dll = LoadLibrary(revxml_dll_loc); 
			/* follows paths:
			 1.	The directory from which the application loaded. 
			 2.	The current directory. 
			 3..6: see the description of LoadLibrary	in InfoViewer */

	if (!revxml_dll)
	{
		sprintf(msg_buf, "DLL %s not found ", revxml_dll_loc);
		return FALSE;
	}
	
	/* identify procedures from revxml.dll */
	revxml_on_files_function = (revxml_on_files_type) GetProcAddress(revxml_dll, "revxml_on_files");
	revxml_on_buffers_function = (revxml_on_buffers_type) GetProcAddress(revxml_dll, "revxml_on_buffers");

	return TRUE;
}

#endif


/* =====================  readfile_in_buffer  ======================================
Purpose:		read a xml/sgml-file in a buffer and optionally remember the position 
						of the eol's			
Parameters:				
	input:
		- textspace_p: pointer to the buffer
		- eol_proc: boolean, indicates if the position of the eol's
								has to be remembered					
	output:
		- len: the resulting length of the buffer					
	input/output:		
Global variables:	
Algorithm:
		- read the file line by line
		- add a space character to the right of the line
		- if eol_proc: remember the position of the eol in cr_data[] 				
====================================================================*/
static void readfile_in_buffer(char * textspace_p, long eol_proc, 
														long * len) 
/* reads a textfile into a buffer */
{
	char s[inputline_size];
  int length_of_s;
	char last_char_of_s;

/* --- code change 18-4-2002 */   

	boolean s_full_size, feof_infile, s_ends_on_eol;

  /* readfile_in_buffer */
  *len = 0;
	cr_data_len = 0;
  do 
  {
    fgets(s, inputline_size, infile); /* next input line */

		/* We need to know the following conditions:
				- s ends on an eol: s_ends_on_eol
				- s is of size inputline_size-1: s_full_size
				- eof condition: feof_infile
					N.B.: with eof there is no eol
		*/
		length_of_s = strlen(s);
		last_char_of_s = s[length_of_s - 1];
		s_full_size = (length_of_s == (inputline_size - 1) );
		feof_infile = feof(infile);
		s_ends_on_eol = (last_char_of_s == '\n');

		/* if s is not empty: 
				- if there is an eol, remove it
				- supply a space at the end of the line in order to separate two words,
				  but not if the last character is already a space or a '>'
		*/
		
		if (length_of_s > 0)
		{ 
			/* if there is an eol, remove it */
			if (s_ends_on_eol)
				length_of_s = length_of_s - 1;
 
	   /* supply a space at the end of the line in order to separate two words,
				 but not if the last character is already a space or a '>';
				 must be a line of reasonable small size (not machine generated) */ 
			last_char_of_s = s[length_of_s - 1];
			if ( (length_of_s < 150) && 
					 (last_char_of_s != BLANK) &&
					 (last_char_of_s != '>') &&
					 !s_full_size )	
	    {   
	      s[++length_of_s - 1] = BLANK;
	    }

/* --- end of code change 18-4-2002 */
 
      if (*len + length_of_s >= (long)textsize)
      {
        sprintf(buffer_sprintf, 
								"File buffer too small. Increase the constant Textsize, which is now %d\n", textsize);
				return_code = file_buffer_too_small;
				longjmp(exception_status, file_buffer_too_small);
      };
      strncpy(textspace_p + *len, s, length_of_s);
      *len = *len + length_of_s;
      if (eol_proc)
      { 
				cr_data[cr_data_len] = *len;
        cr_data_len++;
        if (cr_data_len >= (long)crsize)
        {
          sprintf(buffer_sprintf, 
									"Cr_data buffer too small. Increase the constant crsize, which is now %d\n", crsize);
					return_code = cr_data_buffer_too_small;
					longjmp(exception_status, cr_data_buffer_too_small);
        };
      }
    }
		s[0] = 0;
  } while (!feof(infile));
} /* readfile_in_buffer */


/* =====================  writefile_from_buffer  ======================================
Purpose:		write a xml/sgml-file from a buffer			
Parameters:				
	input:					
	output:					
	input/output:		
Global variables:	
	- rev_data: buffer of revised text
	- rev_len: length of rev_data
Algorithm:				
====================================================================*/
static void writefile_from_buffer(void) 
		/* writes the revised text from the revised buffer onto 
																the revised output file */
{	long l;
	char c;
	int nr_of_chars_on_line = 0;

	for (l = 0; l < len_c; l++)
	{ 
		c = data_c[l];
		if ( (c == eol_char) || 
					(nr_of_chars_on_line == max_outputline_size) )
		{ 
			fprintf(outfile, "\n");
			nr_of_chars_on_line = 0;
		};
		if ( c != eol_char)
    { 
			putc( c, outfile);
			nr_of_chars_on_line++;
		}
	};
	fprintf(outfile, "\n");
} /* writefile_from_buffer */


static void error_file_open_for_clnt(FILE *openfile, char *filename, long *result)
{
  if (openfile == NULL)
	{
	  sprintf(buffer_sprintf, "The file %s could not be opened. Program terminated", filename);
		*result = file_not_opened;
		/* longjmp(exception_status, file_not_opened); */
	}
	else
	{
	  /* sprintf(buffer_sprintf, "File %s opened.", filename);
		wwrite(buffer_sprintf, 1); */
	}
}


#if ((! defined(WINDOWS)) && (! defined(WIN32)))
#define wwrite(buf, dummy) { fprintf(stderr, "%s\n", buf); }
#endif

time_t start_time, finish_time;

void display_time_used()
{
	long hours, minutes, seconds;
	double elapsed_time;
	long remainder;

	time( &finish_time );
	sprintf(buffer_sprintf, "Finish:\t%s", ctime( &finish_time ) );
	wwrite(buffer_sprintf, 1);
	elapsed_time = difftime( finish_time, start_time );
	hours = (long)elapsed_time / (long)3600;
	remainder = (long)elapsed_time % (long)3600;
	minutes = remainder / (long)60;
	seconds = remainder % (long)60;
	sprintf(buffer_sprintf, "Elapsed time %d:%2d:%2d", hours, minutes, seconds );
	wwrite(buffer_sprintf, 1);
}	/* display_time_used */


/* =============== start main program ================================================= */

#if defined(WINDOWS) || defined(WIN32)
BOOL WINAPI WinMain(hInstance, hPrevInstance, lpCmdLine, nCmdShow)
HINSTANCE  hInstance;
HINSTANCE  hPrevInstance;
LPSTR lpCmdLine;
int nCmdShow;
#else
int main(int argc, char *argv[])
#endif

{
#if defined(WINDOWS) || defined(WIN32)
  string256 argv[20];
	my_int argc;
#endif
	long return_code = 0;
	string256	return_message, filename,
			filename_def="revxml.def", 		/* filename of file with settings */
			filename_emp="revxml.emp",		/* filename of file with empty elements and discriminative attributes */
			filename_log="revxml.log",		/* filename of logfile */
			filename_lic="revxml.lic"; 		/* filename of license file */
	boolean ok;


#if defined(WINDOWS) || defined(WIN32)
  if (!InitMainWindow(hInstance, hPrevInstance, nCmdShow, "RevXml", 20, 20, 700, 400))
    return FALSE;
/*  if (!InitChildInstance(0, "Debug info", 0, 0, 240, 354))
    return FALSE; 
  if (!InitChildInstance(1, "General info", 240, 0, 452, 354))
    return FALSE; */
  if (!InitChildInstance(1, "General info", 0, 0, 700, 400))
    return FALSE;

	sprintf(buffer_sprintf, "RevXml is a batch program. This window shows timing information and,");
	wwrite(buffer_sprintf, 1);
	sprintf(buffer_sprintf, "eventually, error messages. The program may be halted by File|Exit.");
	wwrite(buffer_sprintf, 1); wwrite("", 1);

#endif

	time( &start_time );
	sprintf(buffer_sprintf, "Start: \t%s", ctime( &start_time ) );
	wwrite(buffer_sprintf, 1);

	if (setjmp(exception_status) != 0)
		 /* a call "longjmp(exception_status, n)" with n != 0
		 will return processing to this position; we transfer
		 here the control back to the calling program; before
		 the call to "longjmp" a status code should have been
		 entered into *return_code and a message in
		 return_message */
	{
		sprintf(buffer_sprintf, "Error in revxml, return code = %d. %s.", return_code, return_message);
		wwrite(buffer_sprintf, 1);
		sprintf(buffer_sprintf, "Program aborted");
		wwrite(buffer_sprintf, 1);
#ifndef BATCH_mode
#if defined(WINDOWS) || defined(WIN32)
		Main_Loop(); /* to keep the window open for inspection */
#endif
#endif
		exit(return_code);
	}

#if defined(WINDOWS) || defined(WIN32)

  get_args_commandline(lpCmdLine, &argc, argv);
  /* values are now available as argc and *argv[n] */   
		
	ok = load_dll();
	if (!ok)
	{
		sprintf(buffer_sprintf, "revxml.dll cannot be loaded");
		wwrite(buffer_sprintf, 1);
#ifndef BATCH_mode
#if defined(WINDOWS) || defined(WIN32)
		Main_Loop(); /* to keep the window open for inspection */
#endif
#endif
		exit(1);
	}
#endif

#if defined(WINDOWS) || defined(WIN32)
  if (argc < 4) 
	{
    fprintf(stderr, "%s\n", USAGE);
    exit(EXIT_FAILURE);
  }
	strcpy(filedir, argv[0]);	 strcat(filedir, "\\");
	strcpy(filename, argv[1]); strcpy(filename_a, filedir); strcat(filename_a, filename);
	strcpy(filename, argv[2]); strcpy(filename_b, filedir); strcat(filename_b, filename);
	strcpy(filename, argv[3]); strcpy(filename_c, filedir); strcat(filename_c, filename);
#else
  if (argc < 5) 
	{
    fprintf(stderr, "%s\n", USAGE);
    exit(EXIT_FAILURE);
  }
	revxml_on_files_function = &revxml_on_files;

	strcpy(filedir, argv[1]);	 strcat(filedir, "\\");
	strcpy(filename, argv[2]); strcpy(filename_a, filedir); strcat(filename_a, filename);
	strcpy(filename, argv[3]); strcpy(filename_b, filedir); strcat(filename_b, filename);
	strcpy(filename, argv[4]); strcpy(filename_c, filedir); strcat(filename_c, filename);
#endif /* WINDOWS */

	/* specify here the names of files, evt. including their path */
	strcpy(filename_def, filedir); strcat(filename_def, "revxml.def");
	strcpy(filename_emp, filedir); strcat(filename_emp, "revxml.emp");
	strcpy(filename_log, filedir); strcat(filename_log, "revxml.log");
	strcpy(filename_lic, filedir); strcat(filename_lic, "revxml.lic");

	/* specify here whether end of lines in file B should be kept in file C */
	eol_proc = true; 

// #define operate_on_files /* put between comments if operation on internal buffers is required */

#ifdef operate_on_files

	revxml_on_files_function( 
				filename_a,
		    filename_b, 
		    filename_c,
				filename_def,
				filename_emp,
				filename_log,
				filename_lic,
		    return_message,
				eol_proc,
		    &return_code 
		    ); 

	if (return_code != 0)
	{ 
		longjmp(exception_status, return_code);
	}

#else /* operate_on_buffers */

/* allocate memory */
		textsize = 2000000; /* maximum size of files; also in def-file */
		crsize = textsize/10;
		data_a = malloc(textsize);
		data_b = malloc(textsize);
		data_c = malloc(2*textsize);
		cr_data  = malloc(sizeof(long)*crsize); 
		eol_char = '\n';

/* read files A and B */
		
    infile = fopen(filename_a, "r");
		error_file_open_for_clnt(infile, filename_a, &return_code);
		if (return_code != 0)
		{ 
			longjmp(exception_status, file_not_opened);
		}
    readfile_in_buffer(data_a, /* eol_proc = */ false, &len_a);
			/* do not retain positions of eol in data_a */
    fclose(infile);

    infile = fopen(filename_b, "r");
		error_file_open_for_clnt(infile, filename_b, &return_code);
		if (return_code != 0)
		{ 
			longjmp(exception_status, file_not_opened);
		}
    readfile_in_buffer(data_b, eol_proc, &len_b);
			/* if eol_proc: retain positions of eol in data_b */
    fclose(infile);

		if (!eol_proc)
		{
			cr_data_len = 0;
			cr_data[0] = 0;
		}
		
#if ((! defined(WINDOWS)) && (! defined(WIN32)))
	revxml_on_buffers_function = &revxml_on_buffers;
#endif

	revxml_on_buffers_function(
		    (long)data_a, 		/* pointer to old doc */
		    (long)len_a,			/* length of old doc */
				(long)data_b,			/* pointer to new doc */
				(long)len_b,			/* length of new doc */
				(long)data_c,			/* pointer to doc with revision markers */
				(long*)(&len_c),	/* length of doc with revision markers */
				(long*)cr_data,		/* array which contains the positions of the cr's in data_b */
				(long)cr_data_len,/* length of the array cr_data */
				(long*)(&eol_char),	/* end of line character on current platform */
				filename_def, 		/* filename of file with settings, by default "revxml.def" */
				filename_emp, 		/* filename of file with empty elements and discriminative 
																attributes, by default "revxml.emp" */
				filename_log, 		/* filename of logfile */
				filename_lic, 		/* filename of license file */
		    return_message,		/* string with message after abnormal termination */
				eol_proc,				/* boolean: true if end-of-lines in file B should be kept in file C */
		    &return_code 			/* error code after abnormal termination */
		    ); 
	
	if (return_code != 0)
	{ 
		longjmp(exception_status, return_code);
	}

/* write file C */
	strcpy(return_message, buffer_sprintf);
	if (return_code == 0)
	{
		outfile = fopen(filename_c, "w");
		error_file_open_for_clnt(outfile, filename_c, &return_code);
		if (return_code != 0)
		{ 
			longjmp(exception_status, file_not_opened);
		}
		writefile_from_buffer();
		fclose(outfile);
	}

	free(data_a);
	free(data_b);
	free(data_c);
	free(cr_data); 

#endif /* operate_on_buffers */

 	sprintf(buffer_sprintf, "Program successfully terminated");
	wwrite(buffer_sprintf, 1);

	display_time_used();

#if defined(WINDOWS) || defined(WIN32)
	FreeLibrary(revxml_dll);    /* free the DLL */ 
#endif  

#ifndef BATCH_mode
#if defined(WINDOWS) || defined(WIN32)
	Main_Loop(); /* to keep the window open for inspection */
#endif
#endif

	exit(return_code);
}
