/* RevXml.c 

	Update history
	Version
			29-10-2002 Renaming within diff_marker_type
	2.1	10-8-2002	Complete handling of mixed content
	2.0	1-1-2002	Start of history RevXml. Former versions are of predecessor Sgdiff

*/

/* ====================================================================+========
         1         2         3         4         5         6         7         8
12345678901234567890123456789012345678901234567890123456789012345678901234567890
The program revxml creates a file which contains indications
("revision markers") for the differences of two input files.


Example of call:  revxml st1_a st1_b st1_c
  where:  - 'st1_a' is the filename of the original document (input)
          - 'st1_b' is the filename of the revised document (input)
          - 'st1_c' is the filename of the revised document with
              revision markers (output)

						Input to CALLDIFF are:
						- the texts a and b
						- a file called revxml.emp' which contains the names of empty tags,
							in lower case
						- a file which contains parameters, by default called revxml.def
						Output from CALLDIFF are:
						- the text c ( = the text b with revision markers)
						- a tracefile called 'revxml.trc' (if required)
						- a logging file, called 'revxml.log'.

            The tracing serves two purposes: 1) to list, in a tree
            structure, the original document and the revised document
            2) to trace and to demonstrate the calculation process.
            Warning. The tracing file can grow disproportionally and
            its creation will cost much more cpu-time than the
            calculations themselves. Therefore, it is advised not to
            create a trace file for large input files.
          - the deffile contains the characterstrings of the 8 revision markers.


            Initially, these strings are written in the deffile as
            follows for use outside attributes:
              <del> start of a section which _a has, but not _b
              </del>  end of a section which _a has, but not _b
              <add> start of a section which _b has, but not _a
              </add>  end of a section which _b has, but not _a
                    for use inside attributes:
              -   start of a section which _a has, but not _b
              -   end of a section which _a has, but not _b
              +   start of a section which _b has, but not _a
              +   end of a section which _b has, but not _a
===================================================================== */

/* =====================================================================
Update: print also the deleted text for A
===================================================================== */

/* =====================================================================
Update on the processing and datastructure for tags and
attributes.
Principles:
- names of tags, attributes and attributevalues will be coded by a
  b-tree
- the information for the attributes will be stored in a record; the
  record will be refered from a node
- the record consists of an array with as many elements as is necessary. The
  array contains triples: (attributename, attributevalue, sequencenumber
  of attr.).
  Within the array the triples will be sorted on their first element.
- It is assumed that both inputfiles conform to the same DTD. That means
  that the number of names of attributes for a specific tag are the
  same. It is also assumed that all the attributes which are specified
  in the DTD are present in the input files (which are non-minimized).
  Therefore an error message will emerge when the number of attributes
  is different for the two nodes which are compared.
  Only the sequence of the attributes and their values may differ, in
  accordance with the standard for SGML.
- Comparison of attributes is performed as follows. If two nodes are
  compared the successive (coded) values in the array are compared
  (they are already sorted on attributename). The datastructure
  "compared_with_type" will get an extra field which will contain the
  result of the comparison.
===================================================================== */

#ifdef MVS
#pragma linkage (revxml,COBOL)
/*
  for mainframe MVS: do a global replace
    [ => ??(	(ascii 5B)
    ] => ??)	(ascii 5D)
    ^ => ??'	(ascii 5E)
*/
#include <Ctype.h>
#endif

#ifndef MVS
#include <fcntl.h>
#endif

#include <string.h>
#include <stdio.h>
#include <setjmp.h>
#include <stdlib.h>
#include <time.h>
#include <limits.h>
#include <sys/stat.h>

#if defined(WIN32)
#include <io.h>
#include <windows.h>
#include <sys/types.h>
#endif

#if defined(unix)
#include <unistd.h>
#endif

#include "revxml.typ"
#include "castatus.h"
#include "licensing.h"
#include "cm.h"
#include "errno.h"

#include "revxml.h"

#if ((!defined(WINDOWS)) && (!defined(WIN32)))
/* Conditional compilation for WINDOWS/WIN32 
   Note: min() declared as function instead of a macro to avoid side-effects.
*/
   static int min(int a, int b) { return ( (a < b) ? a : b); }
#endif

/*  #define LONG_MAX (2147483647)  remove when LONG_MAX is defined in
                                    limits.h */
/* The Sentinel function completed successfully                                    */
#define LS_SUCCESS                   0x0
#define max_symbol_size  (255)
#define maxlen_symbol (11) /* maximum length of a symbol in printout */
#define len_pointers (7) /* maximum length of a pointer in printout */
#define max_length_of_hashed_text (1000) /* maximum hash coded text
      units within a #pcdata */
#define max_length_of_lexeme (25000) /* maximum length of a lexeme; can be
			the length of a complete pcdata (with atomic = pcdata) */

/* types of lexemes in text */
#define lexeme_to_be_coded (1) /* to be coded with a hash_code
																		(if atomic unit <> 1) */
#define lexeme_not_to_be_coded (2) /* not to be coded with a hash_code
																		(if atomic unit <> 1),
																			e.g. a whitespace */
/* end of types of lexemes in text */

enum kind_of_diff_type
	{ diff_open_a_has_more,
		diff_close_a_has_more,
	  diff_open_b_has_more,
		diff_close_b_has_more,
		diff_open_change,
	  diff_close_change}; /* indicates the four different revisionmarkers plus
                        the two artificial change markers */

enum chars_or_node_type	/* indicates of the deleted text of A is to be written as: */
	{
		chars,							/*	a sequence of characters	*/
		node								/*	a complete node */
	};

enum discriminative_type /* indicates if an attribute is: */
	{
		not_discriminative,
		positive_discriminative,
		negative_discriminative
	};

enum to_be_compared_type /* indicates if an attribute is: */
	{
		compare,
		consider_as_equal,
		consider_as_different
	};

/* acting procedure: compress_diffs */
typedef struct compressed_from_type
  { my_byte kind_of_from;                 /* internal code 1, 2 or 3 */
    struct compressed_from_type  *next;  /* next element in the list */
  } compressed_from_type;
typedef compressed_from_type *compressed_from_type_p;

typedef struct diff_of_attrs_type
  { short diffattr[40];   /* contains the numbers of the attributes
                              which differ */
  } diff_of_attrs_type;
typedef diff_of_attrs_type *diff_of_attrs_type_p;


/* acting procedure: reconstruct_diffs */
/* Compared_with_type functions as a record in a list. The list is
   pointed to by the field 'compared_with' of a node. Each record of the
   list describes the result of the comparison of that node with
   another node. */

typedef struct compared_with_type
  { struct node_type            *node_compared_with;
                                  /* node which was compared with */
    short                       nr_diff_of_attrs;
                                /* the number of different attributes */
		boolean                     mixed_nodes;
																	/* the two nodes compared are of mixed mode */
    long                        result_of_comparison,
                                  /* the total result of the comparison,
                                    taking into account the comparison
                                    of either the atomic units or
                                    subnodes and of the attributes */
																start_of_difference_a,
                                  /* if node is a pcdata node: the
                                     number of the first character
                                     in the text for a which differs */
                                start_of_difference_b,
                                  /* if node is a pcdata node: the
                                     number of the first character
                                     in the text for b which differs */
                                end_of_difference_a,
                                  /* if node is a pcdata node: the
                                     number of the last atomic element
                                     in the text for a which differs */
                                end_of_difference_b;
                                  /* if node is a pcdata node: the
                                     number of the last atomic element
                                     in the text for b which differs */
    enum to_be_compared_type		nodes_to_be_compared;
																	/* because of discriminative attributes the nodes may be considered
																		 as equal or as different; the possible values indicate:
																				compare: outcome of comparison not caused by discriminative attributes
																				consider_as_equal: to be considered as equal
																				consider_as_different: to be considered as different
																	*/
		pos_type										next;  /* next element in the list, as a cm pointer */
  } compared_with_type;
typedef compared_with_type      *compared_with_type_p;

typedef struct compared_with_buffer_type
  { char buffer[max_compared_with_buffer_size];   /* contains compared_with, the numbers of the attributes
                              which differ, and the compressed froms */
  } compared_with_buffer_type;
typedef compared_with_buffer_type      *compared_with_buffer_type_p;

/* acting procedure: add_diff */
/* diff_marker acts as a record in a list which contains the difference
   markers together with their positions. The list is constructed in the
   procedure reconstruct_diffs. This procedure selects the nodes which
   are on the path of the lcs. Nodes which are not on that path will not
   be attached and no list with diff_markers will be constructed for
   them. The list is attached to a node of text b. No reference to a
   node of text a has to be kept. The list will be consulted by the
   output procedure in order to produce the correct difference markers
   at the right positions. */
typedef struct diff_marker_type
  { enum kind_of_diff_type      kind_of_diff;
                                  /* the type of difference marker */
		enum chars_or_node_type			chars_or_node_A,
																chars_or_node_B;
																	/* for A and B: concerns a sequence of chars (value = chars)
																		 or a complete node (value = node) */
    long                        position_char_a,
                                  /* if chars_or_node_A == chars then position of a character in buffer_a,
																			or the hash-position, before the relevant token; else 0 */
																position_char_b;
                                  /* if chars_or_node_B == chars then position of a character in buffer_b,
																			or the hash-position, before the relevant token; else 0 */
		struct node_type						*position_node_a, /* position_node_a == 29-10-2002 */
																  /* position_node_a contains the address of the node in A */
																*position_node_b; /* mixed_child_lh_node_b == 29-10-2002 */
																  /* position_node_a contains the address of the node in B */
    struct diff_marker_type     *next;
                                  /* next element in the list */
  } diff_marker_type;
typedef diff_marker_type        *diff_marker_type_p;
typedef diff_marker_type_p			*diff_marker_type_p_p;

/* keeps the position of line-separator in original file data_b */
typedef struct crs_in_original_type
  { long                        position;
                                  /* position in B before the character */
    struct crs_in_original_type *next;
                                  /* next element in the list */
  } crs_in_original_type;
typedef crs_in_original_type    *crs_in_original_type_p;

typedef struct discriminating_attributes_of_element_type
	{	long																								attribute;
	  enum discriminative_type														discrimination;
		struct discriminating_attributes_of_element_type		*next;
	} discriminating_attributes_of_element_type;
typedef discriminating_attributes_of_element_type *discriminating_attributes_of_element_type_p;

/* record in array attr_info */
typedef struct attr_triple_type
  { long                        attr_name;
                                  /* coded name of the attribute */
    long                        attr_value;
                                  /* coded value of the attribute */
    short                       sequence;
                                  /* original sequence number of the
                                     attribute */
		enum discriminative_type					discrimination;
																	/* if true: if the compared values 
																		 for this attribute are different than the
																		 nodes are considered to be different */
  } attr_triple_type;

/* array with info about the attributes of a node */
typedef struct attributes_type
  { attr_triple_type            attr_info[max_symbol_size];
                                  /* array with attr_triple's */
  } attributes_type;
typedef attributes_type         *attributes_type_p;

/* record with info about corresponding shadow_nodes
	 of nodes in A and B */
typedef struct shadow_node_pair_type
  { struct node_type							*node_a,
																	*node_b;
		diff_marker_type_p						*start_of_diff_markers_to_be_replaced_p_p,
																	*end_of_diff_markers_to_be_replaced_p_p;
		struct shadow_node_pair_type	*next;
  } shadow_node_pair_type;
typedef shadow_node_pair_type			*shadow_node_pair_type_p;

/* acting procedures on node: node_diff, complete_current_node,
 write_for_node
 acting procedures on tree of nodes: add_daughter_node, print_tree,
 print_internal_structure */
typedef struct node_type
{  long                       tagcode;
                                /* coded name of the tag */
   char                       tag[max_symbol_size];
                                /* name of the tag itself, as a string,
																   in upper case */
   long                       start_of_tagstring,
                                /* position within the text buffer where
                                   the name of the tag starts */
                              end_of_tagstring;
                                /* position within the text buffer where
                                   the name of the tag ends */
   boolean                    leaf,
                                /* if true: node is a leaf; this is the
                                   case if the node contains #pcdata, a PI, a declaration, 
																	 a comment, or if the node is EMPTY */
					                    pcdata,
																/* if true: the tag contains #pcdata */
															empty,
                                /* if true: the tag is an XML empty one */
															mixed_mode,
																/* if true: contains at least one #pcdata node and one non-#pcdata node */
															node_has_to_be_compared;
	 short                      number_of_attributes;
                                /* number of attributes */
   attributes_type_p          attributes;
                                /* array with info about the attributes */

																/* if true: the node has to be compared with other nodes */
	 diff_marker_type_p         start_diff_marker,
                                /* points to a list of difference markers;
                                   it contains the difference markers
                                   together with their positions. The list
                                   is constructed in the procedure
                                   reconstruct_diffs. This procedure selects
                                   the nodes which are on the path of the
                                   lcs. Nodes which are not on that path
                                   will not be attached and no list with
                                   diff_markers will be constructed for them.
                                   The list is attached to a node of text b.
                                   The list will be consulted by the
                                   output procedure in order to produce the
                                   correct difference markers at the right
                                   positions. */
															current_diff_marker;
																/* points to a member of the list of difference markers;
																	 in use when mixed_mode is true, because the members may
																	 apply to all children nodes */
   crs_in_original_type_p     start_crs_in_original,
                                /* points to a list which contains the
                                   relative positions of the carriage returns
                                   in the text for the current b_node */
                              last_crs_in_original;
                                /* points to the end of the list */
   pos_type							      compared_with;
                                /* points to a list in CM which describes the
                                   result of the comparison of
                                   the current node with another node. */
	 struct node_type           *up,
                                /* points to the father node */
                              *former,
                                /* points to the node immediately left
                                   of the current node */
                              *next;
                                /* points to the node immediately right
                                   of the current node */
	 long												sequence_nr_of_node,
																/* the sequence number of being the child of the
																	 parent node */
															number_of_atoms,
																/* the number of child nodes + the number of atomic elements 
																	 within contained pcdata nodes */
															number_of_pcdata_nodes,
																/* the number of contained pcdata nodes */
															number_of_nonpcdata_nodes;
																/* the number of contained nonpcdata nodes */
   union
    {  struct
        { /* applicable if node is a leaf */
        long                  start_of_text,
                                /* start position of the text in the
                                   text buffer for a (data_a) or for b
                                   (data_b) */
                              end_of_text,
                                /* end position of the text */
                              length_of_text,
                                /* length of the text = end_of_text -
                                   start_of_text + 1 */
                              length_of_hash_coded_text,
                                /* if the text of the node is hash coded:
                                   its length */
                              *hashcode_text;
                                /* if the text of the node is hash coded:
                                   points to an array with the hash
                                   codes of the text */
        } t;
       struct
        { /* applicable if node is not a leaf */
        struct node_type			*down_start,
                                /* start of the list with daughter nodes */
                              *down_end;
                                /* end of the list with daughter nodes */
        long                  nr_of_nodes_down;
                                /* number of daughter nodes */
        } f;
    } u;
  } node_type;
 typedef node_type *node_type_p;

/* definition of a node of a binary tree */
/* acting procedures: found_in_tree, add_to_tree */
typedef struct btnode
  { char            symbol_string[max_symbol_size]; /* key */
    long            code; /* code of the key, to be assigned
                              incrementally */
    struct          btnode *left,
                           *right; /* pointers to subtrees */
  } *btree;

/*  data_a and data_b will receive the text of the files in the
    procedure readfile_from_buffer;
    data_c returns the text of data_b with revision markers;
    cr_data returns the positions of the cr's in data_b */

static char *data_a,   /* text of old data */
            *data_b,   /* text of new data */
            *data_c,   /* text of revised data */
            eol_char,  /* the end of line character that has to be
                          written in the new text with difference
                          markers */
						cache_filename[255] = "revxml.cm",	/* name of the cache file */
						def_filename[255], 		/* filename of file with settings, by default "revxml.def" */
						emp_filename[255], 		/* filename of file with empty elements and discriminative 
																										attributes, by default "revxml.emp" */
						log_filename[255], 		/* filename of logfile */
						lic_filename[255]; 		/* filename of noSentinel license file */

static long len_a,   /* length of the text in data_a */
            len_b,   /* length of the text in data_b */
            cr_data_len,   /* length of the array cr_data */
            *cr_data,   /* array which contains the positions of the
                           cr's in data_b */
            len_c;  /* length of the text in data_c, to be returned in the parameter return_len_c */
static jmp_buf exception_status;  /* ANSII C construction:
                          exception_status will be defined
                          later as a program address to which is jumped
                          if an exceptional status occurs */
static long result;   /* error code; to be assigned before a jump is
                          made to exception_status */
static my_int atomic_unit_of_text;  /* possible values: rev_chars,
                                 rev_words, rev_sentences, rev_pcdata */

static long *d_value;  /* array at which the matrices d will be
                          projected */
static my_byte  *d_comes_from;  /* array at which the matrices from will be
                               projected */
static long alloc_size, /* length of the stack for simulated alloc's */
						d_size,   /* size of the arrays d_value and d_comes_from */
						max_nr_of_names_of_elements,  /* maximum number of names for tags */
						max_nr_of_names_of_attributes,  /* maximum number of names for tags */
						max_nr_of_names_of_attribute_values;  /* maximum number of names for tags */

static compressed_from_type_p compr_from_last_p;  /* pointer to tail of
                        the list compressed_from of the current node */
char helpline[255],
		 templine[255];   /* reads lines from the 'def' file */
static FILE *infile;  /* used for the 'empty' files  */
static FILE *deffile;  /* used for the 'def' file */

static long bufferp,   /* pointer in the current text buffer */
            bufferp_char_total_a,   /* pointer to the last position in the
                                  text buffer for a */
            bufferp_char_total_b;   /* pointer to the last position in the
                                  text buffer for b */
static long start_of_diff_a,   /* for the current node: the
                                  number of the first character
                                  in the text for a which differs  */
            start_of_diff_b;   /* for the current node: the
                                  number of the first character
                                  in the text for b which differs */
static boolean  closed_for_b,      /* no revision tag for b is open */
            closed_for_a,      /* no revision tag for a is open */
            trace,             /* if true: a tracing is required */
						combtags,					 /* if true: combine revision tags */
						ignore_diff_in_tags,/* if true: ignore the difference in tag names */
						print_a,					 /* if true: print the deleted text of A */
						compare_pi,				 /* if true: compare processing instructions */
						compare_comment,	 /* if true: compare comments */
						preproc,					 /* if true: preprocess with atomic_unit_of_text=pcdata */
            tag_change;				 /* if true: a 'change' sequence is detected */
static long free_d,  /* stack pointer for d_value and d_comes_from */
            distance,   /* the resulting distance between the two top
                           nodes for a and b */
            total_allocated,    /* statistic for the total amount of
                                   bytes that has been allocated */
            total_d;   /* statistic for the total amount of space for
                d_value and d_comes_from that has actually been used */
static node_type_p top_node_a,    /* the top node for text a */
                   top_node_b,    /* the top node for text b */
                   current_node;    /* the current node which is
                                       processed */
static long barposition[40]; /* a global stack for the procedure
                            'printstack', which contains the number of
                             spaces in between two vertical bars in the
                             printout */
static char barsigns[40]; /* contains a ' ' or a '|' */

static char char_end[20];   /* set of separator characters in between
                  atomic units of text; if the atomic unit is #pcdata
                  then char_end contains only x\hh, the highest possible
                  character; it is assumed that this character will
                  not appear in the text; therefore, the algorithm
                  proceeds in the text indefinitely until position_end
                  is reached */

static btree	root_of_element_names, /* top of btrees for names  */
							root_of_attribute_names,
							root_of_attribute_values;

static long		last_code_of_element_names, /* code subsequently the names in the btrees */
							last_code_of_attribute_names,
							last_code_of_attribute_values;

static btree  *names_of_elements, /* for the coding of names */
							*names_of_attributes,
							*names_of_attribute_values;

static discriminating_attributes_of_element_type_p *discriminating_attributes_of_element_p;
static boolean *empty_element_p;

#define symbol_is_element (2)
#define symbol_is_attribute (3)
#define symbol_is_attribute_value (4)

short diff_tag_open_a,
			diff_tag_open_b;   /* signals if a revision tag is open;
                                is used in order to put the revisiontags
                                as close as possible around a lexeme */
static short after_an_end_revtag_nothing_has_been_written;
															/* signals if after an endtag anything has been written;
																 used in the simulation of the printing of nodes */
static char tag_open_for_a_has_more[20], /* as present in the def file */
          tag_close_for_a_has_more[20],  /* as present in the def file */
          tag_open_for_b_has_more[20],   /* as present in the def file */
          tag_close_for_b_has_more[20],  /* as present in the def file */
          tag_open_change[20],           /* as present in the def file */
          tag_close_change[20],          /* as present in the def file */
          tag_open_for_change_attributes[20],  /* as present in the def file */
          tag_close_for_change_attributes[20]; /* as present in the def file */
static int  length_of_open_revision_tag, /* length of
                tag_open_for_a_has_more, tag_open_for_b_has_more and
                tag_open_change */
            length_of_close_revision_tag; /* length of
                tag_close_for_a_has_more, tag_close_for_b_has_more and
                tag_close_change */

static enum kind_of_diff_type type_of_last_open_revision_tag;
      /* type of the last revision open tag that has been added to the
         output buffer */
static long position_of_last_open_revision_tag,
              /* position of the last revision open tag that has been
                 added to the output buffer */
            position_of_last_close_revision_tag;
              /* position of the last revision close tag that has been
                 added to the output buffer */
static diff_marker_type_p *start_of_diff_markers_to_be_replaced_p_p,
													*end_of_diff_markers_to_be_replaced_p_p;
							/* pointers to the diff_markers in preprocessing comparison
								 that have to be replaced during subsequent comparison */

static node_type_p start_of_last_rev_a, /* pointer to childnode in A where the
																		last difference tag started */
						end_of_last_rev_a,   /* pointer to childnode in A where the
																		last difference tag ended */
						start_of_last_rev_b, /* pointer to childnode in B where the
																		last difference tag started */
						end_of_last_rev_b;   /* pointer to childnode in B where the
																		last difference tag ended */

static shadow_node_pair_type_p
						all_shadow_nodes_pair_p,	 /* points to a list which contains corresponding
																	  ranges of children nodes in A and B, detected
																	  with an initial compare in mode=pcdata */
						active_shadow_nodes_p;/* points to an element in all_shadow_nodes_pair_p */

static long textsize_global;  /* textsize specified in def-file */
static long alloc_ptr;    /* stackpointer of the stack for simulated alloc's */
static char *alloc_base;  /* start of the stack for simulated alloc's */
static pos_type filepos_null = {0, 0};

static string80 cache_file; 	/* cache file name */
static long APS = 80;  /* pagesize for cache memory buffer;
										1 page will be of size APS * 512 bytes; 
										because Windows uses a buffersize of 4096 it seems that
										APS = a multiplicity of 8 is a good choice; 
										may be overruled by def-file */
static int cache_page_size; /* will be APS*512 */
							 
static my_int ABS = 10; /* buffersize for cache memory buffer;
										the buffer will be of size ABS * APS * 512 bytes; ABS may be small
										because access to the cache file is more or less sequential;
										ABS = 10 seems to be a good choice	*/ 

#define BLANK	' ' 
static FILE			*outfile;
static char			filename_a[255], filename_b[255], filename_c[255];
static size_t		
								textsize,   /* size of a text buffer to be allocated */
								crsize;

char						buffer_sprintf[255];	/* buffer for printing messages */
my_int					keep_atomic_unit_of_text;	/* temp storage for atomic_unit_of_text */
boolean					keep_combtags;						/* temp storage for combtags */
long						return_code;
long						eol_proc;							/* boolean: true if end-of-lines in file B should be kept in file C */
long						code_of_arb_element;	/* the arb-element is '*'; may occur in the .emp file */
long						code_of_arb_attribute;	/* the arb-attribute is '*'; may occur in the .emp file */


static void short_of_memory(void);
static void add_to_buffer_c(char *s, int l);
static void start_endtag_failure(char tag[max_symbol_size], char * textspace_p,
																 char * A_or_B);
static int repl(char a, char data_b);
static int repl_char(char a, char b);
static int repl_long(long a, long b);
static long maxim3(long k, long l, long m, my_byte *from);
static long d(long start_of_d, long i, long j, long n);
static my_byte d_from(long start_of_d, long i, long j, long n);
static void store_in_d_space(long start_of_d, long i, long j, long n,
                  long v, my_byte f);
static void complete_current_node(char * textspace_p, node_type_p current_node, long end_of_text_in_buffer);
static long found_in_tree(btree root, btree *parent, char *theval);
static long add_to_tree(char *theval, int type_of_name);
static int compare_long(long *p, long *q);
static int compare_short(short *p, short *q);
static node_type_p add_daughter_node(char * textspace_p, char * tagnew,
                                                    boolean node_is_leaf);
static void write_crs(long position,
                  crs_in_original_type_p *current_crs_in_original,
                  long *current_cr_position);
static void getsym(char * textspace_p, long * position,
                  long position_end,
                  char * lexeme, int * type_of_lexeme, int * end_of_text);
static void print_tree(node_type_p top_node, long  top_barposition,
                  char *textspace);
static void print_tree_trace(node_type_p top_node, long top_barposition);
static void print_node_trace(node_type_p top_node);
static void print_internal_structure(node_type_p top_node);
static void readfile_from_buffer(char * textspace_p, long eol_proc,
                  long len, char * A_or_B);
static void write_diff_tag(diff_marker_type_p current_diff_marker_p);
static void write_diff_tags_on_node_level(node_type_p node_b,
										 diff_marker_type_p *current_diff_marker_p,
										 enum chars_or_node_type chars_or_node_a);
static void write_diff_tags_on_text_level(long position_b,
										 node_type_p current_node,
										 diff_marker_type_p *current_diff_marker_p);
static void write_text_for_node(node_type_p current_node);
static void write_for_node_a(node_type_p father_node);
static void write_for_node(node_type_p father_node);
static void write_to_outputbuffer(void);
static void add_compressed(my_byte kind_of_from);
static void compress_diffs(char *compressed_array_p, long max_array, 
										long *size_of_compressed_array,
										long start_of_d, long i, long j);
static long node_diff(node_type_p node_a, node_type_p node_b);
static void add_diff(enum kind_of_diff_type kind_of_diff,
							enum chars_or_node_type chars_or_node_A,
							enum chars_or_node_type chars_or_node_B,
 							node_type_p father_node_a,
              node_type_p father_node_b,
							node_type_p current_node_a_right,
							node_type_p current_node_b_right,
							long bufferp_char_a,
							long bufferp_char_b);
static void reconstruct_diffs(node_type_p node_a, node_type_p node_b,
                  short remember_diff_attrs);
static char upcase(char kar);

/* This function returns the size of a file, in bytes. This function may not be portable
   for every operating system.
*/
static long getFileSize(FILE *f) {
long filesize;
#if defined(WIN32) || defined(unix) 
     {    struct stat statbuf;
          fstat(fileno(f), &statbuf);
          filesize = (long)statbuf.st_size;
     }
#else
     /* An inefficient solution, using the standard library, would be:
          - save the current position of the file using ftell():      long savepos = ftell(f);
          - goto the end of the file using fseek():                   fseek(f, 0L, SEEK_END);
          - ask the current postion:                                  long fileLength = ftell(f);
          - reposition the file to the saved position:                fseek(f, savepos, SEEK_SET);
          - return the length:                                        return fileLength;
     */
     filesize = -1; /* makes the caller aware that something went wrong */
#endif
return filesize;
} /* getFileSize() */

compared_with_type_p filepos_to_ptr(filepos, bufslot)
pos_type  *filepos;
long *bufslot;
/* translate a position in the external file to a physical ptr into the internal buffer */
{ 	compared_with_type_p rampos; 				/* the physical ptr in the buffer */
		*bufslot = -1;
		if ((filepos->blocknr == 0) && (filepos->pos_in_block == 0))
		{ 
			rampos = NULL;
		}
		else
		{
			incmbuffer(cm_cache_p, filepos->blocknr, bufslot);
			rampos = (compared_with_type_p)(&((*cm_cache_p->buf.blocks_p[*bufslot])[filepos->pos_in_block]));
		}
		return rampos;
} /* filepos_to_ptr() */


#include "sgd_trac.inc"

#ifndef mallocs_from_array
/* =====================  my_malloc  ====================================
Purpose:      is malloc plus check on NULL pointer
Parameters:
  input:      number of bytes to be allocated
----------------------------------------------------------------------*/
void *my_malloc(size_t nbytes)
  {
		char * handle;
    handle = malloc(nbytes);
    if (handle == NULL)
    {
      short_of_memory();
    }
    return handle;
  } /* my_malloc */

/* =====================  my_calloc  ====================================
Purpose:      pseudo calloc by taking space from a previously
              allocated stack
Parameters:
  input:      number of objects, size of an object
  return:     the new absolute stack pointer
Algorithm:    increase the stack pointer alloc_ptr by nbytes
----------------------------------------------------------------------*/
void *my_calloc(size_t nobj, size_t objectsize)
  { char *handle;

		handle = calloc(nobj, objectsize);
    if (handle == NULL)
    {
      short_of_memory();
    }
    return handle;
  } /* my_calloc */

#else

/* =====================  my_malloc  ====================================
Purpose:      pseudo malloc by taking space from a previously
              allocated stack
Parameters:
  input:      number of bytes to be allocated
  return:     the new absolute stack pointer
Algorithm:    increase the stack pointer alloc_ptr by nbytes
----------------------------------------------------------------------*/
void *my_malloc(size_t nbytes)
  {
		char * handle;
    handle = (char *)((long)alloc_base + alloc_ptr);
    alloc_ptr += nbytes;
    if (alloc_ptr > alloc_size)
    {
			handle = NULL;
      short_of_memory();
    }
    return handle;
  } /* my_malloc */

/* =====================  my_calloc  ====================================
Purpose:      pseudo calloc by taking space from a previously
              allocated stack
Parameters:
  input:      number of objects, size of an object
  return:     the new absolute stack pointer
Algorithm:    increase the stack pointer alloc_ptr by nbytes
----------------------------------------------------------------------*/
void *my_calloc(size_t nobj, size_t objectsize)
  { char *handle, *help_handle;
 		long n, nbytes;

		handle = (char *)((long)alloc_base + alloc_ptr);
		nbytes = nobj * objectsize;
    alloc_ptr += nbytes;
    if (alloc_ptr > alloc_size)
    {
			handle = NULL;
      short_of_memory();
    }
		/* init at 0 */
		help_handle = handle;
		for (n = 0; n < nbytes; n++)
		{
			*(help_handle++) = '\0';
		}
    return handle;
  } /* my_calloc */

#endif


/* =====================  upcase  ====================================
Purpose:    returns the uppercase representation of a lower case
            character a..z
Parameters:
  input/output:     kar = the character to be converted
Algorithm:
  for ASCII: if kar in a..z: subtract 32
  for EBCDIC: has been determined by customer
----------------------------------------------------------------------*/
char upcase(char kar)
  {
#ifndef MVS
		if ( (kar >= 97) && (kar <= 122) )  /* make lowercase uppercase */
		{
      kar = kar - 32; /* note: ignore a possible compiler warning on
                                      the loss of significant bits */
		}
#endif
#ifdef MVS
    /* take care of EBCDIC; we have to consult the translation table of
    ASCII to EBCDIC; in the meantime we do not perform an upcase */
    kar = kar;
#endif
    return kar;
  } /* UpCase */

/* =====================  short_of_memory  ===========================
Purpose:  called when a dynamic allocation fails
          returns from revxml with message code
          not_enough_memory_on_heap
Parameters:       none
Global variables: *result
Algorithm:    a long jump is made to the postion where "exception_status"
              is defined
----------------------------------------------------------------------*/
void short_of_memory(void)
{
  sprintf(buffer_sprintf, "Sorry, not enough memory on heap. See logfile.\n");
  fprintf(logfile, "Not enough memory. Up till now there has been used:\n");
  fprintf(logfile, " of d space: %d   allocated on heap: %d bytes\n",
									 total_d, alloc_ptr);
  fprintf(logfile, "Increase the heapsize in def-file\n");
  result = not_enough_memory_on_heap;
  longjmp(exception_status, not_enough_memory_on_heap);
}


void error_file_open(FILE *openfile, char *filename, long *result)
{
  if (openfile == NULL)
	{
	  sprintf(buffer_sprintf, "The file %s could not be opened. Program terminated", filename);
		if (logfile)
			fprintf(logfile, "%s\n", buffer_sprintf);
		*result = file_not_opened;
		longjmp(exception_status, file_not_opened);
	}
}


/* =====================  start_endtag_failure  ======================
Purpose:  called when for an endtag no corresponding starttag can be
          found
Parameters:
  input:        tag
Global variables:   *result
Algorithm:    a long jump is made to the postion where
              "exception_status" is defined
----------------------------------------------------------------------*/
void start_endtag_failure(char tag[max_symbol_size], char * textspace_p, char * A_or_B)
{
	char temp[100];

	sprintf(buffer_sprintf,
    "At offset %ld in the textbuffer for %s; surrounding text:\n",
		bufferp, A_or_B);
	fprintf(logfile, buffer_sprintf);

	strncpy(temp, (char *) textspace_p+bufferp-50, 100);
	temp[100] = '\0';
	sprintf(buffer_sprintf, "%s\n", temp);
	fprintf(logfile, buffer_sprintf);

	sprintf(buffer_sprintf,
    "  No corresponding starttag for endtag %s; see logfile.\n", tag);
	fprintf(logfile, buffer_sprintf);

  result = no_corresponding_starttag;
  longjmp(exception_status, no_corresponding_starttag);
}

/* =====================  add_to_buffer_c  =======================
Purpose:          a string is added to the output buffer
Parameters:
  input:          *s: the string;
                  l: length of the string
  output:         none
Global variables: len_c
Algorithm:
----------------------------------------------------------------------*/
void add_to_buffer_c(char *s, int l)
{
  strncpy(data_c + len_c, s, l);
  len_c += l;
} /* add_to_buffer_c */

/* =====================  add_to_buffer_c_with_crs  ==================
Purpose:          a string is added to the output buffer together
									with appropriate CR's
Parameters:
  input:          *s: the string;
                  l: length of the string
				          position : current position in the text of the
                             current node
  input/output:   current_crs_in_original : current element in the list
                                            of crs in the current node
                  current_cr_position     : the position which is stored
                                            in current_crs_in_original
Global variables: len_c
Algorithm:        position is compared with current_crs_in_original,
                  iteratively in subsequent elements of the list of crs;
                  as long as position is larger an eol character is
                  written to the output buffer
----------------------------------------------------------------------*/
void add_to_buffer_c_with_crs(char *s, int l, long position,
                  crs_in_original_type_p *current_crs_in_original_p,
                  long *current_cr_position_p)
{
	my_int i;
	for (	i = 0;
				i < l;
				i++)
	{
		write_crs(position + i, current_crs_in_original_p, current_cr_position_p);
		add_to_buffer_c(s + i, 1);
	}; /* for i */
} /* add_to_buffer_c */


/* =====================  repl_char  ===============================
Purpose:           returns a 1 if a = b, else a 0
Parameters:
  input:          characters a and b
----------------------------------------------------------------------*/
int repl_char(char a, char b)
{ int r;
    if (a == b)
        r = 1;
    else
        r = 0;
    return r;
  }  /* repl_char */

/* =====================  repl_long  ================================
Purpose:          returns a 1 if a = b, else a 0
Parameters:
  input:          longs a and b
----------------------------------------------------------------------*/
int repl_long(long a, long b)
{ int r;
    if (a == b)
        r = 1;
    else
        r = 0;
    return r;
}  /* repl_long */

/* =====================  maxim3  ===================================
Purpose:            - returns the maximum of k, l and m
                    - indicates in 'from' which variable was the maximum:
                      if k then from := 1
                      if l then from := 2, also if k = l
                      if m then from := 3, also if k = m or l = m
Parameters:
  input:          longs k, l, m
  input/output:   *from
Algorithm:        first, in m2 the maximum is taken of k and l;
                  then, in m3, the maximum is taken of m2 and m

----------------------------------------------------------------------*/
long maxim3(long k, long l, long m, my_byte *from)
  {
		long m2, m3;

    if (k > l)
    {
      m2 = k;
      *from = 1;
    }
    else /* then k <= l */
    {
      m2 = l;
      *from = 2;
    };

    if (m2 > m)
		{
      m3 = m2;
		}
    else /* then m2 <= m */
    {
      m3 = m;
      *from = 3;
    };
    return m3;
  }

/* =====================  d  ======================================
Purpose:          the matrix d is projected upon the linear array
                  d_value; because of recursive calls
                  more matrices can be projected upon d_value;
Parameters:
  input:          start_of_d:  the start of the projection of the
                                current d in d_value
                  i: number of column in d
                  j: number of row in d
                  n: the number of columns in d
  returns:       d[i, j]
----------------------------------------------------------------------*/
long d(long start_of_d, long i, long j, long n)
{ /*d*/
    long h;
    h = start_of_d + i * (n + 1) + j;
    return d_value[h];
} /*d*/

/* =====================  d_from  ===================================
Purpose:          the matrix d_from is projected upon the linear array
                  d_comes_from; because of recursive calls
                  more matrices can be projected upon d_comes_from;
Parameters:       see the description of the function d
----------------------------------------------------------------------*/
my_byte d_from(long start_of_d, long i, long j, long n)
  { /*d_from*/
    long h;
    h = start_of_d + i * (n + 1) + j;
    return d_comes_from[h];
  } /*d_from*/

/* =====================  store_in_d_space  =========================
Purpose:          stores v = d[i, j] and f = d_from[i, j] in their
                  respective projected arrays
Parameters:
  input:          start_of_d:  the start of the projection of the
                                current d in d_value
                  i: number of column in d
                  j: number of row in d
                  n: the number of columns in d
                  v: the value of d[i, j] which has to be stored
                  f: the value of d_from[i, j] which has to be stored
----------------------------------------------------------------------*/
void store_in_d_space(long start_of_d, long i, long j, long n, long v,
                                                              my_byte f)
{ /*store_in_d_space*/
    long h;
    h = start_of_d + i * (n + 1) + j;
    d_value[h] = v;
    d_comes_from[h] = f;
} /*store_in_d_space*/

/* =====================  getsym  ===================================
Purpose:          returns the next lexeme from the text which belongs
                  to current_node; will be called when the atomic unit
                  of text is a word, a sentence or #pcdata. The lexeme
                  can be either (1) an atomic unit or (2) a sequence of
                  separators which will not be coded. Separators are:
                  spaces and tabs.
Parameters:
  input:          textspace_p     : pointer to the start of the
                                    global textspace
                  position_end    : pointer to the end of the text for
                                    this call to getsym
  output:         lexeme          : the lexical unit in the text which
                                    has to be returned
                  type_of_lexeme  : a classification of lexeme, out of:
                                      lexeme_to_be_coded (1)
                                      lexeme_not_to_be_coded (2)
                  end_of_text     : a boolean which becomes true when
                                    position >= position_end
  input/output:   position        : pointer to the current character
                                    in the global textspace
Global variables: char_end        : separator character in between
                  atomic units of text; if the atomic unit is #pcdata
                  then char_end = 255, the highest possible
                  character; it is assumed that this character will
                  not appear in the text; therefore, the algorithm
                  proceeds in the text indefinitely until position_end
                  is reached; if the atomic unit is a word then char_end
                  is a space; if it is a sentence then char_end is a '.'
Algorithm:
      3 states are distinguished:
        - state 1 is at the start of getsym; if the current character
          is char_end then we are at the start of a lexeme that we
          do not have to code; we proceed with state 2; else we proceed
          with state 3
        - state 2 continues to read symbols equal to char_end
        - state 3 continues to read symbols unequal to char_end
----------------------------------------------------------------------*/
void getsym(char * textspace_p, long * position, long position_end,
            char * lexeme, int * type_of_lexeme, int * end_of_text)
  {
		int nextstate;
		boolean read_next, proceed;
    long position_start;
    char ch;

    read_next = true;
    proceed = true;
    nextstate = 1;
    lexeme[0] = 0;
    position_start = *position;
    ch = textspace_p[position_start];

		while (proceed)
		{
			switch (nextstate)
			{ case 1:
					if (strrchr(char_end, ch) != NULL) /* ch is in char_end */
					{
						*type_of_lexeme = lexeme_not_to_be_coded;
						nextstate = 2;
					}
					else
					{
						*type_of_lexeme = lexeme_to_be_coded;
						nextstate = 3;
					};
					break;
				case 2:
					if (strrchr(char_end, ch) == NULL) /* ch is not in char_end */
					{
						proceed = false;
						read_next = false;
					};
					break;
				case 3:
					if (strrchr(char_end, ch) != NULL) /* ch is in char_end */
					{
						proceed = false;
						if (atomic_unit_of_text == rev_sentences)
							/* treat '.' as part of the lexeme; change 24-6-94 */
							read_next = true;
						else
							read_next = false;
					};
					break;
			}; /* switch */

			if (read_next)
			{
				(*position)++;
				if (*position > position_end)
				{
					*end_of_text = true;
				}
				else
				{
					ch = textspace_p[*position];
				}
				if (*end_of_text)
				{
					proceed = false;
				}
			};
		}; /* end while proceed */
		strncpy(lexeme,
						textspace_p + position_start,
            *position - position_start);
    lexeme[*position - position_start] = 0;
  } /* getsym */


void calculate_hash_codes_for_text(char * textspace_p, node_type_p current_node)
{
  long h, i, position, position_end, amount_of_information,
                                            hashed_text_index;
  char lexeme[max_length_of_lexeme];
  int type_of_lexeme, end_of_text, len;
  long hashed_text[max_length_of_hashed_text];

	switch (atomic_unit_of_text)
  /* if required, calculate hashcode over the text */
  { case rev_words:
    case rev_sentences:
    case rev_pcdata:
        position = current_node->u.t.start_of_text;
        position_end = current_node->u.t.end_of_text;
        type_of_lexeme = lexeme_to_be_coded;
        end_of_text = false;
        hashed_text_index = 0;

        while (!end_of_text)
        {
					getsym(textspace_p, &position, position_end,
                 lexeme, &type_of_lexeme, &end_of_text);

          /* bug test, remove later */
					if ((type_of_lexeme == lexeme_not_to_be_coded) && (strlen(lexeme) > 5))
          {
						sprintf(buffer_sprintf, "lexeme_not_to_be_coded is too long: %s\n", lexeme);
            fprintf(logfile, "lexeme_not_to_be_coded is too long\n");
            result = hashed_text_too_long;
            longjmp(exception_status, hashed_text_too_long);
          }; /* end bug test */


          if (type_of_lexeme == lexeme_to_be_coded)
          {
						
						/* calculate the hashcode of lexeme; we use the
                algorithm of G.H. Gonnet & R. Baeza-Yates,
                "Handbook of Algorithms and Data
                Structures", page 47-48. We do not take the
                module of a prime, because we do not store
                the text in a separate hash table. The
                hash-code serves only to identify strings
                which differ from each other */
            h = 0;
            len = strlen(lexeme)-1;
            for ( i = 0; i <= len; i++)
              h = 131*h + lexeme[i];
            hashed_text[hashed_text_index++] = h;
            if (hashed_text_index > max_length_of_hashed_text)
            {
							sprintf(buffer_sprintf, "Hashed text too long.\n");
              fprintf(logfile, "Hashed text too long. Recompile with a larger max_length_of_hashed_text\n");
              result = hashed_text_too_long;
              longjmp(exception_status, hashed_text_too_long);
            };
          };
        }; /* end_of_text */
        /* store hash coded text and its length in an allocated
          field of current_node */
        current_node->u.t.length_of_hash_coded_text = hashed_text_index;
        /* allocate only the necessary amount of memory */
        amount_of_information = hashed_text_index*sizeof(long);
        if (amount_of_information > 0)
        {
          current_node->u.t.hashcode_text = my_calloc(1,
                                        amount_of_information);
          /* copy only the necessary amount of information from
            hashed_text to current_node->u.t.hashcode_text */
          memcpy(current_node->u.t.hashcode_text,
                            hashed_text, amount_of_information);
				};
				break;
	}; /* switch atomic_unit_of_text */
} /*calculate_hash_codes_for_text*/


/* =====================  complete_current_node  ======================
Purpose:          called when the end of a leaf node is reached;
                  stores the end position of the text and its length
Parameters:
  input:          textspace_p  : pointer to start of the global
                                  textspace for text a or for text b
Global variables: current_node : pointer to the current node
Algorithm:
      - store in the current node the length of the text and the pointer
        to the end of the text
      - if the atomic unit of text is not character
        - calculate an hashcode over the text
        - store hash coded text and its length in an allocated
          field of current_node
----------------------------------------------------------------------*/
void complete_current_node(char * textspace_p, node_type_p current_node, long end_of_text_in_buffer)
{
  if (current_node->leaf)
  {
    current_node->u.t.end_of_text = end_of_text_in_buffer /* bufferp-2 */;
    current_node->u.t.length_of_text =
      current_node->u.t.end_of_text - current_node->u.t.start_of_text + 1;
		calculate_hash_codes_for_text(textspace_p, current_node);
	}
	/* - if #pcdata:  - add #chars to the father node
										- add 1 to the #pcdata_nodes of the father node
		 - make mixed_mode true if the current node contains one or more pcdata node
										and one or more non-pcdata nodes
	*/
	if (current_node->pcdata)
	{ 
		if (atomic_unit_of_text == rev_chars)
			(current_node->up)->number_of_atoms += current_node->u.t.length_of_text;
		else
			(current_node->up)->number_of_atoms += current_node->u.t.length_of_hash_coded_text;
		(current_node->up)->number_of_pcdata_nodes++;
	}
	else
		(current_node->up)->number_of_nonpcdata_nodes++;
	if ( (current_node->number_of_pcdata_nodes >= 1) &&
			 (current_node->number_of_nonpcdata_nodes >= 1) )
		current_node->mixed_mode = true;
	current_node->number_of_atoms += current_node->number_of_nonpcdata_nodes;
}


/* =====================  found_in_tree  =============================
Purpose:          looks if the string 'theval' is present in the binary
                  tree with root 'root' and returns its code
Parameters:
  input:            root   :  start of the binary tree
                    theval :  string which has to be found
  output:           parent :  if not found: node which should be the
                                parent of the node in which theval should
                                have been present; if found_in_tree is used
                                within add_to_tree then parent can be used
                                immediately in order to attach a new node
                              if found: node which is the parent of
                                the node in which theval is present
                    return :  if not found: 0
                              if found: the code
Algorithm:        straightforward search procedure in a binary tree
----------------------------------------------------------------------*/
long found_in_tree(btree root, btree *parent, char *theval)
  { int found;
    btree tptr;
    long result;

    tptr = root;
    *parent = NULL;
    found = false;
    result = 0;
    while ( (tptr != NULL) && (found == false) )
    {
			if (strcmp(tptr->symbol_string, theval) == 0)
      {
				found = true;
        result = tptr->code;
      }
      else
      {
				*parent = tptr;
        if (strcmp(tptr->symbol_string, theval) > 0)
            tptr = tptr->left;
        else
            tptr = tptr->right;
      };
    };
    return result;
  } /* found_in_tree */

/* =====================  add_to_tree  ===============================
Purpose:          looks if the string 'theval' is present in the binary
                  tree with root 'root'; if not it attaches a new node
                  in the tree with a new code;
                  in either case returns the code of 'theval'
Parameters:
  input:          root                : start of the binary tree
                  theval              : string which has to be found
  output:         return              : the code
  input/output :  last_assigned_code  : last value that has been assigned
                                        to a string in this binary tree
Global variables: names_of_tags_and_attributes : array with references
                                        to nodes in binary trees; indexed
                                        by a code of a string
Algorithm:
    - lookup the word in the tree with the function found_in_tree
    - if not found
      - allocate a new node; store the string in the node
      - increment last_assigned_code and store its value in the node
      - store a reference to the node in names_of_tags_and_attributes
      - attach the new node to the parent node which was returned
        by found_in_tree
----------------------------------------------------------------------*/
long add_to_tree(char *theval, int type_of_name)
{
  btree parent, nptr;
	btree *names_ptr, *root_ptr;
  long result_search, max_nr_of_names;
	long *last_assigned_code_ptr;

	switch (type_of_name)
  { /*case symbol_is_empty_element:
      root_ptr = &root_of_empty_element_names;
			last_assigned_code_ptr = &last_code_of_empty_element_names;
			names_ptr = names_of_empty_elements;
			max_nr_of_names = max_nr_of_names_of_empty_elements;
			break;*/
		case symbol_is_element:
      root_ptr = &root_of_element_names;
			last_assigned_code_ptr = &last_code_of_element_names;
			names_ptr = names_of_elements;
			max_nr_of_names = max_nr_of_names_of_elements;
			break;
		case symbol_is_attribute:
      root_ptr = &root_of_attribute_names;
			last_assigned_code_ptr = &last_code_of_attribute_names;
			names_ptr = names_of_attributes;
			max_nr_of_names = max_nr_of_names_of_attributes;
			break;
		case symbol_is_attribute_value:
      root_ptr = &root_of_attribute_values;
			last_assigned_code_ptr = &last_code_of_attribute_values;
			names_ptr = names_of_attribute_values;
			max_nr_of_names = max_nr_of_names_of_attribute_values;
			break;
	}
 
	if (strlen(theval) > max_symbol_size)
  {
		sprintf(buffer_sprintf, "Name or value too long. See logfile.\n");
		fprintf(logfile, "The name or value %s is longer than %d\n",
					*theval, max_symbol_size);
    result = name_too_long;
    longjmp(exception_status, name_too_long);
  };

  result_search = found_in_tree(*root_ptr, &parent, theval);
  if (result_search == 0)
  {
    nptr = my_malloc(sizeof(*nptr));
    nptr->left = NULL;
    nptr->right = NULL;
    nptr->code = ++*last_assigned_code_ptr;
    names_ptr[*last_assigned_code_ptr] = nptr;
                                          /* refer to name */
		if (*last_assigned_code_ptr >= max_nr_of_names)
		{
			sprintf(buffer_sprintf, "Too many names of tags and attributes. See logfile.\n");
			fprintf(logfile, "Too many names of tags and attributes. Increase max_nr_of_names\n");
			fprintf(logfile, " in def-file, which is now %d\n", max_nr_of_names);
			result = too_many_names;
			longjmp(exception_status, too_many_names);
		}
		else
			strcpy(nptr->symbol_string, theval);

    if (*root_ptr == NULL)
        *root_ptr = nptr;
    else
      if (strcmp(parent->symbol_string, theval) > 0)
          parent->left = nptr;
      else
          parent->right = nptr;
    return nptr->code;
  }
  else
    return result_search;
} /* add_to_tree */

/* =====================  print_tree  ================================
Purpose:          prints a text (sub)tree to the tracefile;
                  for demonstration and debugging purposes
Parameters:
  input:          top_node        : node at the top of the (sub)tree
                  top_barposition : indent
                  textspace       : global text space
Global variables:
Algorithm:
    - print the tag
      - abbreviate a symbol that is too long with a '.' and its last
        character
    - do for all daughter nodes of top_node
      - call print_tree (the function is called recursively)
      - print enough '-'s and '|'s
----------------------------------------------------------------------*/
void print_tree(node_type_p top_node, long  top_barposition, char *textspace)
{
 long i, j, i_end;
 node_type_p current_node;
/* char tag[max_symbol_size];  = tagstring_type tag */

  { /* print the tag */
    if (strlen(top_node->tag) <= maxlen_symbol)
    /*abbreviate a symbol that is too long
     with a '.' and its last character */
    {
      if (top_node->former == NULL)
      {
        i_end = maxlen_symbol - strlen(top_node->tag);
        for (i = 1; i <= i_end; i++)
          fprintf(tracefile, "-");
        fprintf(tracefile, " %s", top_node->tag);
      }
      else
      {
        i_end = maxlen_symbol - strlen(top_node->tag);
        for (i = 1; i <= i_end; i++)
          fprintf(tracefile, " ");
        fprintf(tracefile, "%s", top_node->tag);
      }
    }
    else
    { /* tag too long; abbreviate */
      i_end = maxlen_symbol - 3;
      for (i = 0; i <= i_end; i++)
			{
				fprintf(tracefile, "%c", top_node->tag[i]);
			}
      fprintf(tracefile, ".%c",
      top_node->tag[strlen(top_node->tag)-1]);
    };
    if (top_node->leaf)
    {
      /*write first and last 5 characters of text */
      fprintf(tracefile, "  ");
      if (top_node->u.t.length_of_text < maxlen_symbol)
			{
        for (i = top_node->u.t.start_of_text;
             i <= top_node->u.t.end_of_text;
						 i++)
				{
					fprintf(tracefile, "%c", textspace[i]);
				}
			}
      else
      {
				for (i = 0; i <= 4; i++)
				{
          fprintf(tracefile, "%c",
           textspace[top_node->u.t.start_of_text + i]);
				}
        fprintf(tracefile, "..");
        for (i = 1; i <= 5; i++)
				{
          fprintf(tracefile, "%c",
          textspace[top_node->u.t.end_of_text - 5 + i]);
				}
      };
      fprintf(tracefile, "\n");
    }
    else if (top_node->u.f.down_start != NULL)
    { /* print descendants */
      fprintf(tracefile, " --");
      top_barposition++;
      barposition[top_barposition] = maxlen_symbol + 4;
      current_node = top_node->u.f.down_start;
      if (!top_node->leaf)
			{
        while (current_node != NULL)
        {
          if (current_node->next != NULL)
              barsigns[top_barposition] = '|';
          else
              barsigns[top_barposition] = ' ';
          print_tree(current_node, top_barposition,
             textspace);
          current_node = current_node->next;
          if (current_node != NULL)
          {
						for (j = 1; j <= maxlen_symbol; j++)
              fprintf(tracefile, " ");
            for (i = 1; i <= top_barposition; i++)
            {
              for (j = 1;
                  j <= (barposition[i] - 1);
                  j++)
              fprintf(tracefile, " ");
             fprintf(tracefile, "%c", barsigns[i]);
            };
            fprintf(tracefile, "\n");
            for	(j = 1;
								 j <= maxlen_symbol;
								 j++)
              fprintf(tracefile, " ");
            for (i = 1;
								 i <= (top_barposition - 1);
								 i++)
            {
              for (j = 1;
									 j <= (barposition[i] - 1);
                   j++)
							{
                fprintf(tracefile, " ");
							}
              fprintf(tracefile, "%c", barsigns[i]);
            };
            fprintf(tracefile, "    ");
          }
        }
      }
		}
    else
      fprintf(tracefile, "\n");
  };
} /*of print_tree*/


/* =====================  skip_white_space  =============================
Purpose:    skips white space within a tag and attributes
Parameters:
  input:    s             : pointer to text
  output:   end_of_tag    : true if the textpointer is, at the end,
                            positioned at '/>' (XML) or a '>'
Global variables: bufferp : textpointer
----------------------------------------------------------------------*/
void skip_white_space(char *s, int *end_of_tag)
{
	while ( (s[bufferp] == ' ') || (s[bufferp] == '\t') )
    bufferp++;
   if (s[bufferp] == '/')
	 {
		 *end_of_tag = true;
		 bufferp++;
	 }
	 else if (s[bufferp] == '>')
		*end_of_tag = true;
} /* skip_white_space */

/* =====================  read_name  =============================
Purpose: a name is read from the text;
    if the name starts with an apostroph or quote
    then it may contain spaces; in that case the
      last character should again be an apostroph or quote;
    else the name ends before a space or a '>';
    N.B.: of course the name of an attribute may not contain an
    apostroph; we assume that this condition has been checked
    already by the SGML parser
Parameters:
  input:          s             : pointer to text
  output:         name          : the name that is read
                  code_name     : the coded name
                  end_of_tag    : true if the textpointer is, at the end,
                                  positioned at a '>'
  input/output:
Global variables: bufferp       : textpointer
Algorithm:
      - determine which character will delimit the name at the end
      - increment the textpointer until the delimiting character is found
      - add the name to the name tree (if it is not already there)
----------------------------------------------------------------------*/
void read_name(char *s, char *name, long *code_name, int *end_of_tag, int type_of_name)
  { int length_name;
    char end_char; /* character expected at the end of the name */
    length_name = -1;
    if ((s[bufferp] == '\'') || (s[bufferp] == '\"'))
                                                /* apostrof or quote */
    {
			end_char = s[bufferp];
      bufferp++;
    }
    else end_char = ' ';
    while ( (s[bufferp] != end_char) &&
                          (s[bufferp] != '>') && (s[bufferp] != '=') )
    {
			length_name++;
      name[length_name] = upcase(s[bufferp]);
      bufferp++;
    };
    name[length_name+1] = '\0';
    if ((s[bufferp] == '\'') || (s[bufferp] == '\"'))
      bufferp++;  /* skip the apostrof or quote at the end */
    *end_of_tag = (s[bufferp] == '>');
    *code_name = add_to_tree(name, type_of_name);
  } /* read_name */

/* =====================  compare_long  ==============================
Purpose:  compare 2 longs; the function is called by qsort
Parameters:
  input:          p, q
  output:         return: -1, if p < q
                           0, if p = q
                           1, if p > q
----------------------------------------------------------------------*/
int compare_long(long *p, long *q)
{
	return *p < *q ? -1 : *p > *q ? +1 : 0;
}

/* =====================  compare_short  ============================
Purpose:  compare 2 shorts; the function is called by qsort
Parameters:
  input:          p, q
  output:         return: -1, if p < q
                           0, if p = q
                           1, if p > q
----------------------------------------------------------------------*/
int compare_short(short *p, short *q)
{
	return *p < *q ? -1 : *p > *q ? +1 : 0;
}

/* =====================  add_daughter_node  ========================
Purpose:  reads a tag name and stores it in a new daughter node 'node_p'
         of 'current_node'; reads subsequently the (eventual) attributes.
Parameters:
  input:          textspace_p : pointer to the global text
                  node_is_leaf : true if the node contains text
  output:         tagnew : the read tag name
                            if node_is_leaf then tagnew is the artificial
                            tag "pcdata", "PI" or "comment_decl"; no tagstring has to be read
Global variables:
Algorithm:
      - allocate a new node and initialise it
      - read the tag name and its attribute (name, value) pairs and store
        them in the node
      - sort the attributes on attr_name
      - add to the end a dummy triple with attr_name = LONG_MAX
      - store the buffer position of the end of the tagstring
        and the number of attributes
----------------------------------------------------------------------*/
node_type_p add_daughter_node(char * textspace_p, char * tagnew, boolean node_is_leaf)
{ node_type_p node_p;
  short number_of_attributes;
  attributes_type attributes;
  int end_of_tag, amount_of_information;
  char dummy_name[max_symbol_size];
  int (*compare_p)(); /* becomes pointer to the procedure compare_long;
      necessary for MVS */

  /* add_daughter_node */
  /* current_node is the father_node of the new node_p */
  node_p = (node_type_p)my_calloc(1, sizeof(node_type));
	node_p->node_has_to_be_compared = true; /* by default; may be overruled by indicators in the .emp file */

 /* if not node_is_leaf then tagnew has to be read from the input,
    plus the eventual attributes; together they form the tagstring;
    if node_is_leaf then tagnew is the artificial tag "#pcdata";
    no tagstring has to be read; */

  if (node_is_leaf)
  { /* node is a leaf with the artificial tag "pcdata", "PI" or "comment_decl" */
    node_p->start_of_tagstring = -1;
    node_p->end_of_tagstring = -1;
    node_p->number_of_attributes = 0;
    node_p->attributes = NULL;
    strcpy(node_p->tag, tagnew);
		node_p->tagcode = add_to_tree(tagnew, symbol_is_element);
    node_p->u.t.start_of_text = bufferp;
  }
  else
  { /* node is not a leaf; read the tagstring */

    /* bufferp is positioned at the first character behind "<" */
    node_p->start_of_tagstring = bufferp-1;
    skip_white_space(textspace_p, &end_of_tag);
    read_name(textspace_p, node_p->tag, &node_p->tagcode,
                                &end_of_tag, symbol_is_element);
    strcpy(tagnew, node_p->tag);
    skip_white_space(textspace_p, &end_of_tag);
    number_of_attributes = 0;
    while (!end_of_tag)
    {
			long attr_name_number;
			enum discriminative_type discrimination;
			discriminating_attributes_of_element_type_p help_p, start_of_discr_attrs_p;
			
			number_of_attributes++;
      read_name(textspace_p, dummy_name,
        &attr_name_number,
        &end_of_tag, symbol_is_attribute);
			attributes.attr_info[number_of_attributes-1].attr_name = attr_name_number;
			
			/* is it a discriminating attribute ? */
			discrimination = not_discriminative;
			/* search in the list of discriminating attributes for this element */
			start_of_discr_attrs_p = discriminating_attributes_of_element_p[node_p->tagcode];
			if ( start_of_discr_attrs_p == NULL )
			{	/* no list for this element; maybe there is a list for the element '*' */
				start_of_discr_attrs_p = discriminating_attributes_of_element_p[code_of_arb_element];
			}

			/* search in list of discr. attrs. for this element */
			help_p = start_of_discr_attrs_p;
			while ((help_p != NULL) && (discrimination == not_discriminative))
			{
				if ( help_p->attribute == attr_name_number )
				{
					discrimination = help_p->discrimination;
				}
				help_p = help_p->next;
			}
			
			if (discrimination == not_discriminative)
			{	/* For this attribute no discrimination has been found. Maybe there is an arb. */

				/* Search for an arb in the list of discr. attrs. */
				help_p = start_of_discr_attrs_p;
				while ((help_p != NULL) && (discrimination == not_discriminative))
				{
					if ( help_p->attribute == code_of_arb_attribute )
					{
						discrimination = help_p->discrimination;
					}
					help_p = help_p->next;
				}
			}

			attributes.attr_info[number_of_attributes-1].discrimination = discrimination;

      skip_white_space(textspace_p, &end_of_tag);
      /* bufferp should now be positioned at a '=';
                    if not we will return a status code */
      if (textspace_p[bufferp] != '=')
      {
				char temp[80];
				strncpy(temp, &textspace_p[bufferp - 40], 80);
				sprintf(buffer_sprintf, "Not positioned at equalsign; surrounding text = \n");
				strcat(buffer_sprintf, temp);
        fprintf(logfile, "%s\n", buffer_sprintf);
				result = not_positioned_at_equalsign;
        longjmp(exception_status,
                              not_positioned_at_equalsign);
      };
      bufferp++; /* skip the character '=' */
      skip_white_space(textspace_p, &end_of_tag);
      read_name(textspace_p,
								dummy_name,
								&attributes.attr_info[number_of_attributes-1].attr_value,
                &end_of_tag, symbol_is_attribute_value);
      skip_white_space(textspace_p, &end_of_tag);
      attributes.attr_info[number_of_attributes-1].sequence =
								number_of_attributes;
    };

        /* allocate only the necessary amount of memory */
    amount_of_information = (number_of_attributes+1)*
                                      sizeof(attr_triple_type);
      /* amount_of_information always > 0 */
    node_p->attributes = (attributes_type_p)my_calloc(1,
                                          amount_of_information);
    /* sort the attributes on attr_name */
    compare_p = compare_long;
    qsort((void *)&attributes, number_of_attributes,
                        sizeof(attr_triple_type), compare_p);
												/* ignore compiler warning */
    /* add to the end a dummy triple with attr_name = LONG_MAX
    (the maximum value of a long integer). */
    attributes.attr_info[number_of_attributes].attr_name =
                                                      LONG_MAX;
    /* copy only the necessary amount of information
        from attributes to node_p->attributes */
    memcpy(node_p->attributes, &attributes,
                                        amount_of_information);
    /* store the buffer position of the end of the tagstring
                                and the number of attributes */
    node_p->end_of_tagstring = bufferp;
    node_p->number_of_attributes = number_of_attributes;

    /* in all cases we are located at a ">"; skip that symbol */
    bufferp++;
  };

  /* node_p->compared_with = NULL;  not necessary: allocated with calloc */
  /* put in the tree */
  node_p->up = current_node;
  node_p->former = current_node->u.f.down_end;
  if (node_p->former != NULL)
      node_p->former->next = node_p;
  node_p->next = NULL;
  node_p->empty = false;
  node_p->leaf = node_is_leaf;
  node_p->start_diff_marker = NULL;
	node_p->current_diff_marker = NULL;
  node_p->start_crs_in_original = NULL;
  node_p->last_crs_in_original = NULL;
  /* current_node now acts as the father node */
  if (current_node->u.f.down_start == NULL)
      current_node->u.f.down_start = node_p;
  current_node->u.f.down_end = node_p;
  current_node->u.f.nr_of_nodes_down++;
	node_p->sequence_nr_of_node = current_node->u.f.nr_of_nodes_down;
  current_node = node_p;
	return node_p;
} /*add_daughter_node*/


/* =====================  readfile_from_buffer  =====================
Purpose:          reads a textfile into a tree; keeps also an
   indication of the position of the original carriage returns ('crs')
Parameters:
  input:          textspace_p : pointer to the text buffer
                  eol_proc    : true if the position of the end of
                                lines will have to be remembered
                  long len    : length of the text buffer
Global variables:
Algorithm:
      - the positions of the crs are stored in a list from which the
        first and last element are referenced from the current node
      - if a starttag is encountered
        - complete the current node (if it was a leaf) and position
          at its father as the current node
        - add a new daughter node
        - if the starttag is (in the DTD) an empty one
          - position at the father node
      - if an endtag is encountered
        - complete the current node
        - climb up in the tree until a node is found with the current
          endtag as a starttag; make this node the current one
      - if a normal character is encountered
        - if the current node has no text (up till now)
          - add a new daughter node with the artificial tag "#pcdata"
      - if end of file
        - complete the current node
----------------------------------------------------------------------*/
void readfile_from_buffer(char * textspace_p, long eol_proc,
                            long len, char * A_or_B)
{
  int i, leaf_open;
  long cr_data_index;
  char newtag[max_symbol_size];
  crs_in_original_type_p crs_in_original;
	boolean start_tag, end_tag, start_processing_instruction, 
						start_declaration_or_comment;
	node_type_p daughter_node_p;

  /*readfile_from_buffer*/
  leaf_open = false;
  bufferp = 0;
  cr_data_index = 0;
  while (bufferp < len)
  {
		while (eol_proc && (cr_data[cr_data_index] <= bufferp) )
    {
        /* end of record reached; keep its position */
        crs_in_original = my_malloc(sizeof(crs_in_original_type));
                                            /* create a record */
        if (leaf_open)
          crs_in_original->position = bufferp;
              /* store buffer position */
        else  /* or store the number of the node
                          where the CR occurs */
          crs_in_original->position =
            current_node->u.f.nr_of_nodes_down + 1;
        crs_in_original->next = NULL;
        if (current_node->last_crs_in_original == NULL)
        {
          current_node->last_crs_in_original = crs_in_original;
          current_node->start_crs_in_original = crs_in_original;
        }
        else  /* attach the created record to the tail of
                  the list crs_in_original */
        {
          current_node->last_crs_in_original->next = crs_in_original;
          current_node->last_crs_in_original = crs_in_original;
        };
        cr_data_index++;
      };
			
			if ((strncmp(textspace_p+bufferp, "<", 1) == 0)) 
			{
				start_tag = false;
				end_tag = false;
				start_processing_instruction = false;
				start_declaration_or_comment = false;
				if (strncmp(textspace_p+bufferp+1, "?", 1) == 0)
					start_processing_instruction = true;
				else 
				if (strncmp(textspace_p+bufferp+1, "!", 1) == 0)
					start_declaration_or_comment = true;
				else
				if (strncmp(textspace_p+bufferp+1, "/", 1) == 0)
					end_tag = true;
				else
					start_tag = true;
				bufferp++;

        if (start_tag)
				{
          if (current_node->leaf) /* open #pcdata, pi or comment/decl */
          {
            complete_current_node(textspace_p, current_node, bufferp-2);
            current_node = current_node->up;
            leaf_open = false;
          };
          daughter_node_p = add_daughter_node(textspace_p, newtag, 0);
          if ( (strncmp(textspace_p+bufferp-2, "/", 1) == 0) || /* XML empty tag */
								empty_element_p[current_node->tagcode])
          /* the tag is empty */
          {
						current_node->empty = true;
            current_node = current_node->up;
          }
        }
        if (end_tag)
        { complete_current_node(textspace_p, current_node, bufferp-2);
          bufferp++; /* skip the "/" */
          /* read endtag into newtag */
          sscanf(textspace_p + bufferp, "%[^>]", newtag);
          /* skip the tag and the '>' */
          bufferp = bufferp + strlen(newtag) + 1;

          for (i = 0; i < (int)strlen(newtag); i++)
					{
            newtag[i] = upcase(newtag[i]);
					}

          while (strncmp(newtag, current_node->tag, min(strlen(newtag), strlen(current_node->tag))) != 0 )
					/* Allows that start- and endtags may be prefixes of each other; done in order
						 to process easily doc's where a starttag has been extended with a unique modifier.
						 For complete comparison: exchange the statements above */
					{
						/* for omitted endtags, including #pcdata */
            current_node = current_node->up;
            if ( (current_node == NULL) ||
								 (strcmp(current_node->tag, "top") == 0) )
            {
							start_endtag_failure(newtag, textspace_p, A_or_B);
						}
		        complete_current_node(textspace_p, current_node, bufferp-2); /* moved 24-10-2002 */
          };
/*        complete_current_node(textspace_p, current_node, bufferp-2);  == added 26-7-2002 */
					current_node = current_node->up;
          /* climb up to the father node */
          leaf_open = false;
        };
				if (start_processing_instruction || start_declaration_or_comment)
				{
          if (current_node->leaf)
          {
            complete_current_node(textspace_p, current_node, bufferp-2);
            current_node = current_node->up;
            leaf_open = false;
          };
					bufferp--; /* get back the starting '<' */
          if (start_processing_instruction)
					{
						daughter_node_p = add_daughter_node(textspace_p, "pi", 1);
						if (!compare_pi)
							daughter_node_p->node_has_to_be_compared = false;
					}
					else /* start_declaration_or_comment */
					{
						daughter_node_p = add_daughter_node(textspace_p, "decl_or_comment", 1);
						if (!compare_comment)
							daughter_node_p->node_has_to_be_compared = false;
					}
					leaf_open = true;
					bufferp++; /* but don't fool this scanner */
				};
      }
      else

			if ((strncmp(textspace_p + bufferp, ">", 1) == 0))
			{ /* end of a pi or a declaration or comment ? */
				bufferp++;
        if (current_node->leaf)
        {
          /* complete_current_node expects bufferp to be 1 ahead */
					complete_current_node(textspace_p, current_node, bufferp-1);
          current_node = current_node->up;
					leaf_open = false;
        };
			} 

			else
      { /* no, it is a normal text character */
        if (!leaf_open) /* if this is the first
													 character of the text:
                           add a node with #pcdata */
        {
          daughter_node_p = add_daughter_node(textspace_p, "#pcdata", 1);
					daughter_node_p->pcdata = true;
          leaf_open = true;
        };
        bufferp++;
      }
    }; /* while bufferp < len */

    if (leaf_open)   /* at eof: if there was running text:
                          close all open tags */
    {
			bufferp++; /* complete_current_node expects bufferp to be 1 ahead */
      complete_current_node(textspace_p, current_node, bufferp-2);
      leaf_open = false;
    };

} /*readfile_from_buffer*/

/* =====================  adjust_last_rev_marker  ============================
Purpose:    changes the last open revision marker into the change marker
Parameters:
  input:          kind_of_diff : one of the 2 values:
                      - diff_open_a_has_more
                      - diff_open_b_has_more
Global variables: tag_change: indicates if the last open revision tag is changed into a <rev>
									len_c
									position_of_last_open_revision_tag:  position of the last
																											 revision open tag that has been
																											 added to the output buffer
									position_of_last_close_revision_tag: position of the last
																											 revision close tag that has been
																											 added to the output buffer
									type_of_last_open_revision_tag:	can be
																												diff_open_a_has_more,
																												diff_open_b_has_more,
																												diff_open_change

Algorithm:
  - if the buffer ends on <endX>
      ( in that case the suffix of the buffer would become  <X>...<endX><Y>;
        transform this pattern according to the following rules:
          <X>...<endX><Y>...<endY>  ==>   <X>...<endX>      if X == Y
                                          <rev>...<endrev>  if X != Y
        in order to perform the transformation online the following
            variables are used:
            position_of_last_open_revision_tag (the start of <X>),
            position_of_last_close_revision_tag (the start of <endX>),
            type_of_last_open_revision_tag (X),
            tag_change
      )
    - remove <endX>
    - if X != Y
      - if X != 'rev'
        - replace X by 'rev'
          (assumption: the length of the open revisionmarkers is equal)
      - type_of_last_open_revision_tag = diff_open_change
      - tag_change = true
    - return true
  - else return false (buffer did not end with a revision tag)

----------------------------------------------------------------------*/
int adjust_last_rev_marker(enum kind_of_diff_type kind_of_diff)
  { /* adjust_last_rev_marker */

		if (!combtags) /* no combination of revision tags required */
		{
			return false;
		}

	/*- if the buffer ends on <endX> */
		if (  (position_of_last_close_revision_tag != 0) &&
          (position_of_last_close_revision_tag +
              length_of_close_revision_tag) == len_c)
  /*  - remove <endX>  */
    { len_c = position_of_last_close_revision_tag;
/*  - if X != Y */
/*    - if X != 'rev'  */
      if ( (type_of_last_open_revision_tag != kind_of_diff) &&
           (type_of_last_open_revision_tag != diff_open_change) )
/*    - replace X by '<rev>' (assumption: the length of the open
        revisionmarkers is equal) */
      { memcpy(data_c + position_of_last_open_revision_tag,
                      &tag_open_change, length_of_open_revision_tag);
        type_of_last_open_revision_tag = diff_open_change;
        tag_change = true;
      };
      return true;
    }
    else  /* the buffer does not end on <endX> */
      return false;
  } /* adjust_last_rev_marker */


/* bookmark point */
										
#define init_child_lh_node(A_or_B) \
	if (child_lh_node[A_or_B]->pcdata == true) \
	{ next_atom_is_within_pcdata_node[A_or_B] = true;	 \
		/* go down within pcdata node */ \
		if ( atomic_unit_of_text == rev_chars ) \
		{	position_lh_within_pcdata_node[A_or_B] = child_lh_node[A_or_B]->u.t.start_of_text; \
			last_position_lh_within_pcdata_node[A_or_B] = child_lh_node[A_or_B]->u.t.end_of_text; \
		} \
		else \
		{	position_lh_within_pcdata_node[A_or_B] = 0; \
			last_position_lh_within_pcdata_node[A_or_B] = child_lh_node[A_or_B]->u.t.length_of_hash_coded_text - 1; \
		}; \
	} \
	else \
		next_atom_is_within_pcdata_node[A_or_B] = false;

#define position_at_next_child_lh_node(A_or_B) \
  /* increment position_within_mixed_node[A_or_B] */ \
	if (child_lh_node[A_or_B] != NULL) /* added 1-11-2002 17:40 */ \
		child_lh_node[A_or_B] = child_lh_node[A_or_B]->next; \
	if (child_lh_node[A_or_B] == NULL) \
		current_mixed_node_is_exhausted[A_or_B] = true; \
	else \
	{	init_child_lh_node(A_or_B); \
		next_atom_is_in_new_child[A_or_B] = true; \
	}

#define get_next_atom(A_or_B)  /* result: result_char, result_long or result_node */ \
	result_char[A_or_B] = 0; \
	result_long[A_or_B] = 0; \
	result_node[A_or_B] = NULL; \
	result_pcdatanode[A_or_B] = NULL; \
	if (next_atom_is_within_pcdata_node[A_or_B]) \
	{ /* get next atom from pcdata_node[A_or_B] */ \
\
		if ( atomic_unit_of_text == rev_chars ) \
			result_char[A_or_B] = data[A_or_B][position_lh_within_pcdata_node[A_or_B]]; \
		else \
		{ \
			if (child_lh_node[A_or_B]->u.t.length_of_hash_coded_text == 0) \
				result_long[A_or_B] = -1; /* there is only white space */ \
			else \
				result_long[A_or_B] = child_lh_node[A_or_B]->u.t.hashcode_text[position_lh_within_pcdata_node[A_or_B]]; \
		} \
		result_pcdatanode[A_or_B] = child_lh_node[A_or_B]; \
		next_atom_is_in_new_child[A_or_B] = false; \
		position_lh_within_pcdata_node[A_or_B]++; \
		if (position_lh_within_pcdata_node[A_or_B] > last_position_lh_within_pcdata_node[A_or_B]) \
		{	position_at_next_child_lh_node(A_or_B); \
		} \
	} \
	else \
	{ /* get next atom as non_pcdata_node */ \
		result_node[A_or_B] = child_lh_node[A_or_B]; \
		position_at_next_child_lh_node(A_or_B); \
	}; 

/* example of use for doc B:
		child_lh_node[1] = mixed_node[1]->u.f.down_start;
		init_child_lh_node(1);
		current_mixed_node_is_exhausted[1] = false;
		while (!current_mixed_node_is_exhausted[1])
		{
			get_next_atom(1);
			--> returns one of result_char[1], result_long[1] and result_node[1]
				 as a non-zero value <--
			...
		}
*/

#define startcondition \
if (!current_mixed_node_is_exhausted[0]) \
{ \
	proceed = true; \
	if ( (start_chars_or_node_a == chars) && (result_node[0] == NULL) ) \
	{	\
		if ( (position_lh_within_pcdata_node[0] /* -1: former one 5-9-2002*/ == start_position_char_a ) \
				&& (result_pcdatanode[0] == start_position_node_a) ) \
			proceed = false; \
	} \
	else  \
	if ( (start_chars_or_node_a == node) && (result_node[0] != NULL) ) \
	{ if (result_node[0] == position_node_a ) \
			proceed = false; \
	} \
} \
else \
proceed = false;


#define endcondition \
if (!current_mixed_node_is_exhausted[0]) \
{ \
	proceed = true; \
	if ( (end_chars_or_node_a == chars) && (result_node[0] == NULL) ) \
	{	boolean t; \
		t = ( (position_lh_within_pcdata_node[0] /*-1 4-9-2002*/) > end_position_char_a ) \
				&& (result_pcdatanode[0] == end_position_node_a); \
		if (t == true) /* unbelievable: with "if (t)" then proceed becomes false ! */ \
			proceed = false; \
	} \
	else  \
	if ( (end_chars_or_node_a == node) && (result_node[0] != NULL) ) \
	{ if (result_node[0] == end_position_node_a) \
			proceed = false; \
	} \
} \
else \
proceed = false;

/* =====================  write_diff_tag  ============================
Purpose:          writes a marker tag for a difference
Parameters:
  input:          current_diff_marker : one of the 4 values:
                      - diff_open_a_has_more
                      - diff_close_a_has_more
                      - diff_open_b_has_more
                      - diff_close_b_has_more
									chars_or_node_a : one of the 2 values:
											- chars
											- node
Global variables:     none
Algorithm:
  - dependent on one of the 4 values a marker tag is written in the
    output buffer

  - let us call Y = '<add>' or '<del>'; call the corresponding <endY>

  - upon adding <Y> to the output buffer:
    - adjust_last_rev_marker

  - upon adding <endY> to the output buffer:
    - if tag_change
      - add '<endrev>' to the buffer, instead of <endY>

  - upon adding <del>
		- if print_a (the deleted text has to be printed)
			- if atom = character
				- write the text for A from (*current_diff_marker_p)->position_a
					to ((*current_diff_marker_p)->next)->position_a
				else
				- position in the hash-coded text of position_node_a
					- by first skipping lexemes (with getsym) until the start position is reached
						- not counting the lexemes which are lexeme_not_to_be_coded
					- and then printing the selected lexemes
						- also not counting the lexemes which are lexeme_not_to_be_coded
						- and keeping at the start and end only lexemes wich are lexeme_to_be_coded
----------------------------------------------------------------------*/
void write_diff_tag(diff_marker_type_p current_diff_marker_p)
{ 
  	long start_hash_a, end_hash_a, start_position_char_a, end_position_char_a;
		node_type_p node_a, position_node_a, node_a_last, start_position_node_a, end_position_node_a;
		enum chars_or_node_type chars_or_node_a;
		boolean current_mixed_node_is_exhausted[2],
						next_atom_is_within_pcdata_node[2],
						next_atom_is_in_new_child[2],
						mixed_node_a,
						proceed;
		node_type_p mixed_node[2], 
								child_lh_node[2],
								result_node[2],
								result_pcdatanode[2];
		char result_char[2],
				 *data[2];
		long position_lh_within_pcdata_node[2],
				 last_position_lh_within_pcdata_node[2],
				 result_long[2],
				 charpos_in_pcdata[2],
				 end_charpos_in_pcdata[2];
		my_int A_or_B = 0; /* only for debugging */

/* bookmark point */

		chars_or_node_a = current_diff_marker_p->chars_or_node_A;
		node_a = current_diff_marker_p->position_node_a;
		
		switch (current_diff_marker_p->kind_of_diff)
    { case diff_open_a_has_more:
			 
			 diff_tag_open_a = false;
			 diff_tag_open_b = false;

			 if (!adjust_last_rev_marker(diff_open_a_has_more))
        {
					position_of_last_open_revision_tag = len_c;
          type_of_last_open_revision_tag = diff_open_a_has_more;
          
					if ((chars_or_node_a != chars) &&
								(node_a->node_has_to_be_compared == false))
						/* if node_a does not have to be compared then no tag_open_for_a_has_more
							 has to be written */
					{}
					else
					{
						add_to_buffer_c(tag_open_for_a_has_more, strlen(tag_open_for_a_has_more));
						diff_tag_open_a = true;
					}
          tag_change = false;
        };
				if (print_a)
				{
					long i, j;
					long position_char_a, position_char_a_end;
					int type_of_lexeme, end_of_text;
					char lexeme[max_length_of_lexeme],
							 lexeme2[max_length_of_lexeme]	;
					enum chars_or_node_type start_chars_or_node_a, end_chars_or_node_a;
					/*
					write the deleted text for A:
					- if diff_marker = <del>
						- if atom = character
							- write the text for A from (*current_diff_marker_p)->position_char_a
								to ((*current_diff_marker_p)->next)->position_char_a
							else
							- position in the hash-coded text of position_node_a
								- by first skipping lexemes (with getsym) until the start position is reached
									- not counting the lexemes which are lexeme_not_to_be_coded
								- and then printing the selected lexemes
									- also not counting the lexemes which are lexeme_not_to_be_coded
								 	- and keeping at the start and end only lexemes wich are lexeme_to_be_coded
					10-8-2002: Update for mixed mode: use the algorithm for next atoms at printing of a node_B
					
					*/

					position_node_a = current_diff_marker_p->position_node_a;
					position_char_a = current_diff_marker_p->position_char_a;
					mixed_node_a = false;
					if ( (position_node_a != NULL) && (position_node_a->up != NULL) )
					{
						mixed_node[0] = position_node_a->up;
						mixed_node_a = mixed_node[0]->mixed_mode;
					}
										
/* bookmark point */
										
					if (mixed_node_a)
					{ /* special treatment for mixed nodes, 11-8-2002 */
						
						data[0] = data_a;
						start_position_char_a = current_diff_marker_p->position_char_a;
							/* position_char_a is the start position of the (eventually hash coded) text in position_node_a,
									counted from position_node_a->u.t.start_of_text as hash-coded elements */
						if (atomic_unit_of_text != rev_chars)
							start_position_char_a++; /* is a bug-fix; but should be repaired in an earlier phase; to be done 19-8-2002 */
						start_hash_a = 1;
						start_position_node_a = current_diff_marker_p->position_node_a;
						start_chars_or_node_a = current_diff_marker_p->chars_or_node_A;

						end_position_char_a = (current_diff_marker_p->next)->position_char_a;
							/* position_char_a_end is the end position of the hash coded text in position_node_a,
									counted from position_node_a->u.t.start_of_text as hash-coded elements */
						end_position_node_a = (current_diff_marker_p->next)->position_node_a;
						end_chars_or_node_a = (current_diff_marker_p->next)->chars_or_node_A;
						
						child_lh_node[0] = mixed_node[0]->u.f.down_start;
						init_child_lh_node(0);
						current_mixed_node_is_exhausted[0] = false;
						end_of_text = false;
						lexeme[0] = '\0';


						/* skip leading atoms; implies: skip leading hashes;
							principle: get first the symbol/lexeme of the lh-node and then the atom, because
							  the pointers and indexes needed for getsym are known for the lh-node */

						do 
						{								
							/* if child_lh_node[0] is pcdata, get a symbol/lexeme */
							if (child_lh_node[0]->pcdata)
							{ if (atomic_unit_of_text == rev_chars)
								{
									lexeme[0] = data_a[position_lh_within_pcdata_node[0]];
									lexeme[1] = '\0';
								}
								else
								{ /* atomic_unit_of_text != rev_chars */
									type_of_lexeme = lexeme_to_be_coded;
									end_of_text = false;

									/* get the lexeme of the next atom; we do this only to keep track of 
											the right count_char_position_in_pcdata_a */
									if (next_atom_is_in_new_child[0])
									{
										charpos_in_pcdata[0] = child_lh_node[0]->u.t.start_of_text; 
										end_charpos_in_pcdata[0] = child_lh_node[0]->u.t.end_of_text; 
									}
									getsym(data_a, &charpos_in_pcdata[0], end_charpos_in_pcdata[0],
													lexeme, &type_of_lexeme, &end_of_text);
									if ( (type_of_lexeme == lexeme_not_to_be_coded) && (!end_of_text) )
									{	getsym(data_a, &charpos_in_pcdata[0], end_charpos_in_pcdata[0],
														lexeme2, &type_of_lexeme, &end_of_text);
										strcat(lexeme, lexeme2);
									}
								}
							}
							get_next_atom(0);
							startcondition; 
						} while (proceed);

						/* write the selected part of the mixed node */
						do 
						{	/* write the current atom */
							if (result_node[0] == NULL)
								/* it is a pcdata node; write out the current char/lexeme */
								add_to_buffer_c(lexeme, strlen(lexeme));
							else
								/* it is a node */
								write_for_node_a(result_node[0]);

							/* if child_lh_node[0] is pcdata, get a symbol/lexeme */
							if ( (child_lh_node[0] != NULL) && (child_lh_node[0]->pcdata) )
							{ if (atomic_unit_of_text == rev_chars)
								{
									lexeme[0] = data_a[position_lh_within_pcdata_node[0]];
									lexeme[1] = '\0';
								}
								else
								{ /* atomic_unit_of_text != rev_chars */
									type_of_lexeme = lexeme_to_be_coded;
									end_of_text = false;

									/* get the lexeme of the next atom; we do this only to keep track of 
											the right count_char_position_in_pcdata_a */
									if (next_atom_is_in_new_child[0])
									{
										charpos_in_pcdata[0] = child_lh_node[0]->u.t.start_of_text; 
										end_charpos_in_pcdata[0] = child_lh_node[0]->u.t.end_of_text; 
									}
									getsym(data_a, &charpos_in_pcdata[0], end_charpos_in_pcdata[0],
													lexeme, &type_of_lexeme, &end_of_text);
									if ( (type_of_lexeme == lexeme_not_to_be_coded) && (!end_of_text) )
									{	getsym(data_a, &charpos_in_pcdata[0], end_charpos_in_pcdata[0],
														lexeme2, &type_of_lexeme, &end_of_text);
										strcat(lexeme, lexeme2);
									}
								}
							}
							else
								lexeme[0] = '\0';

							get_next_atom(0);
							endcondition;

						} while (proceed);

						if (current_mixed_node_is_exhausted[0]) /* write out remaining atom */
						{
							if (result_node[0] == NULL)
								/* it is a pcdata node; write out the current char/lexeme */
								add_to_buffer_c(lexeme, strlen(lexeme));
							else
								/* it is a node */
								write_for_node_a(result_node[0]);
						}	
					}


					else /* not mixed node */

					if (chars_or_node_a == chars)
					{	/* text in a leaf node */
					
						if (atomic_unit_of_text == rev_chars)
						{
							for (	j = current_diff_marker_p->position_char_a;
										j < (current_diff_marker_p->next)->position_char_a;
										j++)
							{
								add_to_buffer_c(data_a + j + 1, 1);
							}
						}
						else  /*  atomic_unit_of_text != rev_chars */
						{
							start_hash_a = current_diff_marker_p->position_char_a;
								/* position_char_a is the start position of the hash coded text in position_node_a,
										counted from position_node_a->u.t.start_of_text as hash-coded elements */
							end_hash_a = (current_diff_marker_p->next)->position_char_a;
								/* position_char_a_end is the end position of the hash coded text in position_node_a,
										counted from position_node_a->u.t.start_of_text as hash-coded elements */
							type_of_lexeme = lexeme_to_be_coded;
							end_of_text = false;

							/* within position_node_a: skip hash-coded text up to position_char_a */
							position_char_a = position_node_a->u.t.start_of_text;
							position_char_a_end = position_node_a->u.t.end_of_text;

							i = 0;
							/* while (i <= start_hash_a) == 7-11-2002 */
							while (i < start_hash_a) /* == 5-11-2002 */
							{
								/* get the lexeme which was coded (and eventually
										the not to be coded lexemes) */
								getsym(data_a, &position_char_a, position_char_a_end,
												lexeme, &type_of_lexeme, &end_of_text);
								if (type_of_lexeme == lexeme_to_be_coded)
									i++;
							}

							i = start_hash_a;
							while (i < end_hash_a)
							{
								/* get the lexeme which was coded (and eventually
										the not to be coded lexemes) */
								getsym(data_a, &position_char_a, position_char_a_end,
												lexeme, &type_of_lexeme, &end_of_text);
								if (type_of_lexeme == lexeme_to_be_coded)
									i++;
								if ( (type_of_lexeme == lexeme_not_to_be_coded)
											&& (!end_of_text) )
								{
#ifdef tight_revisionmarkers
									/* If atomic_unit_of_text == rev_words we want
											the revision markers as close to a word as
											possible. The variable diff_tag_open
											monitors if a revision tag is open.
											If (i == start_hash_a) then we skip this
											lexeme_not_to_be_coded */
					 				if (i != start_hash_a)
#endif
										add_to_buffer_c(lexeme, strlen(lexeme));
									if (!end_of_text)
									{
										getsym(data_a, &position_char_a, position_char_a_end,
														lexeme, &type_of_lexeme, &end_of_text);
										if (type_of_lexeme == lexeme_to_be_coded)
											i++;
										add_to_buffer_c(lexeme, strlen(lexeme));
									}
								}
								else  /* type_of_lexeme == lexeme_to_be_coded */
								{
									add_to_buffer_c(lexeme, strlen(lexeme));
								}
							} /* while i */
						}
					diff_tag_open_a = true;
				}
				else
				{	/* chars_or_node_a == node */
					 boolean finished;
					 node_a = current_diff_marker_p->position_node_a;
					 /* current_diff_marker_p->next contains the <enddel> marker, and
							current_diff_marker_p->next)->position_char_a is the last node of A which
							was deleted
					 */
					 
					 if ( (current_diff_marker_p->next != NULL)  /* 5-11-2002 */
								&& (current_diff_marker_p->next->chars_or_node_B == node) )
						 node_a_last = (current_diff_marker_p->next)->position_node_a;
					 else
						 node_a_last = node_a->next; 
						
					 if (node_a->node_has_to_be_compared == true) /* write at least once */
						 write_for_node_a(node_a);
					 else
	  				 diff_tag_open_a = false;

					 node_a = node_a->next;
//					 finished = false;
//					 while ((node_a != NULL) && !finished)   == 6-11-2002
  				 while ((node_a != NULL) && (node_a != node_a_last))   /* == 6-11-2002 */
					 {
						 if (node_a->node_has_to_be_compared == true)
						 /* a node_a which does not have to be compared has to be left out */
						 {
							 if (!diff_tag_open_a)
							 {
								 add_to_buffer_c(tag_open_for_a_has_more, strlen(tag_open_for_a_has_more));
								 diff_tag_open_a = true;
							 }
							 write_for_node_a(node_a);
						 }
//						 finished = (node_a == node_a_last);  == 6-11-2002
						 node_a = node_a->next;
					 }
					}
				}
      break;

      case diff_close_a_has_more:
       position_of_last_close_revision_tag = len_c;
        if (tag_change == true)
				{
          add_to_buffer_c(tag_close_change, strlen(tag_close_change));
        }
				else
				{
          if (diff_tag_open_a == true)
						add_to_buffer_c(tag_close_for_a_has_more,
													strlen(tag_close_for_a_has_more));
        }
				diff_tag_open_a = false;
      break;

      case diff_open_b_has_more:
       if (!adjust_last_rev_marker(diff_open_b_has_more))
        {
					position_of_last_open_revision_tag = len_c;
          type_of_last_open_revision_tag = diff_open_b_has_more;
          add_to_buffer_c(tag_open_for_b_has_more,
													strlen(tag_open_for_b_has_more));
          tag_change = false;
        };
        diff_tag_open_b = true;
      break;

    case diff_close_b_has_more:
     position_of_last_close_revision_tag = len_c;
      if (tag_change == true)
      {
			  if (diff_tag_open_b == true)
				/* could be false in case the printing of a 
						tag_open_for_b_has_more has been inhibited because of a node
						with node->node_has_to_be_compared = false */
					add_to_buffer_c(tag_close_change,
												strlen(tag_close_change));
      }
			else
      {
			  if (diff_tag_open_b == true)
				/* could be false in case the printing of a 
						tag_open_for_b_has_more has been inhibited because of a node
						with node->node_has_to_be_compared = false */
					add_to_buffer_c(tag_close_for_b_has_more,
												strlen(tag_close_for_b_has_more));
      }
			diff_tag_open_b = false;
    break;
  }; /* switch */
} /* write_diff_tag */

/* =====================  write_crs  ============================
Purpose:          generates in the output buffer a number of carriage
                  returns
Parameters:
  input:          position : current position in the text of the
                             current node
  input/output:   current_crs_in_original : current element in the list
                                            of crs in the current node
                  current_cr_position     : the position which is stored
                                            in current_crs_in_original
Algorithm:        position is compared with current_crs_in_original,
                  iteratively in subsequent elements of the list of crs;
                  as long as position is larger an eol character is
                  written to the output buffer
----------------------------------------------------------------------*/
void write_crs(long position,
               crs_in_original_type_p *current_crs_in_original,
               long *current_cr_position)
{ while ( (position >= (*current_cr_position)  ) &&
            (*current_cr_position != -2) )
      /* this is a 'while' and not an 'if'
      because there may appear subsequent CR's */
  {
		add_to_buffer_c(&eol_char, 1);
    *current_crs_in_original = (*current_crs_in_original)->next;
    if (*current_crs_in_original != NULL )
		{
      *current_cr_position =
        (*current_crs_in_original)->position;
		}
    else
		{
      *current_cr_position = - 2; /* no more cr's */
		}
  };
} /* write_crs */

/* =====================  write_diff_tags_on_node_level  ============================
Purpose:          generates in buffer_c a number of difference
                  tags
Parameters:
  input:          node_b : current node
  input/output:   current_diff_marker_p : current element in the list of
                                        diff markers in the current node
                  current_diff_position : the position which is stored
                                          in current_diff_marker_p
Algorithm:        node_b is compared with
									(*current_diff_marker_p)->position_b, iteratively in
									subsequent elements of the list of diff markers;
                  as long as they are equal
									a difference tag is written to buffer_c
----------------------------------------------------------------------*/
void write_diff_tags_on_node_level(node_type_p node_b,
										 diff_marker_type_p *current_diff_marker_p,
										 enum chars_or_node_type chars_or_node_a)
{
	if (node_b->node_has_to_be_compared == false)
	{	
		while ( (*current_diff_marker_p != NULL) &&
				( (*current_diff_marker_p)->position_node_b == node_b)
				)
				/* this is a 'while' and not an 'if'
				   because there may appear subsequent diff's */
		{
			if ((*current_diff_marker_p)->kind_of_diff != diff_open_b_has_more)
				/* if node_b does not have to be compared then no diff_open_b_has_more has
			 to be written; but all other revision markers have to be written */
				write_diff_tag(*current_diff_marker_p);
			*current_diff_marker_p = (*current_diff_marker_p)->next;
		};
	}
	else
/* bookmark point */
	{
		while ( (*current_diff_marker_p != NULL) 
						/* && ((*current_diff_marker_p)->chars_or_node_B == node) == 1-11-2002 */
						&& ( (*current_diff_marker_p)->position_node_b == node_b )
					)
		{
			write_diff_tag(*current_diff_marker_p);
			*current_diff_marker_p = (*current_diff_marker_p)->next;
			if ( (node_b->up != NULL) && (node_b->up->mixed_mode) )
			{
				node_b->up->current_diff_marker = *current_diff_marker_p;
			} 
		};

	}

} /* write_diff_tags_on_node_level */


/* =====================  write_diff_tags_on_text_level  ============================
Purpose:          generates in buffer_c a number of difference
                  tags
Parameters:
  input:          position_b : current position in the text of the
															 current node
  input/output:   current_diff_marker_p : current element in the list of
                                        diff markers in the current node
Algorithm:        position_b is compared with
									*current_diff_marker_p)->position_b, iteratively in
									subsequent elements of the list of diff markers;
                  as long as they are equal
									a difference tag is written to buffer_c
----------------------------------------------------------------------*/
void write_diff_tags_on_text_level(long position_b,
										 node_type_p current_node,
										 diff_marker_type_p *current_diff_marker_p)
{
	if (current_node->up->mixed_mode)
	{ 
		while ( (*current_diff_marker_p != NULL) 
						&& ((*current_diff_marker_p)->chars_or_node_B == chars)
						&& (current_node == (*current_diff_marker_p)->position_node_b) 
						&& ( (*current_diff_marker_p)->position_char_b == position_b) ) 
		{
			write_diff_tag(*current_diff_marker_p);
			*current_diff_marker_p = (*current_diff_marker_p)->next;
			current_node->up->current_diff_marker = *current_diff_marker_p;
		}
	}
	else
	while ( (*current_diff_marker_p != NULL) 
					&& ((*current_diff_marker_p)->position_char_b) == position_b )
			/* this is a 'while' and not an 'if'
			because there may appear subsequent diff's */
	{
		write_diff_tag(*current_diff_marker_p);
		*current_diff_marker_p = (*current_diff_marker_p)->next;
		current_node->current_diff_marker = *current_diff_marker_p; /* 8-11-2002 */
	};
} /* write_diff_tags_on_text_level */


void write_for_node_a(node_type_p father_node)
{ /* write_for_node */
  node_type_p current_node;
  long i;
  long position_a, position_a_end;
  int type_of_lexeme, end_of_text;
  char lexeme[max_length_of_lexeme];

  if ( (!father_node->leaf) &&
      (strcmp(father_node->tag, "top") != 0) )
  {   /* write tagstring */
    add_to_buffer_c(data_a + father_node->start_of_tagstring,
										father_node->end_of_tagstring -
												father_node->start_of_tagstring + 1);
  };
  if (father_node->leaf)
          /* text symbol */
  {
	  if (atomic_unit_of_text == rev_chars)
		{
			for (	i = father_node->u.t.start_of_text;
						i <= father_node->u.t.end_of_text;
						i++)
			{
				add_to_buffer_c(data_a + i, 1);
			}; /* for i */
		}
		else  /*  atomic_unit_of_text != rev_chars */
		{
			position_a = father_node->u.t.start_of_text;
			position_a_end = father_node->u.t.end_of_text;
			type_of_lexeme = lexeme_to_be_coded;
			end_of_text = false;
			for (	i = 0;
						i < father_node->u.t.length_of_hash_coded_text;
						i++)
			{
				/* get the lexeme which was coded (and eventually
						the not to be coded lexemes) */
				getsym(data_a, &position_a, position_a_end,
								lexeme, &type_of_lexeme, &end_of_text);
				if ( (type_of_lexeme == lexeme_not_to_be_coded)
							&& (!end_of_text) )
				{ /* If atomic_unit_of_text == rev_words we want
							the revision markers as close to a word as
							possible. The variable diff_tag_open
							monitors if a revision tag is open */
					add_to_buffer_c(lexeme, strlen(lexeme));
					if (!end_of_text)
					{
						getsym(data_a, &position_a, position_a_end,
										lexeme, &type_of_lexeme, &end_of_text);
						add_to_buffer_c(lexeme, strlen(lexeme));
					}
				}
				else  /* type_of_lexeme == lexeme_to_be_coded */
				{
					add_to_buffer_c(lexeme, strlen(lexeme));
				};
			}; /* for i */

			/* Maybe (an) uncoded lexeme(s) was (were) waiting.
				However, in the current schema of coding there can be
				only 0 or 1 uncoded lexemes left */
			while (!end_of_text)
			{
				getsym(data_a, &position_a, position_a_end,
											lexeme, &type_of_lexeme, &end_of_text);
				add_to_buffer_c(lexeme, strlen(lexeme));
			};
		}
  }
  else 	/* father_node is not a leaf */
  {
    if (father_node->u.f.down_start != NULL )
      {
        current_node = father_node->u.f.down_start;
        while (current_node != NULL)
        {
          write_for_node_a(current_node);
          current_node = current_node->next;
        };
      };
  };
  /* fprintf endtag  */
  if ( (!father_node->leaf) &&
       (strcmp(father_node->tag, "top") != 0) &&
       !father_node->empty )   /* the tag is not empty */
  {
		add_to_buffer_c("</", 2);
    add_to_buffer_c(father_node->tag, strlen(father_node->tag));
    add_to_buffer_c(">", 1);
  }
} /* write_for_node_a */


/* =====================  write_text_for_node  =======================
Purpose:  writes out the text of a node, with revision indicators;
          takes into account the type of atomic_unit_of_text and
          the original position of the cr's
Parameters:
  input:          current_node : the current node
Algorithm:
    - if atomic_unit_of_text = rev_chars
      - do for all characters in the text of the node
        - write_crs
        - write_diff_tags
        - write character to the output buffer
    - else (the text is compressed)
      - do for all compressed text elements of the node
        a. write_crs
        b. get, with getsym, the corresponding uncompressed text element
        c. write the uncompressed text element to the output buffer
        d. write_diff_tags
        NB: because diff marker tags have to surround uncompressed
            words as tight as possible the sequence of the calls a, c
            and d differs when atomic_unit_of_text = rev_words
    - process eventual remaining crs and diffs
----------------------------------------------------------------------*/
void write_text_for_node(node_type_p current_node)
{ /* write_text_for_node */
  long i, current_cr_position,
      position_b, position_b_end;
  int type_of_lexeme, end_of_text;
  crs_in_original_type_p current_crs_in_original;
  diff_marker_type_p current_diff_marker_p;
  char lexeme[max_length_of_lexeme];

	current_crs_in_original = current_node->start_crs_in_original;
	if (current_crs_in_original != NULL )
	{
		current_cr_position = current_crs_in_original->position;
	}
	else
	{
		current_cr_position = - 2; /* no more cr's */
		/* the value -2 is motivated because of the
		 automatic failure of
		 the next while statements */
	}

	if ( (current_node->up != NULL) && current_node->up->mixed_mode)
		current_diff_marker_p = current_node->up->current_diff_marker;
	else
		current_diff_marker_p = current_node->start_diff_marker;

	if (atomic_unit_of_text == rev_chars)
	{
		for (	i = current_node->u.t.start_of_text;
					i <= current_node->u.t.end_of_text;
					i++)
		{
			write_crs(i, &current_crs_in_original,
										&current_cr_position);
			write_diff_tags_on_text_level(i - 1, current_node, &current_diff_marker_p);
			add_to_buffer_c(data_b + i, 1);
		}; /* for i */

		/* maybe a diff marker was waiting at the end of the node */

#ifdef proefje //1-11-2002
		if ( (current_node->up != NULL) && (current_node->up->mixed_mode) )
				/* if in mixed_mode, get the updated current_diff_marker_p */
		{
			current_diff_marker_p = current_node->up->current_diff_marker;
			while ( (current_diff_marker_p != NULL) 
					 && ( ( (current_diff_marker_p->chars_or_node_B == node) 
							 /* && ((current_diff_marker_p->position_node_b) == node_b ) */
								)
						 || ( (current_diff_marker_p->chars_or_node_B == chars) 
							 && (current_node == current_diff_marker_p->position_node_b) 
							 /* && ( current_diff_marker_p->position_b == position_b) */
								)
							) 
						)   /* nog even bijgewerkt ============= 4-9-2002, ochtend */
			{
				write_diff_tag(current_diff_marker_p);
				current_diff_marker_p = (current_diff_marker_p)->next;
				current_node->up->current_diff_marker = current_diff_marker_p;
			}
		}
		else
#endif

		while (current_node->current_diff_marker != NULL)
		{
			write_diff_tag(current_node->current_diff_marker);
			current_node->current_diff_marker = (current_node)->current_diff_marker->next;
		};


		/* maybe more cr's were waiting at the end of the node */
		while (current_crs_in_original != NULL)
		{
			add_to_buffer_c(&eol_char, 1);
			current_crs_in_original =
														current_crs_in_original->next;
		};
	}
	else  /*  atomic_unit_of_text != rev_chars */
	{
		long start_position_lexeme_in_b;

		position_b = current_node->u.t.start_of_text;
		position_b_end = current_node->u.t.end_of_text;
		type_of_lexeme = lexeme_to_be_coded;
		end_of_text = false;
		for (	i = 0;
					i < current_node->u.t.length_of_hash_coded_text;
					i++)
		{
			/* get the lexeme which was coded (and eventually
					the not to be coded lexemes) */
			start_position_lexeme_in_b = position_b;
			getsym(data_b, &position_b, position_b_end,
							lexeme, &type_of_lexeme, &end_of_text);
			if ( (type_of_lexeme == lexeme_not_to_be_coded)
						&& (!end_of_text) )
			{
#ifdef tight_revisionmarkers
				 /* If atomic_unit_of_text == rev_words we want
						the revision markers as close to a word as
						possible. The variable diff_tag_open
						monitors if a revision tag is open */

				if (!(diff_tag_open_b && (atomic_unit_of_text == rev_words)))
				{
					add_to_buffer_c_with_crs(lexeme,
																	 strlen(lexeme),
																	 start_position_lexeme_in_b,
																	 &current_crs_in_original,
																	 &current_cr_position);
					write_diff_tags_on_text_level(i, &current_diff_marker_p, chars);

				}
				else
#endif
				{
					if (atomic_unit_of_text == rev_words) /* gek */
						write_diff_tags_on_text_level(i /* 22-10-2002 */, current_node, &current_diff_marker_p);
					else
						write_diff_tags_on_text_level(i- 1 /* 4-9-2002 */, current_node, &current_diff_marker_p);
					add_to_buffer_c_with_crs(lexeme,
																	 strlen(lexeme),
																	 start_position_lexeme_in_b,
																	 &current_crs_in_original,
																	 &current_cr_position);
				};
				if (!end_of_text)
				{
					start_position_lexeme_in_b = position_b;
					getsym(data_b, &position_b, position_b_end,
									lexeme, &type_of_lexeme, &end_of_text);
					add_to_buffer_c_with_crs(lexeme,
																	 strlen(lexeme),
																	 start_position_lexeme_in_b,
																	 &current_crs_in_original,
																	 &current_cr_position);
				}
			}
			else  /* type_of_lexeme == lexeme_to_be_coded */
			{
				write_diff_tags_on_text_level(i - 1 /* 4-9-2002; 
						gek, als A een node extra heeft moet de -1 weg! Waarom? */, current_node, &current_diff_marker_p);
				add_to_buffer_c_with_crs(lexeme,
																 strlen(lexeme),
																 start_position_lexeme_in_b,
																 &current_crs_in_original,
																 &current_cr_position);
			};
		}; /* for i */

		/* maybe a current_diff_marker was waiting at the end of the node */

#ifdef proefje //31-10-2002
		if ( (current_node->up != NULL) && (current_node->up->mixed_mode) )
			/* if in mixed_mode, get the updated current_diff_marker_p */
		{
			current_diff_marker_p = current_node->up->current_diff_marker;
			while ( (current_diff_marker_p != NULL) 
					 && ( ( (current_diff_marker_p->chars_or_node_B == node) 
							 /* && ((current_diff_marker_p->position_b) == node_b ) */
								)
						 || ( (current_diff_marker_p->chars_or_node_B == chars) 
							 && (current_node == current_diff_marker_p->position_node_b) 
							 /* && ( current_diff_marker_p->position_b == position_b) */
								)
							) 
						)   /* nog even bijgewerkt ============= 4-9-2002, ochtend */
			{
				write_diff_tag(current_diff_marker_p);
				current_diff_marker_p = current_diff_marker_p->next;
				current_node->up->current_diff_marker = current_diff_marker_p;
			}
		}
		else
#endif

		while (current_node->current_diff_marker != NULL)
		{
			write_diff_tag(current_node->current_diff_marker);
			current_node->current_diff_marker = (current_node->current_diff_marker)->next;
		}

		/* maybe (an) uncoded lexeme(s) was (were) waiting.
			However, in the current schema of coding there can be
			only 0 or 1 uncoded lexemes left */
		while (!end_of_text)
		{
			start_position_lexeme_in_b = position_b;
			getsym(data_b, &position_b, position_b_end,
										lexeme, &type_of_lexeme, &end_of_text);
			add_to_buffer_c_with_crs(lexeme,
															 strlen(lexeme),
															 start_position_lexeme_in_b,
															 &current_crs_in_original,
															 &current_cr_position);
		};

		/* maybe more cr's were waiting at the end of the node */
		while (current_crs_in_original != NULL)
		{
			add_to_buffer_c(&eol_char, 1);
			current_crs_in_original =
														current_crs_in_original->next;
		};
	};
	
	if ( (current_node->up != NULL) && current_node->up->mixed_mode)
		current_node->up->current_diff_marker	= current_diff_marker_p;

} /* write_text_for_node */


/* =====================  write_for_node  ============================
Purpose:  writes out the tag, attributes and text for a node.
Parameters:
  input:          father_node : the top node for the (sub)tree
                                which has to be written out
Global variables:
Algorithm:
  - if there is no revision tag open
    - print the revision of attributes (if any)
  - print starttag for the node (called father_node)
  - print the text for the node (with write_text_for_node)
  - do for all the daughter nodes of the node
    - process crs
    - process diff markers
    - call write_for_node (recursively)
  - print remaining crs and diff markers
  - print endtag for the node
----------------------------------------------------------------------*/
void write_for_node(node_type_p father_node)
{ /* write_for_node */
  node_type_p child_node;
  long i, current_cr_position, temp_bufferpos;
  short seq_nr, proceed;
  int j;
  crs_in_original_type_p current_crs_in_original;
  diff_marker_type_p current_diff_marker_p;
  short nr_diff_of_attrs;
  diff_of_attrs_type *diff_of_attrs;
  btree help_name;
  attributes_type_p attributes;
	compared_with_buffer_type_p compared_with_buffer_p;
	my_int bufslot;
	
	current_crs_in_original = father_node->start_crs_in_original;
  if (current_crs_in_original != NULL )
  {
		current_cr_position = current_crs_in_original->position;
  }
	else
  {
		current_cr_position = - 2; /* no more cr's */
  }
	/* if (father_node->mixed_mode) == 1-11-2002 */
	/* init the current_diff_marker */
	father_node->current_diff_marker = father_node->start_diff_marker;

	if ( father_node->pcdata &&  /* 7-11-2002 */
			 (father_node->up != NULL) && 
			 (father_node->up->mixed_mode) )
		current_diff_marker_p = father_node->up->current_diff_marker;
	else
		current_diff_marker_p = father_node->start_diff_marker;

  /* write, eventually, the difference of attributes;
      diff_of_attrs contains the sorted sequence numbers
      of the attributes which differ; we will locate in
      father_node->attributes->attr_info these
      sequencenumbers and will print the corresponding names
  */
  /* print revision of attributes, but only if there is no
      difference tag open */

  if (!position_is_zero(father_node->compared_with) && (!diff_tag_open_b) )

  {
		struct compared_with_type *father_node_compared_with_p;

		compared_with_buffer_p = 
						(compared_with_buffer_type_p)filepos_to_ptr(&(father_node->compared_with), &bufslot);

		father_node_compared_with_p = (compared_with_type_p)compared_with_buffer_p;

		nr_diff_of_attrs =
            father_node_compared_with_p->nr_diff_of_attrs;

    if (nr_diff_of_attrs > 0)
    {
			diff_of_attrs = (diff_of_attrs_type_p)(&compared_with_buffer_p->buffer[sizeof(compared_with_type)]);
      add_to_buffer_c(tag_open_for_change_attributes,
                      strlen(tag_open_for_change_attributes));
      attributes = father_node->attributes;

      for (i = 1; i <= nr_diff_of_attrs; i++)
      {
				if (i > 1)
          add_to_buffer_c(" ", 1);
        seq_nr = diff_of_attrs->diffattr[i-1];
        proceed = true;
        j = 0;
        while (proceed)
        {
					if (seq_nr ==
                    attributes->attr_info[j].sequence)
          {
						help_name =
              (names_of_attributes)[attributes->
                          attr_info[j].attr_name];
            proceed = false;
          };
          j++;
        };
        add_to_buffer_c(help_name->symbol_string,
                      strlen(help_name->symbol_string));
      };
      add_to_buffer_c(tag_close_for_change_attributes,
                      strlen(tag_close_for_change_attributes));
    };
  };

  if ( (!father_node->leaf) &&
       (strcmp(father_node->tag, "top") != 0) )
  {   /* write tagstring */
    add_to_buffer_c(data_b + father_node->start_of_tagstring,
										father_node->end_of_tagstring -
												 father_node->start_of_tagstring + 1);
  };
  if (father_node->leaf) /* text symbol */
  {
    write_text_for_node(father_node);
  }
  else
  { /* if the text is empty; then father_node->u.f.down_start == NULL */
    if (father_node->u.f.down_start != NULL )
    {
      child_node = father_node->u.f.down_start;
      i = 1;
      while (child_node != NULL)
      {
        write_crs(i,
									&current_crs_in_original,
                  &current_cr_position);
				if (father_node->mixed_mode)
					/* if in mixed_mode, get the updated current_diff_marker_p */
					current_diff_marker_p = father_node->current_diff_marker;
				write_diff_tags_on_node_level(/* i */ child_node,
									&current_diff_marker_p,
									node);
				write_for_node(child_node);
	      child_node = child_node->next;
        i++;
      };
    };
 
   /* maybe a current_diff_marker was waiting */

#ifdef proefje //31-10-2002
		if ( (father_node->up != NULL) && (father_node->up->mixed_mode) )
			/* if in mixed_mode, get the updated current_diff_marker_p */
		{
			current_diff_marker_p = father_node->up->current_diff_marker;
			while ( (current_diff_marker_p != NULL) 
					 && ( ( (current_diff_marker_p->chars_or_node_B == node) 
							 /* && ((current_diff_marker_p->position_b) == node_b ) */
								)
						 || ( (current_diff_marker_p->chars_or_node_B == chars) 
							 && (child_node == current_diff_marker_p->position_node_b) 
							 /* && ( current_diff_marker_p->position_b == position_b) */
								)
							) 
						)   /* nog even bijgewerkt ============= 4-9-2002, ochtend */
			{
				write_diff_tag(current_diff_marker_p);
				current_diff_marker_p = (current_diff_marker_p)->next;
				father_node->up->current_diff_marker = current_diff_marker_p;
			}
		}
		else
#endif

		if (father_node->mixed_mode) /* 6-11-2002 */
			/* if in mixed_mode, get the updated current_diff_marker_p */
			current_diff_marker_p = father_node->current_diff_marker;
		while (current_diff_marker_p != NULL) /* 6-11-2002 */
		{
		 write_diff_tag(current_diff_marker_p);
		 current_diff_marker_p = current_diff_marker_p->next;
		}

    /* maybe more cr's were waiting */
    while (current_crs_in_original != NULL)
    {
      add_to_buffer_c(&eol_char, 1);
      current_crs_in_original = current_crs_in_original->next;
    };
  };
  /* print the endtag */
  if ( (!father_node->leaf) &&
       (strcmp(father_node->tag, "top") != 0) &&
       !father_node->empty )   /* the tag is not empty */
  {
		add_to_buffer_c("</", 2);
		/* determine in temp_bufferpos the last position of starttag in data_b */
		temp_bufferpos = father_node->start_of_tagstring;
		while ((data_b[temp_bufferpos] != ' ') &&
					 (data_b[temp_bufferpos] != '>') )
		{
			temp_bufferpos++;
		}
    /* print the tag */
		add_to_buffer_c(data_b + father_node->start_of_tagstring + 1,
										temp_bufferpos - father_node->start_of_tagstring - 1);
    add_to_buffer_c(">", 1);
  }
} /* write_for_node */

/* =====================  write_to_outputbuffer  =====================
Purpose:  writes a complete texttree to the outputbuffer
Parameters:   none
Algorithm:    performs the initial call to write_for_node
----------------------------------------------------------------------*/
void write_to_outputbuffer(void)
{
  /* len_c = 0; */
  tag_change = false;
  diff_tag_open_a = false;
  diff_tag_open_b = false;
  write_for_node(top_node_b);
} /* write_to_outputbuffer */


/* =====================  interpret_last_rev_marker_as_a_change  ============================
Purpose:    interprets the last open revision marker as a change marker
Parameters:
  input:          kind_of_diff : one of the 2 values:
                      - diff_open_a_has_more
                      - diff_open_b_has_more
Global variables: tag_change: indicates if the last open revision tag is changed into a <rev>
									after_an_end_revtag_nothing_has_been_written
									type_of_last_open_revision_tag:	can be
																												diff_open_a_has_more,
																												diff_open_b_has_more,
																												diff_open_change

Algorithm:
  - if the buffer ends on <endX>
    - if X != Y and X != 'rev'
      - remember <X> as the last open <chg>
    - return true
  - else return false (buffer did not end with a revision tag)
----------------------------------------------------------------------*/
int interpret_last_rev_marker_as_a_change(enum kind_of_diff_type kind_of_diff)
  { /* interpret_last_rev_marker_as_a_change */

	/*- if the buffer ends on <endX> */
		if (after_an_end_revtag_nothing_has_been_written)
    {
/*  - if X != Y and X != 'rev' */
      if ( (type_of_last_open_revision_tag != kind_of_diff) &&
           (type_of_last_open_revision_tag != diff_open_change) )
/*    - remember <X> as the last open <chg> */
      {
        type_of_last_open_revision_tag = diff_open_change;
        tag_change = true;
      };
      return true;
    }
    else  /* the buffer does not end on <endX> */
      return false;
  } /* interpret_last_rev_marker_as_a_change */


/* =====================  check_for_rev_sequence  ============================
	(corresponds to function write_diff_tag)
Purpose:          adds a pair of shadow nodes upon the detection of a rev sequence of diff markers
Parameters:
  input:          current_diff_marker : one of the 4 values:
                      - diff_open_a_has_more
                      - diff_close_a_has_more
                      - diff_open_b_has_more
                      - diff_close_b_has_more
									node_b : current node
Global variables:
	- start_of_last_rev_a: position in A when the last addition to B started;
			to be overwritten each time a deletion of A starts
	- start_of_last_rev_b: position in B when the last addition to B started;
			to be overwritten each time a deletion of A starts
	- end_of_last_rev_a: position in A when the last deletion to B ended;
			to be overwritten each time a change has been detected
	- end_of_last_rev_b: position in B when the last deletion to B ended;
			to be overwritten each time a change has been detected
	- shadow_node_a, to be created on the end of a change
	- shadow_node_b, to be created on the end of a change
	- shadow_node_pair, contains a shadow_node_a, a shadow_node_b and the start
			and the end of a sequence of diff_markers which have to be replaced by
			a recalculated set of diff_markers
	- all_shadow_nodes_pair: set of shadow_node_pair(s) which have to be processed with the
			required atomic_unit_of_text_for_readfile

Algorithm:
	- dependent on one of the 4 values a node is added as a daughter node
			to its shadow parent
  - let us call Y = '<add>' or '<del>'; call the corresponding <endY>
  - upon adding <Y> to the output buffer:
    - adjust_last_rev_marker:
  - upon adding <endY> to the output buffer:
    - if tag_change
			- remember the start and end of the last deletion of A as the indexnumbers of
			  the nodes in A: start_of_last_rev_a and end_of_last_rev_a
			- remember the start and end of the last addition to B as the indexnumbers of
				the nodes in B: start_of_last_rev_b and end_of_last_rev_b
			- remember the start and end of the diff_markers which indicate the revision:
				start_of_diff_markers_to_be_replaced and end_of_diff_markers_to_be_replaced;
				the diff_markers in between will have to be replaced by the recalculated
				diff_markers for shadow_node_b
			- create a new artificial parent in A: shadow_node_a
				- add to shadow_node_a the children start_of_last_rev_a and end_of_last_rev_a
			- create a new artificial parent in B: shadow_node_b
				- add to shadow_node_b the children start_of_last_rev_b and end_of_last_rev_b
			- construct a shadow_node_pair with (shadow_node_a, shadow_node_b,
					start_of_diff_markers_to_be_replaced, end_of_diff_markers_to_be_replaced)
			- add shadow_node_pair to all_shadow_nodes_pair: the set of shadow nodes which
				have to be processed with the required atomic_unit_of_text_for_readfile
----------------------------------------------------------------------*/
void check_for_rev_sequence(diff_marker_type_p *current_diff_marker_p_p)
{
	enum kind_of_diff_type kind_of_diff;

	kind_of_diff = (*current_diff_marker_p_p)->kind_of_diff;
	switch (kind_of_diff)
  {
		case diff_open_a_has_more:
		case diff_open_b_has_more:
      if (!interpret_last_rev_marker_as_a_change(kind_of_diff))
      {
				start_of_last_rev_a = (*current_diff_marker_p_p)->position_node_a;
				start_of_last_rev_b = (*current_diff_marker_p_p)->position_node_b;
        type_of_last_open_revision_tag = kind_of_diff;
				start_of_diff_markers_to_be_replaced_p_p = current_diff_marker_p_p;
      }
			else
      {
        type_of_last_open_revision_tag = diff_open_change;
			};
			/* if the last rev marker has to be interpreted as a change then the two
				 start_of_last_rev_(a,b) remain the same */
      if (kind_of_diff == diff_open_a_has_more)
				diff_tag_open_a = true;
			else
				diff_tag_open_b = true;
    break;

    case diff_close_a_has_more:
    case diff_close_b_has_more:
 			end_of_last_rev_a = (*current_diff_marker_p_p)->position_node_a;
 			if (end_of_last_rev_a != NULL)
				end_of_last_rev_a = (end_of_last_rev_a)->former;
			end_of_last_rev_b = (*current_diff_marker_p_p)->position_node_b;
 			if (end_of_last_rev_b != NULL)
				end_of_last_rev_b = end_of_last_rev_b->former;

			if (type_of_last_open_revision_tag == diff_open_change)
			{	/* a revision is closed; add a shadow_node element */
				node_type_p shadow_node_a_p, shadow_node_b_p;
				long sequence_nr_end_of_last_rev_a, sequence_nr_end_of_last_rev_b;
				shadow_node_pair_type_p shadow_node_pair_p;

				shadow_node_a_p	= my_calloc(1, sizeof(node_type));
				shadow_node_b_p	= my_calloc(1, sizeof(node_type));
				shadow_node_a_p->node_has_to_be_compared = true;
				shadow_node_b_p->node_has_to_be_compared = true;
				
				shadow_node_pair_p	= my_calloc(1, sizeof(shadow_node_pair_type));
				shadow_node_pair_p->node_a = shadow_node_a_p;
				shadow_node_pair_p->node_b = shadow_node_b_p;
				shadow_node_pair_p->start_of_diff_markers_to_be_replaced_p_p =
					start_of_diff_markers_to_be_replaced_p_p;
				/* remember the current_diff_marker_p_p as the end of diff_markers
				   that will have to be replaced */
				end_of_diff_markers_to_be_replaced_p_p = &((*current_diff_marker_p_p)->next);
				shadow_node_pair_p->end_of_diff_markers_to_be_replaced_p_p =
					end_of_diff_markers_to_be_replaced_p_p;
				shadow_node_pair_p->next = all_shadow_nodes_pair_p;
				all_shadow_nodes_pair_p = shadow_node_pair_p;

				shadow_node_a_p->u.f.down_start = start_of_last_rev_a;
				shadow_node_b_p->u.f.down_start = start_of_last_rev_b;
				if (end_of_last_rev_a == NULL)
				{
					sequence_nr_end_of_last_rev_a = (start_of_last_rev_a->up)->u.f.nr_of_nodes_down;
						/* which is the number of children nodes of the father */
					shadow_node_a_p->u.f.down_end = (start_of_last_rev_a->up)->u.f.down_end;
				}
				else
				{
					sequence_nr_end_of_last_rev_a = end_of_last_rev_a->sequence_nr_of_node;
					shadow_node_a_p->u.f.down_end = end_of_last_rev_a;
				}
 				shadow_node_a_p->u.f.nr_of_nodes_down = sequence_nr_end_of_last_rev_a -
							start_of_last_rev_a->sequence_nr_of_node + 1;

				if (end_of_last_rev_b == NULL)
				{
					sequence_nr_end_of_last_rev_b = (start_of_last_rev_b->up)->u.f.nr_of_nodes_down;
						/* which is the number of children nodes of the father */
					shadow_node_b_p->u.f.down_end = (start_of_last_rev_b->up)->u.f.down_end;
				}
				else
				{
					sequence_nr_end_of_last_rev_b = end_of_last_rev_b->sequence_nr_of_node;
					shadow_node_b_p->u.f.down_end = end_of_last_rev_b;
				}
 				shadow_node_b_p->u.f.nr_of_nodes_down = sequence_nr_end_of_last_rev_b -
							start_of_last_rev_b->sequence_nr_of_node + 1;
      }
			else
			{
        after_an_end_revtag_nothing_has_been_written = true;
      }
      if (kind_of_diff == diff_close_a_has_more)
				diff_tag_open_a = false;
			else
				diff_tag_open_b = false;
    break;
  }; /* switch */
} /* check_for_rev_sequence */

/* =====================  process_diff_markers  ============================
Purpose:          detect a rev sequence in a list of diff markers
Parameters:
  input:          node_b : current (sub)node
  input/output:   current_diff_marker_p : current element in the list of
                                          diff markers in the current node
Algorithm:        position_b is compared with
									current_diff_marker_p->position_b, iteratively in
									subsequent elements of the list of diff markers;
                  as long as current_diff_marker_p->position_b is equal
									a check is performed if a revision sequence is present
----------------------------------------------------------------------*/
void process_diff_markers(node_type_p node_b,
													diff_marker_type_p_p *current_diff_marker_p_p)
{
	while ( (**current_diff_marker_p_p != NULL) &&
				(node_b == (**current_diff_marker_p_p)->position_node_b) )
			/* this is a 'while' and not an 'if'
			because there may appear subsequent diff's */
	{
		check_for_rev_sequence(*current_diff_marker_p_p);
		*current_diff_marker_p_p = &((**current_diff_marker_p_p)->next);
	};
} /* process_diff_markers */


/* =====================  follow_diffs_and_create_shadow_nodes  ============================
	(Corresponds with write_for_node)
Purpose:  detect and mark nodes which are different for atomic_unit_of_text = 4
Parameters:
  input:  father_node : the top node for the (sub)tree
Global variables:
Algorithm:
  - do recursively for all subnodes: process_diff_markers
  - process remaining diff markers
----------------------------------------------------------------------*/
void follow_diffs_and_create_shadow_nodes(node_type_p father_node)
{
  node_type_p child_node;
	int i;
  diff_marker_type_p *current_diff_marker_p_p;

	current_diff_marker_p_p = &(father_node->start_diff_marker);

  /* forget the difference of attributes */

  /* simulate the printing of the starttag of the node */
	after_an_end_revtag_nothing_has_been_written = false;

	if (!father_node->leaf)
  {
    if (father_node->u.f.down_start != NULL )
    {
      child_node = father_node->u.f.down_start;
      i = 1;
      while (child_node != NULL)
      {
        process_diff_markers(child_node, &current_diff_marker_p_p);
        follow_diffs_and_create_shadow_nodes(child_node);
        child_node = child_node->next;
        i++;
      };
    };
    /* maybe a current_diff_marker was waiting */
    while (*current_diff_marker_p_p != NULL)
    {
      check_for_rev_sequence(current_diff_marker_p_p);
      current_diff_marker_p_p = &((*current_diff_marker_p_p)->next);
    };
  };

  /* simulate the printing of the endtag of the node */
	after_an_end_revtag_nothing_has_been_written = false;

} /* follow_diffs_and_create_shadow_nodes */


/* =====================  create_all_shadow_nodes  =====================
		(Corresponds with write_to_outputbuffer)
Purpose:  detect and mark all nodes which are different for atomic_unit_of_text = 4
Parameters:   none
Algorithm:    performs the initial call to mark_different_nodes
---------------------------------------------------------------------- */
void create_all_shadow_nodes(void)
{
  diff_tag_open_a = false;
  diff_tag_open_b = false;
	after_an_end_revtag_nothing_has_been_written = false;
  follow_diffs_and_create_shadow_nodes(top_node_b);
	free_d = 0;
} /* create_all_shadow_nodes */


/* =====================  recalculate_atomics  =====================
Purpose:  recalculate the atomics for a subtree
  input:  node_p : the top node for the (sub)tree for which the atomics have to
										be recalculated
Algorithm: recursively recalculate the atomic units of text for all the
						 nodes in the subtree
---------------------------------------------------------------------- */
void recalculate_atomics(char * textspace_p, node_type_p node_p)
{
	node_type_p child_node_p;

/* no, else information about attribute difference disappears  ==21-9-2001==
	node_p->compared_with = filepos_null;
*/
	if (node_p->leaf)
	{
		calculate_hash_codes_for_text(textspace_p, node_p);
	}
	else
	{
    child_node_p = node_p->u.f.down_start;
    while (child_node_p != NULL)
    {
      recalculate_atomics(textspace_p, child_node_p);
      child_node_p = child_node_p->next;
    };
	}
}


/* =====================  add_compressed  ==========================
Purpose:  allocates and fills a record for a compressed d value
Parameters:
  input:          kind_of_from : values 1, 2 and 3
Global variables: total_allocated : total amount of allocated heap space
                  compr_from_last_p : pointer to tail of the list
                                      compressed_from of the current node
Algorithm:
    - allocates a record of type compressed_from_type
    - fills the record with kind_of_from
    - attaches the record at the tail of the list compressed_from
      of the current node, pointed to by compr_from_last_p
----------------------------------------------------------------------*/
void add_compressed(my_byte kind_of_from)
{ compressed_from_type_p compr_from_p;

  compr_from_p = my_malloc(sizeof(compressed_from_type));
  compr_from_p->kind_of_from = kind_of_from;
  compr_from_p->next = NULL;
  compr_from_last_p->next = compr_from_p;
  compr_from_last_p = compr_from_p;
} /* add_compressed */

/* =====================  compress_diffs  ==========================
Purpose:  compresses a d-array into a list according to the
    optimal path that can be retraced in d, starting at the end of d
Parameters:
  input:          start_of_d : index in d_value
                  i : number of columns in d
                  j : number of rows in d
  output:         return: pointer to start of created list of
                          compressed d values
Global variables: total_allocated :
                  compr_from_last_p : pointer to end of created list of
                                      compressed d values
Algorithm:
    - follow the optimal path in d, from d[i, j] to d[0, 0]
    - store the encountered values of d in a list
    - store the end of the list in compr_from_last_p
    - return the start of the list
----------------------------------------------------------------------*/
void compress_diffs(char *compressed_array_p, long max_array, 
			long *size_of_compressed_array,
			long start_of_d, long i, long j)
{
	long n;
    *size_of_compressed_array = -1;
		
    n = j;
    while ( (j > 0) && (i > 0) )
    {
			switch (d_from(start_of_d, i, j, n))
      {
				case 1:
          {
						compressed_array_p[++*size_of_compressed_array] = '1';
            i--;
            j--;
          };
					break;
        case 2:
          {
						compressed_array_p[++*size_of_compressed_array] = '2';
            i--;
          };
					break;
        case 3:
          {
 						compressed_array_p[++*size_of_compressed_array] = '3';
           j--;
          };
					break;
				default:
					{
						sprintf(buffer_sprintf, "Internal error. See logfile.\n");
						fprintf(logfile, "Value of d_from not 1, 2 or 3.\n");
						fprintf(logfile, "Report to Palstar.\n");
						fflush(logfile);
						result = value_of_d_out_of_range;
						longjmp(exception_status, result);
					}
      } /* switch */ ;
			if (*size_of_compressed_array >= max_array)
			{
				sprintf(buffer_sprintf, "Sorry, compared_with_buffer too small. See logfile.\n");
				fprintf(logfile, "Compared_with_buffer too small\n");
				fprintf(logfile, "Increase the program constant max_compared_with_buffer_size\n");
				fflush(logfile);
				result = compared_with_buffer_too_small;
				longjmp(exception_status, result);
			}

    } /* while (j > 0) && (i > 0) */ ;

    /* walk towards the (0, 0) cell in the matrix d */
    if (i > 0 )
		{
      while (i > 0)
      {
				compressed_array_p[++*size_of_compressed_array] = '2';
        i--;
      };
		}
    if (j > 0 )
		{
      while (j > 0)
      {
   			compressed_array_p[++*size_of_compressed_array] = '3';
				j--;
      };
		}
		compressed_array_p[++*size_of_compressed_array] = '0';

	} /* compress_diffs */

/* =====================  compare_attributes  =======================
Purpose:  compares two arrays A and B of attribute triples and
          creates an array C with sorted sequencenumbers of the
          elements of B which do not appear in A or which have a
          different value
Parameters:
  input:          A, B : array of triples
                          ( attr_name, attr_value, sequence); the last
                          triple is always a dummy with
                          attr_name = LONG_MAX. The triples of an array
                          are sorted on attr_name.
  output:         C : array of sequencenumbers in B
                  nr_of_c : number of elements of C
Algorithm:
    The criterion for the appearance in C of a sequencenumber is :
      - its corresponding attr-name appears in list B only
      - or: if its corresponding attr_name appears in both lists but the
            corresponding attr_values are different
    The algorithm is straightforward by comparing A and B
    sequentially and adding elements to C if appropriate.
    At the end C is sorted.
----------------------------------------------------------------------*/
void compare_attributes(attr_triple_type *A,
                        attr_triple_type *B,
                        diff_of_attrs_type *C,
                        short *nr_of_c,
												enum to_be_compared_type *to_be_compared)
{ short ia, ib, ic, proceed;
  boolean nodes_may_be_considered_as_equal, 
					nodes_may_be_considered_as_different,
					nodes_have_to_be_compared_for_positive_discrimination,
					nodes_have_to_be_compared_for_negative_discrimination,
					nodes_have_to_be_compared,
					attribute_values_are_equal;
	int (*compare_p)(); /* becomes pointer to the procedure compare_short;
        necessary for MVS */

  ia = 0;
  ib = 0;
  ic = 0;
	nodes_may_be_considered_as_different = false;
	nodes_may_be_considered_as_equal = false;
	nodes_have_to_be_compared = false;
	nodes_have_to_be_compared_for_positive_discrimination = false;
	nodes_have_to_be_compared_for_negative_discrimination = false;

  proceed = true;
  while (proceed)
  {
		while (A[ia].attr_name < B[ib].attr_name)
    { /* At this point it is possible to bring A[ia].sequence
          also to C; however, in that case we get mixed
          sequencenumbers of A and B; a further elaboration
          will be necessary. ( Maybe storing negative ? ) */
      ia++;
    };
    while ( (A[ia].attr_name == B[ib].attr_name) && proceed)
    {
			if (A[ia].attr_name == LONG_MAX)
			{
        proceed = false;
			}
      else
      {	/* if an attribute is discriminative the decisions are:

											if attribute values   are equal					are not equal
				-	if attribute is
						not_discriminative							compare							compare
						positive_discriminative					compare							don't compare; nodes may be considered as different
						negative_discriminative					don't compare;			compare
																						nodes may be considered as equal

				- there may be combinations of attributes which influence each other. 
					Example: an element chapter with a version
					- the chapter attribute CH is positive_discriminative
					- the version attribute VERSION is negative_discriminative

																	if	VERSION is equal		VERSION is not equal
						if CH is equal						!compare, equal				compare
						if CH is not equal				!compare, different		!compare, different

					- the 3 resulting combinations are represented by three booleans:
						- nodes_have_to_be_compared:						compare
						- nodes_may_be_considered_as_equal:			!compare, equal  
						- nodes_may_be_considered_as_different:	!compare, different

				-	if for all positive discriminative attributes holds that the nodes have to be compared then the comparison 
					has to take place; else the nodes may be considered as being different (check: consider the values of 
							pos. discr. attributes as being appended to the element name)
					- implementation with nodes_may_be_considered_as_different is TRUE: logical AND of all occurrences

				- if for all negative discriminative attributes holds that the nodes do not have to be compared then the comparison 
					does not have to take place and the nodes may be considered as equal (check: consider the values of 
							neg. discr. attributes as being appended together to form one value of one neg. discr. attribute)
					- implementation with nodes_may_be_considered_as_equal is TRUE: logical AND of all occurrences

				- at the end: if there is a conflict between nodes_may_be_considered_as_equal and nodes_may_be_considered_as_different
						(that is, both are TRUE) then the conflict is resolved by letting nodes_may_be_considered_as_different win.
						An alternative could be to favor a comparison. However, we can imagine a situation where revxml is solely used to react on 
						version indicators. The decision could be made explicit by introducing a user definable 
						switch in the def-file. To be done if anyone asks for it. (20-9-2001)
				*/
				
				attribute_values_are_equal = (A[ia].attr_value == B[ib].attr_value);
				if (B[ib].discrimination == positive_discriminative)
				{
					if (attribute_values_are_equal && !nodes_may_be_considered_as_different)
					{
						nodes_have_to_be_compared_for_positive_discrimination = true;
					}
					else
					{
						nodes_may_be_considered_as_different = true;
					}
				}
				else if (B[ib].discrimination == negative_discriminative)
				{
					if (attribute_values_are_equal && !nodes_have_to_be_compared_for_negative_discrimination)
					{
						nodes_may_be_considered_as_equal = true;
					}
					else
					{
						nodes_have_to_be_compared_for_negative_discrimination = true;
					}
				}

				if (!attribute_values_are_equal)
				{
					C->diffattr[++ic-1] = B[ib].sequence;
				}
        ia++;
        ib++;
      };
    };
    while (A[ia].attr_name > B[ib].attr_name && proceed)
    {
			C->diffattr[++ic-1] = B[ib].sequence;
      ib++;
    };
  };

	if (nodes_have_to_be_compared)
	{
		*to_be_compared = compare;
	}
	else if (nodes_may_be_considered_as_different)
	{
		*to_be_compared = consider_as_different ;
	}
	else if (nodes_may_be_considered_as_equal)
	{
		*to_be_compared = consider_as_equal;
	}
	else
	{
		*to_be_compared = compare;
	}

	*nr_of_c = ic;
  compare_p = compare_short;
  qsort((void *)C, (size_t)ic, sizeof(short), compare_p); /* sort C */
												/* ignore compiler warning */
}  /* end compare_attributes */

/* =====================  node_diff  ===============================
Purpose:          compares two (sub)trees A and B for difference
Parameters:
  input:          node_a, node_b : top nodes of A and B
  output:         return : number of differences of A and B
Global variables:
                  start_of_d :
                  free_d :
                  total_d :
                  total_allocated :
Algorithm:
  - if node_a and node_b contain distinct tags
    - node_diff_result := 0
    - exit function
  - m := the number of the daughter nodes of A
         or the number of characters in A
         or the number of hash coded pieces of text in A
  - n := the number of the daughter nodes of B
         or the number of characters in B
         or the number of hash coded pieces of text in B
  - if in A and B characters are compared: m and n are taken for
    the infixes of the texts A and B that are different;
    in start_of_diff_a, end_of_diff_a, start_of_diff_b and end_of_diff_b
    the positions of the two infixes are determined
  - d[i, 0] := 0 for i = 0,..., m
  - d[0, j] := 0 for j = 0,..., n
  - for i := 1 to m do
    - for j := 1 to n do
      - d[i, j] := max (d[i-1, j-1] + w[i, j], d[i-1, j], d[i, j-1])
        where w[i, j]
          = node_diff(Ai, Bj) if Ai and Bj are the
            ith and jth daughter nodes of A and B;
          = repl_char(Ai, Bj) if Ai and Bj are characters
          = repl_long(Ai, Bj) if Ai and Bj are compressed parts
            of #pcdata;
            repl_char[i, j] and repl_long(Ai, Bj)
            = 1 if Ai = Bj
            = 0 otherwise
      - from[i, j] := '\' or '^' or '<'
        depending on the parameter of max which was the maximum
        (1st, 2nd or 3rd)
  - node_diff_result := d[m, n] + 1
    the + 1 happens because of the equality of the tags at the root
    (except for artificial tags)

  - allocate a record compared_with
  - store in compared_with
    - the reference to A
    - compare the attributes of the two nodes and store
      the sequencenumbers of the attributes which differ in
      compared_with, field diff_of_attrs
    - subtract the number of different attributes from node_diff_result
      and store node_diff_result in compared_with
    - the bounderies of the infixes of unequal text
    - compress the matrix d of the comparison in a compressed list
      and store a reference to the list in compared_with
  - insert compared_with in front of the list of compared_with's
    of node_b

  - give back space of d which was used in this call to node_diff

________________________

Improved algorithm for the treatment of mixed content.
Date: 25 July 2002

Problem.
Currently, the indication of revisions for mixed content is sometimes too wide. It is not incorrect, but 
sometimes too much is indicated.

General solution
- identify in A and B nodes with mixed content
- if two mixed nodes are to be compared, and the atomic unit is not pcdata, simulate the following:
	- treat the mixed node as consisting of atomic units: treat a node as an atomic unit
	- if a node is to be compared with an atomic unit
		- the difference is 1
	- else the normal algorithm

The simulation has the following consequence:
	- the pcdata does not end at the end of a pcdata-node, it continues with the next node

Algorithm:
- position_within_mixed_node = 0 (init at local variable)
- last_position_within_mixed_node = (init at local variable)
- current_mixed_node_is_exhausted = false
- call position_at_next_atom_within_mixed_node
- while not current_mixed_node_is_exhausted
	- in A: get_next_atom
	- in B: get_next_atom
	- if a node is to be compared with an atomic unit
		- the difference is 1
	- else the normal algorithm

Function get_next_atom (guardian: current_mixed_node_is_exhausted)
- position_at_next_atom_within_mixed_node = false
- if next_atom_is_within_pcdata_node
	- get next atom from pcdata_node
	- increment position_lh_within_pcdata_node
	- if position_lh_within_pcdata_node > last_position_lh_within_pcdata_node
		- position_at_next_atom_within_mixed_node = true
- else
	- get next atom as non_pcdata_node
	- position_at_next_atom_within_mixed_node = true

- if position_at_next_atom_within_mixed_node == true
	- increment position_within_mixed_node
	- if position_within_mixed_node > last_position_within_mixed_node
		- current_mixed_node_is_exhausted = true
	- else
		- if atom at position_within_mixed_node is pcdata_node
			- next_atom_is_within_pcdata_node = true	
				-- go down within pcdata node --
			- position_lh_within_pcdata_node = (use local variable)
			- last_position_lh_within_pcdata_node = (use local variable)
		- else
			- next_atom_is_within_pcdata_node = false

Local variables used:

current_mixed_node_is_exhausted
position_within_mixed_node
last_position_within_mixed_node

position_lh_within_pcdata_node
last_position_lh_within_pcdata_node

next_atom_is_within_pcdata_node

_____________________ end improved algorithm _________________________

----------------------------------------------------------------------*/

long node_diff(node_type_p node_a, node_type_p node_b)
              /* compares two nodes for differences */
{
	long i, j, m, n, value, w, max_array;
  my_byte from;
  node_type_p node_i, node_j;
  long start_of_d, node_diff_result,
    attributes_length, compared_with_total_length,
    end_of_diff_a, end_of_diff_b, nr_of_eq_in_prefix_and_suffix,
		size_of_compressed_array,
  i_end, j_end;
  int equal, leaf;
  short nr_diff_of_attrs, only_blanks_or_dots;
  compared_with_type_p compared_with_p;
	compared_with_buffer_type compared_with_buffer;
	char *compressed_array_p;
	size_t compared_with_length;
	boolean tags_are_equal, compare_nodes, 
					dont_compare_but_register_as_equal, allocated;
	virtual_ptr_type vp;
	enum to_be_compared_type nodes_to_be_compared;

	boolean compare_mixed_nodes,
					current_mixed_node_is_exhausted[2],
					next_atom_is_within_pcdata_node[2];
	node_type_p mixed_node[2], 
							child_lh_node[2],
							result_node[2],
							result_pcdatanode[2];
	char result_char[2],
			 *data[2];
	long position_lh_within_pcdata_node[2],
			 last_position_lh_within_pcdata_node[2],
			 result_long[2];
	boolean next_atom_is_in_new_child[2];

  /* if the tags are different then it depends upon the
	switch 'ignore_diff_in_tags' in the def-file if the nodes
	are considered to be totally different */

	tags_are_equal = (node_a->tagcode == node_b->tagcode);

	if (ignore_diff_in_tags)	 /* switch in def-file */
	{
		compare_nodes = true;
		if (!tags_are_equal)	/* allow for comparison even if
																 the tags are different */
		{	/* but not when one or both nodes are a leaf */
			compare_nodes = !(node_a->leaf) && !(node_b->leaf);
		}
	}
	else
	{
		compare_nodes = tags_are_equal;
	}
  node_diff_result = 0;

  compared_with_length = sizeof(compared_with_type);
	nr_diff_of_attrs = 0;
  
	if ( compare_nodes &&
				(node_a->node_has_to_be_compared == false) &&
				(node_b->node_has_to_be_compared == false) ) /* from spec in .emp file */
	{
		/* don't compare but register as being compared and being equal; 
		for simulation of equalness make node_diff_result = 1: it concerns one tag */
		node_diff_result = 1; /* see below */
		dont_compare_but_register_as_equal = true;
	}
	else
	{
		nodes_to_be_compared = compare; /* default */
		dont_compare_but_register_as_equal = false;
		if (compare_nodes && (node_b->number_of_attributes > 0) )
		{
			/* compare the attributes of the two nodes and store
				 the sequencenumbers of the attributes which differ
				 in diff_of_attrs; in addition, check the discriminating attributes */
			compare_attributes(&node_a->attributes->attr_info[0],
												 &node_b->attributes->attr_info[0],
												 (diff_of_attrs_type_p)(&compared_with_buffer.buffer[compared_with_length]),
												 &nr_diff_of_attrs,
												 &nodes_to_be_compared);
			
			if (nodes_to_be_compared == compare)
			{
				compare_nodes = true;
			}
			else if (nodes_to_be_compared == consider_as_equal)
			{
				/* don't compare but register as being compared and being equal; 
				for simulation of equalness make node_diff_result equal to the number of atomic units within node_b 
				=> to be done;
				temporarily: fake difference at 100  == 18-9-2001 == */
				node_diff_result = 100;
				dont_compare_but_register_as_equal = true;
			}
			else /* must be consider_as_different */ 
			{
				compare_nodes = false; 
			}
		}
  };  

	if (compare_nodes)
	{
		if (dont_compare_but_register_as_equal)
		{
			m = 0; n = 0; /* fake a d table of 0x0 */
		}
		else
		{
			compare_mixed_nodes = ( ( node_a->mixed_mode && (node_b->number_of_pcdata_nodes >= 1) ) ||
															( node_b->mixed_mode && (node_a->number_of_pcdata_nodes >= 1) ) );

			if (compare_mixed_nodes)
			{
				only_blanks_or_dots = false;
				mixed_node[0] = node_a;
				mixed_node[1] = node_b;
				data[0] = data_a;
				data[1] = data_b;
				
				if ( atomic_unit_of_text == rev_pcdata )
				{	/* in this case pcdata nodes contain only one atom, so the total is simply the nr_of_nodes_down */			
					m = node_a->u.f.nr_of_nodes_down;
					n = node_b->u.f.nr_of_nodes_down;				
				}
				else
				{				
					m = mixed_node[0]->number_of_atoms; /* = all atoms within the pcdata nodes + number_of_nonpcdata_nodes */
					n = mixed_node[1]->number_of_atoms; /* idem */
				}
			
				if ((m == 0) && (n == 0))
				{
					only_blanks_or_dots = true;
				}

				start_of_diff_a = 0;		/* as index in data_a */
				start_of_diff_b = 0;		/* as index in data_b */
				end_of_diff_a = m - 1;		/* as index in data_a */
				end_of_diff_b = n - 1;		/* as index in data_b */

				start_of_d = free_d;
				free_d = free_d + (m + 1) * (n + 1);
				 /* reserve room for the matrix d[m, n] */
				if (free_d > d_size)
				{
					fprintf(logfile,
					"Not enough memory for matrix d. Increase the constant\n");
					fprintf(logfile,
						"  d_size, which is now %d, to at least %d\n",
						d_size, free_d);
					result = not_enough_memory_for_d;
					longjmp(exception_status, not_enough_memory_for_d);
				};
				if (free_d > total_d)
				{
					total_d = free_d;
				}

				for (i = 0; i <= m; i++) /* initialize row d[*, 0] */
				{
					store_in_d_space(start_of_d, i, 0, n, 0, 0);
				}
				for (j = 1; j <= n; j++) /* initialize column d[0, *] */
				{
					store_in_d_space(start_of_d, 0, j, n, 0, 0);
				}
		
				/* Main loop */
				
				child_lh_node[0] = mixed_node[0]->u.f.down_start;
				{ init_child_lh_node(0) }
				current_mixed_node_is_exhausted[0] = false;
				i = 0;
				while (!current_mixed_node_is_exhausted[0])
				{
					i++;
				  { get_next_atom(0); }

					child_lh_node[1] = mixed_node[1]->u.f.down_start;
					{init_child_lh_node(1)}
					current_mixed_node_is_exhausted[1] = false;
					j = 0;
					while (!current_mixed_node_is_exhausted[1])
					{
						j++;
						{ get_next_atom(1); }

						/* experimentje, voor het geval er alleen een whitespace is gelezen -- 22-10-2002 
						overblijvend probleem: de <ADD> komt 1 te ver, maar alleen als er 2 elementen staan */
						while (result_long[1] == -1)
						{	get_next_atom(1);
						}

						if ( (result_node[0] != 0) && (result_node[1] != 0) )
						{ /* 2 nodes are the result, compare these in the normal way */
							w = node_diff(result_node[0], result_node[1]);
						}
						else
						if ( ( (result_char[0] != 0) || (result_long[0] != 0) )
									&& ((result_char[1] != 0) || (result_long[1] != 0)) )
						{ /* 2 atoms are the result, compare these in the normal way;
									if equal: w becomes 1, else 0 */
							if ( atomic_unit_of_text == rev_chars )
								w = repl_char(result_char[0], result_char[1]);
							else
								w = repl_long(result_long[0], result_long[1]);
						}
						else
						{/* comparison of an atom with a node, or in reverse; result = 1 */
							w = 0;
						};

						value = maxim3((d(start_of_d, i - 1, j - 1, n) + w),
													 (d(start_of_d, i - 1, j, n)),
													 (d(start_of_d, i, j - 1, n)), &from);
						store_in_d_space(start_of_d, i, j, n, value, from);
					};
				};
				
				node_diff_result = d(start_of_d, m, n, n);

				if ( (!node_b->leaf) &&
						(strcmp(node_b->tag, "top") != 0) )
				{
					node_diff_result++;
					/* because the tag was also equal; but not for artificial tags */
				}
				/* give back space of d which was used in this call of node_diff */
				free_d = free_d - (m + 1) * (n + 1);


			} /* end compare_mixed_nodes */				
			
			else

			{	/* not compare_mixed_nodes */
				leaf = node_b->leaf; /* which is also node_a->leaf */
				only_blanks_or_dots = false;
				if (leaf )
					/* We will compare text. If
					atomic_unit_of_text != rev_chars then
					we use the hashcoding to determine if the two
					#pcdata's are different. We initialize m and n to be
					prepared for the calculation of d */
				{
					if ( atomic_unit_of_text != rev_chars )
					{
						m = node_a->u.t.length_of_hash_coded_text;
						n = node_b->u.t.length_of_hash_coded_text;
						start_of_diff_a = 0;
						start_of_diff_b = 0;
						end_of_diff_a = m - 1;
						end_of_diff_b = n - 1;
						nr_of_eq_in_prefix_and_suffix = 0;
						/* it is possible that no lexical units were coded
								because the text consisted solely of spaces
								(if words are coded) or dots (if sentences are
								coded). In that case m = n = 0. The result of the
								comparison will be 0, which is not fair because
								we do not count spaces or dots for the difference.
								Therefore, we will check this condition at the
								end of the function node_diff and will repair the
								result to 1. The condition is called
								only_blanks_or_dots */
						if ((m == 0) && (n == 0))
						{
							only_blanks_or_dots = true;
						}
					}
					else
					{ /* atomic_unit_of_text = rev_chars. We will
							compare character by character. As an optimization,
							we first determine the common prefixes
							and suffixes.
							Then we calculate the differences in between the common
							prefixes and suffixes. */
						if ( (node_a->u.t.length_of_text <= 2) ||
							 (node_b->u.t.length_of_text <= 2) )
								 /* avoid boundery problems: do not determine
										common prefixes and suffixes
										when the length of at least one of the
										compared texts is <= 2 */
						{
							start_of_diff_a = node_a->u.t.start_of_text;
							start_of_diff_b = node_b->u.t.start_of_text;
							end_of_diff_a = node_a->u.t.end_of_text;
							end_of_diff_b = node_b->u.t.end_of_text;
							nr_of_eq_in_prefix_and_suffix = 0;
						}
						else /* determine start and end of difference */
						{
							/* first, determine start of difference */
							i_end = node_a->u.t.end_of_text;
							j_end = node_b->u.t.end_of_text;
							i = node_a->u.t.start_of_text;
							j = node_b->u.t.start_of_text;
							equal = true;
							while ( (i <= i_end) && (j <= j_end) && equal )
							{
								equal = (data_a[i] == data_b[j]);
								if (equal)
								{ i++;
									j++;
								};
							};
							if (equal) /* guarantee correct bounds */
							{
								start_of_diff_a = i - 1;
								start_of_diff_b = j - 1;
							}
							else
							{
								start_of_diff_a = i;
								start_of_diff_b = j;
							};
							/* 1 <= start_of_diff_a <= node_a->u.t.end_of_text */
							/* 1 <= start_of_diff_b <= node_b->u.t.end_of_text */

							/* determine end of difference */
							i_end = start_of_diff_a;
											/* 1 <= i_end <= node_a->u.t.end_of_text */
							j_end = start_of_diff_b;
											/* 1 <= j_end <= node_b->u.t.end_of_text */
							i = node_a->u.t.end_of_text;
							j = node_b->u.t.end_of_text;
							equal = true;
							while ( (i > i_end) && (j > j_end) && equal)
							{
								equal = (data_a[i] == data_b[j]);
								if (equal)
								{ i--;
									j--;
								};
							};
							end_of_diff_a = i;
								/* start_of_diff_a <= end_of_diff_a
																		<= node_a->u.t.end_of_text */
							end_of_diff_b = j;
								/* start_of_diff_b <= end_of_diff_b
																		<= node_a->u.t.end_of_text */
						};
						m = end_of_diff_a - start_of_diff_a + 1;
							/* m = length of the infix of data_a */
						n = end_of_diff_b - start_of_diff_b + 1;
							/* n = length of the infix of data_b */
						nr_of_eq_in_prefix_and_suffix =
																		node_a->u.t.length_of_text - m;
							/* should be the same as
								 (node_b->u.t.length_of_text - n); has to be added
								 to the difference of the infixes */
					}
				}
				else /* not a leaf */
				{
					m = node_a->u.f.nr_of_nodes_down;
					n = node_b->u.f.nr_of_nodes_down;
					nr_of_eq_in_prefix_and_suffix = 0;
				};

				start_of_d = free_d;
				free_d = free_d + (m + 1) * (n + 1);
				 /* reserve room for the matrix d[m, n] */
				if (free_d > d_size)
				{
					fprintf(logfile,
					"Not enough memory for matrix d. Increase the constant\n");
					fprintf(logfile,
						"  d_size, which is now %d, to at least %d\n",
						d_size, free_d);
					result = not_enough_memory_for_d;
					longjmp(exception_status, not_enough_memory_for_d);
				};
				if (free_d > total_d)
				{
					total_d = free_d;
				}

				for (i = 0; i <= m; i++) /* initialize row d[*, 0] */
				{
					store_in_d_space(start_of_d, i, 0, n, 0, 0);
				}
				for (j = 1; j <= n; j++) /* initialize column d[0, *] */
				{
					store_in_d_space(start_of_d, 0, j, n, 0, 0);
				}
				if (!leaf )
				{
					node_i = node_a->u.f.down_start;
				}
				for (i = 1; i <= m; i++)
				{
					if (!leaf )
					{
						node_j = node_b->u.f.down_start;
					}
					for (j = 1; j <= n; j++)
					{
						if (leaf)
						{ if ( atomic_unit_of_text == rev_chars )
								w = repl_char(data_a[start_of_diff_a + i - 1],
												data_b[start_of_diff_b + j - 1]);
							else
								w = repl_long(node_a->u.t.hashcode_text[i - 1],
												node_b->u.t.hashcode_text[j - 1]);
						}
						else
						{
							w = node_diff(node_i, node_j);
						}
						value = maxim3((d(start_of_d, i - 1, j - 1, n) + w),
													 (d(start_of_d, i - 1, j, n)),
													 (d(start_of_d, i, j - 1, n)), &from);
						store_in_d_space(start_of_d, i, j, n, value, from);
						if (!leaf )
						{
							node_j = node_j->next;
						}
					};
					if (!leaf )
					{
						node_i = node_i->next;
					}
				};
				node_diff_result = d(start_of_d, m, n, n)
															+ nr_of_eq_in_prefix_and_suffix;
			} /* if not mixed_content */

			if (only_blanks_or_dots)
			{
				node_diff_result = 1;
			}

			if ( (!node_b->leaf) &&
					(strcmp(node_b->tag, "top") != 0) )
			{
				node_diff_result++;
				/* because the tag was also equal;
				 but not for artificial tags */
			}
			/* give back space of d which was used in this
					call of node_diff */
			free_d = free_d - (m + 1) * (n + 1);
		} 		/* end if (!dont_compare_but_register_as_equal) */
		
		/* equal attributes contribute to node_diff_result */
    node_diff_result += node_b->number_of_attributes - nr_diff_of_attrs;
		
		/* "allocate space" for the result of the comparison */
		compared_with_p = (compared_with_type_p)&(compared_with_buffer.buffer);
		compared_with_length = sizeof(compared_with_type);

    compared_with_p->node_compared_with = node_a;
		compared_with_p->mixed_nodes = compare_mixed_nodes;
		compared_with_p->nr_diff_of_attrs = nr_diff_of_attrs;
    attributes_length = nr_diff_of_attrs*sizeof(short);
    compared_with_p->result_of_comparison = node_diff_result;
    compared_with_p->nodes_to_be_compared = nodes_to_be_compared;
    compared_with_p->start_of_difference_a = start_of_diff_a;
    compared_with_p->start_of_difference_b = start_of_diff_b;
    compared_with_p->end_of_difference_a = end_of_diff_a;
    compared_with_p->end_of_difference_b = end_of_diff_b;
		
		/* overlay array of compressed froms in compared_with_buffer behind the attribute diff info */
		compressed_array_p = (char *)
								(&(compared_with_buffer.buffer[compared_with_length + attributes_length]));

    /* calculate the compressed froms */
		max_array = cache_page_size - compared_with_length - attributes_length;
		compress_diffs(compressed_array_p, max_array,
										&size_of_compressed_array, start_of_d, m, n);

    /* insert compared_with in front of the list of
        compared_with's of node_b */
    compared_with_p->next = node_b->compared_with;

		/* copy compared_with to cache memory */

		compared_with_total_length = compared_with_length + attributes_length + size_of_compressed_array + 1;

		allocated = alloc_cm_record(cm_cache_p, compared_with_total_length, &vp);
		if (!allocated)
		{
			/* Too large for a page. Report error with advice to enlarge the pagesize. */
			sprintf(buffer_sprintf, "Record too long for page in cache file. Increase pagesize in the def-file");
			result = increase_pagesize;
			longjmp(exception_status, increase_pagesize);
		}

		/* next statement caused segmentation fault under Linux in the past; maybe was not an 8-byte alignment? */
    memcpy(vp.rampos, compared_with_p, compared_with_total_length);

    node_b->compared_with = vp.filepos;

    if (trace)
    {
			fprintf(tracefile, "Node %d  compared with node %d , m = %d n = %d\n",
        (long)node_a, (long)node_b, m, n);
      fprintf(tracefile, " matrix d:\n");
      fprintf(tracefile, "   ");
      for (j = 0; j <= n; j++)
        fprintf(tracefile, "%3d", j);
      fprintf(tracefile, "\n");
      for (i = 0; i <= m; i++)
      {
        fprintf(tracefile, "%1d  ", i);
        for (j = 0; j <= n; j++)
          fprintf(tracefile, "%3d",
          d(start_of_d, i, j, n));
        fprintf(tracefile, "\n");
      };

      fprintf(tracefile, "matrix from:\n");
      fprintf(tracefile, "   ");
      for (j = 0; j <= n; j++)
        fprintf(tracefile, "%3d", j);
      fprintf(tracefile, "\n");
      for (i = 0; i <= m; i++)
      { fprintf(tracefile, "%1d  ", i);
        for (j = 0; j <= n; j++)
          fprintf(tracefile, "%3d",
           d_from(start_of_d, i, j, n));
        fprintf(tracefile, "\n");
      };
  }; /* if trace */
};
return node_diff_result;

} /* node_diff */


//#define old
#ifdef old

/* =====================  add_diff ================================
Purpose:          adds a record in front of the list of differences
Parameters:
  input:          kind_of_diff : 1, 2 or 3
                  node_b : current node
                  i : number of daughter node in A at which the difference
                      occurs (if applicable)
                  j : number of daughter node in B at which the difference
                      occurs (if applicable)
Global variables: total_allocated
Algorithm:
    - allocate a diff marker record diff_marker
    - store in diff_marker:
      - the kind of difference, kind_of_diff
      - the position in A at which the difference occurs:
        - if node_a is not a leaf: i + 1
        - else: bufferp, the position in the text buffer
      - the position in B at which the difference occurs:
        - if node_b is not a leaf: j + 1
        - else: bufferp, the position in the text buffer
    - insert diff_marker in front of the other diff marker records
------------------------------------------------------------------- */
void add_diff(enum kind_of_diff_type kind_of_diff,
							enum chars_or_node_type chars_or_node_A,
							enum chars_or_node_type chars_or_node_B,
              node_type_p father_node_b,
							node_type_p current_node_a,
							node_type_p current_node_b,
							node_type_p position_node_a,
							long bufferp_char_a,
							long bufferp_char_b)
{ /* add_diff */
 /* bookmarker */
  diff_marker_type_p diff_marker;
													
	diff_marker = my_malloc(sizeof(diff_marker_type));
  diff_marker->kind_of_diff = kind_of_diff;
	diff_marker->chars_or_node_A = chars_or_node_A;
	diff_marker->chars_or_node_B = chars_or_node_B;
/* 	if ( (position_node_a != NULL) && 
				( (position_node_a->leaf ) || (position_node_a->up->mixed_mode) ) ) == not sure if this is correct */
	if (chars_or_node_A == chars)
	{
		diff_marker->position_char_a = bufferp_char_a;
	  diff_marker->position_node_a = position_node_a;
	}
	else
    diff_marker->position_char_a = (long)(position_node_a);

	
 	if ( (father_node_b->leaf ) || (father_node_b->up->mixed_mode) )
	{
		diff_marker->position_char_b = bufferp_char_b;
		diff_marker->position_node_b = father_node_b;
	}
	else
		diff_marker->position_char_b = (long)(current_node_b);
  
	if (father_node_b->up->mixed_mode)
	{
		diff_marker->next = father_node_b->up->start_diff_marker;
		father_node_b->up->start_diff_marker = diff_marker;
	}
	else
	{
		diff_marker->next = father_node_b->start_diff_marker;
		father_node_b->start_diff_marker = diff_marker;
	}
} /* add_diff */
#endif


#ifndef old
/* =====================  add_diff ================================
Purpose:          adds a record in front of the list of differences
Parameters:
  input:          kind_of_diff : 1, 2 or 3
                  node_b : current node
                  i : number of daughter node in A at which the difference
                      occurs (if applicable)
                  j : number of daughter node in B at which the difference
                      occurs (if applicable)
Global variables: total_allocated
Algorithm:
    - allocate a diff marker record diff_marker
    - store in diff_marker:
      - the kind of difference, kind_of_diff
      - the position in A at which the difference occurs:
        - if node_a is not a leaf: i + 1
        - else: bufferp, the position in the text buffer
      - the position in B at which the difference occurs:
        - if node_b is not a leaf: j + 1
        - else: bufferp, the position in the text buffer
    - insert diff_marker in front of the other diff marker records
------------------------------------------------------------------- */
void add_diff(enum kind_of_diff_type kind_of_diff,
							enum chars_or_node_type chars_or_node_A,
							enum chars_or_node_type chars_or_node_B,
 							node_type_p father_node_a,
              node_type_p father_node_b,
							node_type_p current_node_a_right,
							node_type_p current_node_b_right,
							long bufferp_char_a,
							long bufferp_char_b)
{ /* add_diff */
 /* bookmarker */
  diff_marker_type_p diff_marker;
	node_type_p up_node;
													
	diff_marker = my_calloc(1, sizeof(diff_marker_type));
  diff_marker->kind_of_diff = kind_of_diff;
	diff_marker->chars_or_node_A = chars_or_node_A;
	diff_marker->chars_or_node_B = chars_or_node_B;

	if (chars_or_node_A == chars)
		diff_marker->position_char_a = bufferp_char_a;

 	if (chars_or_node_B == chars)
		diff_marker->position_char_b = bufferp_char_b;

	diff_marker->position_node_a = /* current_node_a 6-11-2002 */ current_node_a_right;
	diff_marker->position_node_b = /* current_node_b 6-11-2002 */ current_node_b_right;
	
	if (father_node_b != NULL) /* added 1-11-2002 */
		up_node = father_node_b->up;
	else
		up_node = current_node_b_right->up->up;

	if (up_node->mixed_mode)
	{
		diff_marker->next = up_node->start_diff_marker;
		up_node->start_diff_marker = diff_marker;
	}
	else /* hetzelfde == 6-11-2002 */
	{
		diff_marker->next = father_node_b->start_diff_marker;
		father_node_b->start_diff_marker = diff_marker;
	}
} /* add_diff */

#endif
/* =====================  reconstruct_diffs  =========================
Purpose:          reconstructs the minimal differences between two nodes;
                  it uses the compressed arrays d in order to
                  reconstruct the total minimal path of differences
                  through the whole text;
                  the compressed arrays d are already stored from the
                  right-down position to the left-up position;
                  the function reconstruct_diffs calls itself
                  recursively at each node;
                  the reconstructed path within a node is stored in
                  records which are created by calls to the function
                  add_diff; within the records the position
                  within the text or within the daughter nodes where
                  a diff marker has to be inserted is stored; the kind
                  of diff marker is also stored in the record;
                  after the creation of the records the individual
                  compressed matrices d are not longer necessary;
                  at that moment the text tree is prepared to be printed
                  to the output buffer, together with the difference
                  markers

Parameters:
  input:          node_a, node_b : nodes which are compared
                  remember_diff_attrs : boolean
                    if false: node_b is within add markers, therefore it
                              is not necessary to indicate eventual
                              differences of attributes
                    if true:  node_b is compared with node_a; eventual
                              differences of attributes will be
                              indicated
Global variables:
Algorithm:
  - locate in the list node_b->compared_with the record p with node_a
  - in record p start_of_compressed_from_p points to the list
    which represents the compressed d-array for the comparison of
    node_a and node_b
  - process each element of compressed_from: dependent on kind_of_from
    construct, with add_diff, a record with the correct diff_marker
    - i := m (eventually taking care of the common prefix and suffix)
    - j := n (idem)
    - while (i > 0) and (j > 0) do
      - case from[i, j] of
        '\':  - close an eventual open '<del>' or '<add>' with an
                insertion of '<enddel>' or '<endadd>' after the
                j'th character
              - i := i - 1;  j := j - 1
        '^':  - close an eventual open '<add>' with an insertion
                of '<endadd>' after the j'th character
              - if '<del>' not already open: insert a '<del>' after
                the j'th character
              - i := i - 1
        '<':  - close an eventual open '<del>' with an insertion
                of '<enddel>' after the j'th character
              - if '<add>' not already open: insert a '<add>' after
                the j'th character
              - j := j - 1
    - if i > 0
      - sequence A has more initial elements: insert '<del><enddel>'
        in front of B
    - if j > 0
      - sequence B has more initial elements: surround the first j
        elements of B by '<add>' and '<endadd>'

---------------------------------------------------------------------*/
void reconstruct_diffs(node_type_p node_a,
											 node_type_p node_b,
                       short remember_diff_attrs)
{ node_type_p current_node_a,
							current_node_b,
							current_node_a_right,
							current_node_b_right;
  boolean leaf, ready, node_a_mixed, node_b_mixed;
  compared_with_type_p p;
  long i, j, ca_index, compared_with_length, attributes_length;
  long start_of_diff_a, start_of_diff_b, end_of_diff_a, end_of_diff_b,
			 bufferp_char_a, bufferp_char_b, begin_bufferp_char_a, begin_bufferp_char_b;
	compared_with_buffer_type_p compared_with_buffer_p;
	pos_type p_filepos;
	char *compressed_array_p;
	my_int bufslot;
	enum chars_or_node_type chars_or_node_A, chars_or_node_A_right, chars_or_node_B;


  /* reconstruct_diffs */
	compared_with_length = sizeof(compared_with_type);

	node_b->start_diff_marker = NULL;
	node_b->current_diff_marker = NULL;

  /* find the start_of_compressed_from_p for this pair of nodes */
  ready = false;
	p_filepos = node_b->compared_with;
  p = filepos_to_ptr(&p_filepos, &bufslot);
	attributes_length = 0;
  
	while ( (p != NULL) && ! ready )
  {
		if (p->node_compared_with == node_a )
    {
			attributes_length = p->nr_diff_of_attrs * sizeof(short);
			compared_with_buffer_p = (compared_with_buffer_type_p)p;
			compressed_array_p = (char *)
								(&(compared_with_buffer_p->buffer[compared_with_length + attributes_length]));

      ready = true;
      start_of_diff_a = p->start_of_difference_a;
      start_of_diff_b = p->start_of_difference_b;
      end_of_diff_a = p->end_of_difference_a;
      end_of_diff_b = p->end_of_difference_b;
    }
    else
		{
      p_filepos = p->next;
			p = filepos_to_ptr(&p_filepos, &bufslot);
		}
  };
  if (ready)
  { /* if not ready: node_b was not compared with another
    node, for instance
    when it is a non-leaf and the others are #pcdata; we can
    therefore leave the procedure reconstruct_diffs */

    /* node_a and node_b were compared. The
       result of the comparison of the attributes is stored in
       the record which is pointed to by p. After the current
       call to reconstruct_diffs the field "compared_with"
       within node_b will not longer be used.
       Essentially, the list which is pointed to by
       "compared_with" could be released. This will be
       implemented later, if space becomes a problem. In the
       meantime we simply keep p by storing it in "compared_with".
    */
    if (remember_diff_attrs)
		{
      node_b->compared_with = p_filepos;
    }


    ca_index = 0;
		
		/* update 27-7-2002 for mixed nodes */

/* bookmark point */
										
		if (p->mixed_nodes)

		{	/* Observations/Algorithm.
				 
				 - for all child nodes, from last to first
					 - if pcdata
						 - walk from last to first atom
					 - else
						 - take this child node

			*/

			i = node_a->number_of_atoms + 1;
			j = node_b->number_of_atoms + 1;
			node_a_mixed = node_a->mixed_mode;
			node_b_mixed = node_b->mixed_mode;

			current_node_a = node_a->u.f.down_end; /* can be pcdata or other type of node */
			if (current_node_a->pcdata)
			{
				if (atomic_unit_of_text == rev_chars)
				{
					begin_bufferp_char_a = current_node_a->u.t.start_of_text /* -1 8-11-2002 */;
					bufferp_char_a = current_node_a->u.t.end_of_text;
				}
				else	
				{
					begin_bufferp_char_a = 1 /* 0 == 7-11-2002 */;
					bufferp_char_a = current_node_a->u.t.length_of_hash_coded_text /*-1 6-9-2002 */;
				};
				chars_or_node_A = chars;
				chars_or_node_A_right = chars;
			}
			else
			{
				chars_or_node_A = node;
				chars_or_node_A_right = node;
			};


			current_node_b = node_b->u.f.down_end; /* idem */
			if (current_node_b->pcdata)
			{
				if (atomic_unit_of_text == rev_chars)
				{
					begin_bufferp_char_b = current_node_b->u.t.start_of_text /* -1 8-11-2002 */;
					bufferp_char_b = current_node_b->u.t.end_of_text;
				}				
				else	
				{
					begin_bufferp_char_b = 1 /* 0 == 7-11-2002 */;
					bufferp_char_b = current_node_b->u.t.length_of_hash_coded_text /*-1 6-9-2002 */;
				};
				chars_or_node_B = chars;
			}
			else
				chars_or_node_B = node;


			if ( current_node_a != NULL )
			 current_node_a_right = current_node_a->next;
			else
			 current_node_a_right = NULL;

			if ( current_node_b != NULL )
			 current_node_b_right = current_node_b->next;
			else
			 current_node_b_right = NULL;

			while (compressed_array_p[ca_index] != '0')
			{
				boolean goto_former_node;

			
				switch (compressed_array_p[ca_index])
				{ case '1': /* A and B are the same */
						{ 
							if (!closed_for_a)
							{
								if (chars_or_node_B == chars) /* 8-11-2002 */
									add_diff(	diff_open_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								else
									add_diff(	diff_open_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_a = true;
							};
							if (!closed_for_b)
							{
								if (chars_or_node_B == chars) /* 8-11-2002 */
									add_diff(	diff_open_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								else
									add_diff(	diff_open_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_b = true;
							};

							/* =========== block for i ===================*/
							i--;
							/* check if the start of a pcdata node is reached */
							goto_former_node = false;
							if (current_node_a->pcdata)
							{
								bufferp_char_a--;
								if (bufferp_char_a < begin_bufferp_char_a)
								{
									goto_former_node = true;
								}
							}
							else
								goto_former_node = true;
							if (goto_former_node)
							{
								chars_or_node_A_right = chars_or_node_A;
								current_node_a_right = current_node_a;
								current_node_a = current_node_a->former;
								if (current_node_a != NULL)
								{
									if (current_node_a->pcdata)	
									{
										if (atomic_unit_of_text == rev_chars)
										{
											begin_bufferp_char_a = current_node_a->u.t.start_of_text-1;
											bufferp_char_a = current_node_a->u.t.end_of_text /* -1 = 2-11-2002 */;
										}
										else	
										{
											begin_bufferp_char_a = 0;
											bufferp_char_a = current_node_a->u.t.length_of_hash_coded_text-1 /* 6-9-2002 */;
										};
										chars_or_node_A = chars;
									}
									else
									{
										bufferp_char_a = -1; /* 2-11-2002 */
										chars_or_node_A = node;
									}	
								}			
							};
							/* =========== end block for i ===================*/

							/* =========== block for j ===================*/
							j--;
							/* check if the start of a pcdata node is reached */
							goto_former_node = false;
							if (current_node_b->pcdata)
							{
								bufferp_char_b--;
								if (bufferp_char_b < begin_bufferp_char_b)
								{
									goto_former_node = true;
								}
							}
							else
								goto_former_node = true;
							if (goto_former_node)
							{
								current_node_b_right = current_node_b;
								current_node_b = current_node_b->former;
								if (current_node_b != NULL)
								{
									if (current_node_b->pcdata)	
									{
										if (atomic_unit_of_text == rev_chars)
										{
											begin_bufferp_char_b = current_node_b->u.t.start_of_text-1;
											bufferp_char_b = current_node_b->u.t.end_of_text-1;
										}				
										else	
										{
											begin_bufferp_char_b = 0;
											bufferp_char_b = current_node_b->u.t.length_of_hash_coded_text-1 /* 6-9-2002 */;
										};
										chars_or_node_B = chars;
									}
									else
									{
										bufferp_char_b = -1; /* 2-11-2002 */
										chars_or_node_B = node;
									}
								}				
							};
							/* =========== end block for j ===================*/

							if ( (current_node_a != NULL) && (current_node_b != NULL)
										&& (!current_node_a->pcdata) && (!current_node_b->pcdata)
										&& closed_for_b && closed_for_a)
							{
								reconstruct_diffs(current_node_a, current_node_b, true);
								/* the formal parameter remember_diff_attrs has
								to become true because node_a and node_b are
								really compared internally */
							}

						}; break;
					case '2': /* A has more */
						{ 
							if (!closed_for_b)
							{ 
								if (chars_or_node_B == chars) /* 8-11-2002 */
									add_diff(	diff_open_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								else
									add_diff(	diff_open_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_b = true;
							};
							if (closed_for_a )
							{ 
								if (chars_or_node_B == chars) /* 8-11-2002 */
									add_diff(diff_close_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								else
									add_diff(diff_close_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_a = false;
							};
							
							/* cannot be true. Remove later.
							if ( (!current_node_a->pcdata) && (!current_node_b->pcdata)
								 && closed_for_b && closed_for_a)
							{
								reconstruct_diffs(current_node_a, current_node_b, false);
							}
							*/

							/* =========== block for i ===================*/
							i--;
							/* check if the start of a pcdata node is reached */
							goto_former_node = false;
							if (current_node_a->pcdata)
							{
								bufferp_char_a--;
								if (bufferp_char_a < begin_bufferp_char_a)
								{
									goto_former_node = true;
								}
							}
							else
								goto_former_node = true;
							if (goto_former_node)
							{
								chars_or_node_A_right = chars_or_node_A;
								current_node_a_right = current_node_a;
								current_node_a = current_node_a->former;
								if (current_node_a != NULL)
								{
									if (current_node_a->pcdata)	
									{
										if (atomic_unit_of_text == rev_chars)
										{
											begin_bufferp_char_a = current_node_a->u.t.start_of_text-1;
											bufferp_char_a = current_node_a->u.t.end_of_text /* -1 = 2-11-2002 */;
										}
										else	
										{
											begin_bufferp_char_a = 0;
											bufferp_char_a = current_node_a->u.t.length_of_hash_coded_text-1 /* 6-9-2002 */;
										};
										chars_or_node_A = chars;
									}
									else
									{
										bufferp_char_a = -1; /* 2-11-2002 */
										chars_or_node_A = node;
									}	
								}			
							};
							/* =========== end block for i ===================*/


						}; break;
					case '3': /* B has more */
						{ 
							if (!closed_for_a )
							{
								if (chars_or_node_B == chars) /* 8-11-2002 */
									add_diff(	diff_open_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								else
									add_diff(	diff_open_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_a = true;
							};
							if (closed_for_b )
							{
								if (chars_or_node_B == chars) /* 8-11-2002 */
									add_diff(	diff_close_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								else
									add_diff(	diff_close_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_b = false;
							};

							/* cannot be true. Remove later.
							if ( (!current_node_a->pcdata) && (!current_node_b->pcdata)
								 && closed_for_b && closed_for_a)
							{
								reconstruct_diffs(current_node_a, current_node_b, false);
							}
							*/

							/* =========== block for j ===================*/
							j--;
							/* check if the start of a pcdata node is reached */
							goto_former_node = false;
							if (current_node_b->pcdata)
							{
								bufferp_char_b--;
								if (bufferp_char_b < begin_bufferp_char_b)
								{
									goto_former_node = true;
								}
							}
							else
								goto_former_node = true;
							if (goto_former_node)
							{
								current_node_b_right = current_node_b;
								current_node_b = current_node_b->former;
								if (current_node_b != NULL)
								{
									if (current_node_b->pcdata)	
									{
										if (atomic_unit_of_text == rev_chars)
										{
											begin_bufferp_char_b = current_node_b->u.t.start_of_text-1;
											bufferp_char_b = current_node_b->u.t.end_of_text-1;
										}				
										else	
										{
											begin_bufferp_char_b = 0;
											bufferp_char_b = current_node_b->u.t.length_of_hash_coded_text -1 /* 6-9-2002 */;
										};
										chars_or_node_B = chars;
									}
									else
									{
										bufferp_char_b = -1; /* 2-11-2002 */
										chars_or_node_B = node;
									}
								}				
							};
							/* =========== end block for j ===================*/


						}; break;
				} /* switch */ ;
				ca_index++;

				/* buffer swapped out after the (recursive) call to reconstruct_diffs? */
				if (cm_cache_p->buf.adblocks_p[bufslot]->blocknumber != p_filepos.blocknr)
				{	/* yes, swap in again */
					p = filepos_to_ptr(&p_filepos, &bufslot);
					compared_with_buffer_p = (compared_with_buffer_type_p)p;
					compressed_array_p = (char *)
							(&(compared_with_buffer_p->buffer[compared_with_length + attributes_length]));
				}

			} /* while */
  
			if (!closed_for_a)
			{
				if (chars_or_node_B == chars) /* 8-11-2002 */
					add_diff(	diff_open_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
				else
					add_diff(	diff_open_a_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
				closed_for_a = true;
			};

			if (!closed_for_b)
			{
				if (chars_or_node_B == chars) /* 8-11-2002 */
					add_diff(	diff_open_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, current_node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
				else
					add_diff(	diff_open_b_has_more, chars_or_node_A_right, chars_or_node_B,
														node_a, node_b, current_node_a_right, /* 7-11-2002 */
														current_node_b_right, bufferp_char_a, bufferp_char_b);
				closed_for_b = true;
			};
		}  /* end mixed nodes were compared */


		 
		else
		{ /* not mixed nodes were compared */
			
			leaf = node_a->leaf; /* which is also node_b->leaf */

	  	current_node_a_right = NULL;
		  current_node_b_right = NULL;
			if (leaf )
			{
				i = end_of_diff_a - start_of_diff_a + 1;
				j = end_of_diff_b - start_of_diff_b + 1;
				bufferp_char_a = end_of_diff_a;
				bufferp_char_b = end_of_diff_b;
				chars_or_node_A = chars;
				chars_or_node_B = chars;
			}
			else
			{
				i = node_a->u.f.nr_of_nodes_down;
				j = node_b->u.f.nr_of_nodes_down;
				current_node_a = node_a->u.f.down_end;
				current_node_b = node_b->u.f.down_end;
				chars_or_node_A = node;
				chars_or_node_B = node;

				if ( current_node_a != NULL )
				 current_node_a_right = current_node_a->next;

				if ( current_node_b != NULL )
				 current_node_b_right = current_node_b->next;
			};
			
			while (compressed_array_p[ca_index] != '0')
			{
				switch (compressed_array_p[ca_index])
				{ case '1': /* A and B are the same */
						{ 
							if (!closed_for_a)
							{
								add_diff(	diff_open_a_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_a = true;
							};
							if (!closed_for_b)
							{
								add_diff(	diff_open_b_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_b = true;
							};
							if (!leaf && closed_for_b && closed_for_a)
							{
								reconstruct_diffs(current_node_a, current_node_b, true);
							}
							/* the formal parameter remember_diff_attrs has
							to become true because node_a and node_b are
							really compared internally */
							i--;
							j--;
							if (leaf )
							{
								bufferp_char_a--;
								bufferp_char_b--;
							}
							else
							{
								current_node_a_right = current_node_a;
								current_node_b_right = current_node_b;
								current_node_a = current_node_a->former;
								current_node_b = current_node_b->former;
							};
						}; break;
					case '2': /* A has more */
						{ 
							if (!closed_for_b)
							{ add_diff(	diff_open_b_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_b = true;
							};
							if (closed_for_a )
							{ add_diff(diff_close_a_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_a = false;
							};
							/* cannot be true. Remove later.
							if (!leaf && closed_for_b && closed_for_a )
							{
								reconstruct_diffs(current_node_a, current_node_b, false);
							}
							*/
							i--;
							if (leaf)
							{
								bufferp_char_a--;
							}
							else
							{
								current_node_a_right = current_node_a;
								current_node_a = current_node_a->former;
							}
						}; break;
					case '3': /* B has more */
						{ 
							if (!closed_for_a )
							{
								add_diff(	diff_open_a_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_a = true;
							};
							if (closed_for_b )
							{
								add_diff(	diff_close_b_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
								closed_for_b = false;
							};

							/* cannot be true. Remove later.
							if (!leaf && closed_for_b && closed_for_a )
							{
								reconstruct_diffs(current_node_a, current_node_b, false);
							}
							*/

							j--;
							if (leaf)
							{
								bufferp_char_b--;
							}
							else
							{
								current_node_b_right = current_node_b;
								current_node_b = current_node_b->former;
							}
						}; break;
				} /* switch */ ;
				ca_index++;

				/* buffer swapped out after the (recursive) call to reconstruct_diffs? */
				if (cm_cache_p->buf.adblocks_p[bufslot]->blocknumber != p_filepos.blocknr)
				{	/* yes, swap in again */
					p = filepos_to_ptr(&p_filepos, &bufslot);
					compared_with_buffer_p = (compared_with_buffer_type_p)p;
					compressed_array_p = (char *)
							(&(compared_with_buffer_p->buffer[compared_with_length + attributes_length]));
				}
				} /* while */
			if (!closed_for_a)
			{
				add_diff(	diff_open_a_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
				closed_for_a = true;
			};

			if (!closed_for_b)
			{
				add_diff(	diff_open_b_has_more, chars_or_node_A, chars_or_node_B,
													node_a, node_b, current_node_a_right, /* 7-11-2002 */
													current_node_b_right, bufferp_char_a, bufferp_char_b);
				closed_for_b = true;
			};
		} /* end if not mixed nodes were compared */
	}; /* end if ready */ 


} /* reconstruct_diffs */


void get_switch_from_deffile(char *sw)
{
	char tempstring[255];
	fgets(tempstring, 255, deffile);
  sscanf(tempstring, "%s", sw);
}


void read_switches(long *result)
{	

#ifdef MVS
		deffile = fopen("dd:diffdef", "r");
#endif
#ifndef MVS
		deffile = fopen(def_filename, "r");
#endif
		error_file_open(deffile, def_filename, result);
		get_switch_from_deffile(tag_open_for_a_has_more);
		get_switch_from_deffile(tag_close_for_a_has_more);
		get_switch_from_deffile(tag_open_for_b_has_more);
		get_switch_from_deffile(tag_close_for_b_has_more);
		get_switch_from_deffile(tag_open_change);
		get_switch_from_deffile(tag_close_change);
		get_switch_from_deffile(tag_open_for_change_attributes);
		if (tag_open_for_change_attributes[strlen(tag_open_for_change_attributes)-1] != '>')
		{
			/* if it does not end on a ">" we add a " " because it is supposed to be the start
					of a PI */
			strcat(tag_open_for_change_attributes, " ");
		}

		get_switch_from_deffile(tag_close_for_change_attributes);
			/* remove the nl */
			/* the lengths of the start revision tags should be equal
				 the lengths of the end revision tags should be equal */
		if ( (strlen(tag_open_for_a_has_more) !=
																				strlen(tag_open_for_b_has_more))
				 || (strlen(tag_open_for_a_has_more) !=
																				strlen(tag_open_change))
				 || (strlen(tag_close_for_a_has_more) !=
																				strlen(tag_close_for_b_has_more))
				 || (strlen(tag_close_for_a_has_more) !=
																				strlen(tag_close_change)) )
		{
			sprintf(buffer_sprintf,
				"lengths of start or end revision markers should be equal\n");
			fprintf(logfile, buffer_sprintf);
			fprintf(logfile, "Lengths of all start revision markers should be equal, and\n");
			fprintf(logfile, "  lengths of all end revision markers should be equal.\n");
			*result = unequal_lengths_of_rev_markers;
			return;
		}
		length_of_open_revision_tag = strlen(tag_open_for_a_has_more);
		length_of_close_revision_tag = strlen(tag_close_for_a_has_more);

		get_switch_from_deffile(helpline);
		combtags = (strcmp(helpline, "combtags") == 0);

		get_switch_from_deffile(helpline);
		ignore_diff_in_tags = (strcmp(helpline, "ignore_diff_in_tags") == 0);

		get_switch_from_deffile(helpline);
		compare_pi = (strcmp(helpline, "compare_pi") == 0);

		get_switch_from_deffile(helpline);
		compare_comment = (strcmp(helpline, "compare_comment") == 0);

		get_switch_from_deffile(helpline);
		print_a = (strcmp(helpline, "printa") == 0);

		get_switch_from_deffile(helpline);
		sscanf(helpline, "%i", &atomic_unit_of_text);
		if (print_a && combtags)
		{
			sprintf(buffer_sprintf, "Combtags and printa cannot be combined");
			fprintf(logfile, "%s\n", buffer_sprintf);
			*result = combtags_and_printa_both_true;
			return;
		}


#ifdef DEMO_limited_size
		trace = false;
		*alloc_size = 200000;
		*textsize = 3000;
		d_size = 25000;
		max_nr_of_names = 200;
		preproc = false;
#endif

#ifndef DEMO_limited_size
		get_switch_from_deffile(helpline);
		preproc = (strcmp(helpline, "preproc") == 0);
		if (atomic_unit_of_text == 4)
		{
		  preproc = false; /* because preprocessing is also done with
													atomic_unit_of_text = 4 */
		}
		get_switch_from_deffile(helpline);
		trace = (strcmp(helpline, "trace") == 0);

		if (trace)
		{
#ifndef MVS
			tracefile = fopen("revxml.trc", "w");
#endif
#ifdef MVS
			tracefile = fopen("dd:trace", "w");
#endif
			error_file_open(tracefile, "revxml.trc", result);
		}
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &textsize);
		textsize_global = textsize;
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &alloc_size);
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &d_size);
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &max_nr_of_names_of_elements);
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &max_nr_of_names_of_attributes);
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &max_nr_of_names_of_attribute_values);
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &APS);
		cache_page_size = APS << 9;
		if (max_compared_with_buffer_size < cache_page_size)
		{
			sprintf(buffer_sprintf, "Compared_with_buffer too small. See logfile.");
			fprintf(logfile, "Compared_with_buffer too small\n");
			fprintf(logfile, "Increase the program constant max_compared_with_buffer_size\n");
			fflush(logfile);
			*result = compared_with_buffer_too_small;
			longjmp(exception_status, compared_with_buffer_too_small);
		}
		/* ABS now fixed. 21-9-2001
		fgets(helpline, 255, deffile);
		sscanf(helpline, "%ld", &ABS); */
		ABS = 10;
#endif /* DEMO_limited_size */

		fclose(deffile);
  }	/* end read_switches */


/* =============================== init_position =========================================
	 Purpose: Initialize blocknr and pos_in_block of a given position to zero
	 Parameters
	 input:
	 output:
	 input/output: pointer to a position
	 Global variables:
	 Algorithm:
	 Revisions done:
	 Revisions to be done:
	 ---------------------------------------------------------------------------------------- */
void init_position(pos_type_p position)
{
	position->blocknr = 0L;
	position->pos_in_block = 0L;
}

/* =============================== position_is_zero =========================================
	 Purpose: Return TRUE if blocknr and pos_in_block of a given position both equal zero, and 
	 FALSE otherwise
	 Parameters
	 input: position
	 output: boolean result
	 input/output:
	 Global variables:
	 Algorithm:
	 Revisions done:
	 Revisions to be done:
	 ---------------------------------------------------------------------------------------- */
boolean position_is_zero(pos_type position)
{
	return ((position.blocknr == 0L) && (position.pos_in_block == 0L));
}

/* ================================ remove_file ==========================================
	 Purpose: Remove a file, and supply an error message if an error occurred
	 Parameters
	 input: filename
	 output: 
	 input/output:
	 Global variables: pbuffer
	 Algorithm:
	 Revisions done:
	 Revisions to be done:
	 ---------------------------------------------------------------------------------------- */
void remove_file(char *filename)
{
	my_int err_code;

	err_code = remove((char *)filename);
	
/*	3-1-2001 
	if (err_code == 0)
	{  
#ifdef DEBUG
		sprintf(pbuffer, "removing file %s.", filename);
		wwrite(pbuffer, 1);  
#endif
		return;
	}
	else if (errno == EACCES)
	{
		sprintf(pbuffer, "Cannot remove %s; file is read-only.", filename);
		wwrite(pbuffer, 1);
	}
	else if (errno == ENOENT)
	{
		sprintf(pbuffer, "Cannot remove %s; file not found.", filename);
		wwrite(pbuffer, 1);
	}
	else
	{
		sprintf(pbuffer, "Cannot remove file %s.", filename);
		wwrite(pbuffer, 1); 
	} 

*/
}

	/* ================================= exit_store ==========================================
	 Purpose: Close CM and remove the external cache-file if necessary
	 Parameters
	 input: boolean stating whether the external cache-file should be removed or not
	 output:
	 input/output:
	 Global variables: cm_cache_p
	 Algorithm:
	 Revisions done:
	 Revisions to be done:
	 ---------------------------------------------------------------------------------------- */
void exit_store(boolean must_remove)
{
	close_cm(cm_cache_p);
	if (must_remove)
	{
		remove_file(cache_file);
	}
}

/* ================================= open_store ==========================================
	 Purpose: Initialize CM and open the external CM-file
	 Parameters
	 input: page size, buffer size
	 output:
	 input/output:
	 Global variables: cm_cache_p, Cache_PS, Cache_BS
	 Algorithm:
	 Revisions done:
	 Revisions to be done:
	 ---------------------------------------------------------------------------------------- */
void open_store(long PS, my_int BS, long *return_code)
{
	Cache_PS = PS;
	Cache_BS = BS;

	*return_code = 0;
	cm_cache_p = (cm_type_p)malloc(/*calloc 1,*/ sizeof(cm_type));
	if (cm_cache_p == NULL)
	{ 
		*return_code = 1;
		return;
	}
	open_cm(cm_cache_p, cache_file, Cache_PS, Cache_BS, return_code);
}


/* #define TEST_CM */

#ifdef TEST_CM

void test_cache_memory()
{
	typedef struct testrecord_type
		{ long                  identification;
			pos_type							next;  /* next element in the list, as a cm pointer */
			char									vulling[490];
		} testrecord_type;
	typedef testrecord_type      *testrecord_type_p;

	pos_type last_testrecord = {0, 0}, p_filepos;
	long counter, max_count;
	virtual_ptr_type vp;
	testrecord_type_p testrecord_p, p;
	my_int breakpoint;

	/* measurement of influence of memory segmentation on 1, 2, 4, 8, ... bytes */
	boolean	b;
	my_int	m;
	int			i;
	long		l;
	void		*q;
	typedef struct test_sizeof_type
		{ boolean   b;
		} test_sizeof_type;
	test_sizeof_type t;
	my_int	sizeof_b, sizeof_m, sizeof_i, sizeof_l, sizeof_q, sizeof_t;

	sizeof_b = sizeof(b); /* 1 for all segmentations on VC and Alpha */
	sizeof_m = sizeof(m); /* 2 ibid. */ 
	sizeof_i = sizeof(i); /* 4 ibid. */
	sizeof_l = sizeof(l); /* 4 ibid. */
	sizeof_q = sizeof(q); /* 4 ibid. */
  sizeof_t = sizeof(t); /* 1 ibid. */
	/* end of measurement */

	max_count = 300000L;

	for (counter = 1; counter <= max_count; counter++)
	{
		alloc_cm_record(cm_cache_p, sizeof(testrecord_type), &vp);
		testrecord_p = (testrecord_type_p)(vp.rampos);
    /* insert testrecord in front of the list of testrecord's */
    testrecord_p->next = last_testrecord;
    last_testrecord = vp.filepos;
    testrecord_p->identification = counter;
	}
	/* read back */
	p_filepos = last_testrecord;
  p = (testrecord_type_p)filepos_to_ptr(&p_filepos);
	counter = max_count;
	while (p != NULL)
  {
		if (p->identification == counter )
		{
      p_filepos = p->next;
			p = (testrecord_type_p)filepos_to_ptr(&p_filepos);
			counter--;
		}
		else
		{
			breakpoint = 1; /* debugging point */
			exit(1);
		}
	}
	breakpoint = 0; /* debugging point */
	exit(0);
}

#endif


/* =====================  readfile_in_buffer  ======================================
Purpose:		read a xml/sgml-file in a buffer and optionally remember the position 
						of the carriage-returns ("CR")			
Parameters:				
	input:
		- textspace_p: pointer to the buffer
		- eol_proc: boolean, indicates if the position of the CR's
								has to be remembered					
	output:
		- len: the resulting length of the buffer					
	input/output:		
Global variables:	
Algorithm:
		- read the file line by line
		- add a space character to the right of the line
		- if eol_proc: remember the position of the CR in cr_data[] 				
====================================================================*/
static void readfile_in_buffer(char * textspace_p, long eol_proc, 
														long * len) 
/* reads a textfile into a buffer */
{
	char s[inputline_size];
  int length_of_s;
	char last_char_of_s;

 /* --- code change 18-4-2002 */   

	boolean s_full_size, feof_infile, s_ends_on_eol;

  /* readfile_in_buffer */
  *len = 0;
	cr_data_len = 0;
  do 
  {
    fgets(s, inputline_size, infile); /* next input line */

		/* We need to know the following conditions:
				- s ends on an eol: s_ends_on_eol
				- s is of size inputline_size-1: s_full_size
				- eof condition: feof_infile
					N.B.: with eof there is no eol
		*/
		length_of_s = strlen(s);
		last_char_of_s = s[length_of_s - 1];
		s_full_size = (length_of_s == inputline_size - 1);
		feof_infile = feof(infile);
		s_ends_on_eol = (last_char_of_s == '\n');

		/* if s is not empty: 
				- if there is an eol, remove it
				- supply a space at the end of the line in order to separate two words,
				  but not if the last character is already a space or a '>'
		*/
		
		if (length_of_s > 0)
		{ 
			/* if there is an eol, remove it */
			if (s_ends_on_eol)
				length_of_s = length_of_s - 1;
 
	   /* supply a space at the end of the line in order to separate two words,
				 but not if the last character is already a space or a '>';
				 must be a line of reasonable small size (not machine generated) */ 
			last_char_of_s = s[length_of_s - 1];
			if ( (length_of_s < 150) && 
					 (last_char_of_s != BLANK) &&
					 (last_char_of_s != '>') &&
					 !s_full_size )	
	    {   
	      s[++length_of_s - 1] = BLANK;
	    }

/* --- end of code change 18-4-2002 */
 
      if (*len + length_of_s >= (long)textsize)
      {
        sprintf(buffer_sprintf, 
								"File buffer too small. Increase the constant Textsize, which is now %d\n", textsize);
				result = file_buffer_too_small;
				longjmp(exception_status, file_buffer_too_small);
      };
      strncpy(textspace_p + *len, s, length_of_s);
      *len = *len + length_of_s;
      if (eol_proc)
      { 
				cr_data[cr_data_len] = *len;
        cr_data_len++;
        if (cr_data_len >= (long)crsize)
        {
          sprintf(buffer_sprintf, 
									"Cr_data buffer too small. Increase the constant crsize, which is now %d\n", crsize);
					result = cr_data_buffer_too_small;
					longjmp(exception_status, cr_data_buffer_too_small);
        };
      }
    }
		s[0] = 0;
  } while (!feof(infile));
} /* readfile_in_buffer */


/* =====================  writefile_from_buffer  ======================================
Purpose:		write a xml/sgml-file from a buffer			
Parameters:				
	input:					
	output:					
	input/output:		
Global variables:	
	- data_c: buffer of revised text
	- len_c: length of data_c
Algorithm:				
====================================================================*/
static void writefile_from_buffer(void) 
		/* writes the revised text from the revised buffer onto 
																the revised output file */
{	long l;
	char c;
	int nr_of_chars_on_line = 0;

	for (l = 0; l < len_c; l++)
	{ 
		c = data_c[l];
		if ( (c == eol_char) || 
					(nr_of_chars_on_line == max_outputline_size) )
		{ 
			fprintf(outfile, "\n");
			nr_of_chars_on_line = 0;
		};
		if ( c != eol_char)
    { 
			putc( c, outfile);
			nr_of_chars_on_line++;
		}
	};
	fprintf(outfile, "\n");
} /* writefile_from_buffer */


void revxml_common1()
/* common to functions revxml and revxml_p */
{
long return_code;

	/* open logfile */
#ifdef MVS
  logfile = fopen("dd:log", "w");
#endif
#ifndef MVS
  logfile = fopen(log_filename, "w");
#endif
  error_file_open(logfile, log_filename, &result);

	result = checkLicense(buffer_sprintf, lic_filename);
	if (result != 0)
	{
		fprintf(logfile, buffer_sprintf, "%s\n");
		fflush(logfile);
		longjmp(exception_status, result);
	}
	
	read_switches(&result);
	if (result != 0)
	{ 
		sprintf(buffer_sprintf, "%s.\n See logfile.", buffer_sprintf);
		result = switches_not_correct;
		longjmp(exception_status, switches_not_correct);
	}
		crsize = textsize/10;

#ifdef mallocs_from_array    
		alloc_base = malloc(alloc_size);
		if (alloc_base == NULL)
		{
			sprintf(buffer_sprintf, "Heapsize %ld too large", alloc_size);
			result = heapsize_too_large;
			longjmp(exception_status, heapsize_too_large);
		}
#endif
		
		len_c = 0;

/* open Cache Memory file */
	sprintf(cache_file, cache_filename);
	remove_file(cache_file); 
	open_store(APS, ABS, &return_code);
	if (return_code)
	{
		/* give message */
	}

#ifdef TEST_CM
	test_cache_memory();
#endif

} /* end revxml_common1 */


void read_elem_atts_specs()
{  /* read file with names of empty elements and discriminating attributes 
			and store them in a btree */

#ifdef MVS
  infile = fopen("dd:empties", "r");
#endif
#ifndef MVS
  infile = fopen(emp_filename, "r");
#endif
	error_file_open(infile, emp_filename, &result);

	/* read empty tags and discriminating attributes from revxml.emp */
  while (!feof(infile))
  {
		string80 name_element, name_attribute;
		long attribute_nr, element_nr;
		discriminating_attributes_of_element_type_p help_p;
		boolean element_is_empty;
		enum discriminative_type discrimination;
		char* p;

		fgets(helpline, max_symbol_size, infile);
		while ( (helpline[0] == 10) && !feof(infile) )
		{
			fgets(helpline, max_symbol_size, infile);
		}
		if (feof(infile))
		{
			fclose(infile);
			return;
		}
		
		p = strtok(helpline, "/[ \t\n");
	  if (p == NULL)        
		{
			sprintf(buffer_sprintf, "Error in .emp file\n");
      fprintf(logfile, "%s", buffer_sprintf);
      result = error_in_emp_file;
      longjmp(exception_status, error_in_emp_file);
    };
	
		strcpy(name_element, p);
		if (name_element[strlen(name_element)-1] == '\\')
		{	/* empty element */
			name_element[strlen(name_element)-1] = 0;
			element_is_empty = true;
		}
		else
			element_is_empty = false;

		element_nr = add_to_tree(name_element, symbol_is_element);
		empty_element_p[element_nr] = element_is_empty;

		p = strtok(NULL, "[@* \t\n"); /* delivers token ']' */
		p = strtok(NULL, "{|} \t\n"); /* delivers attribute */

		while (p != NULL)
		{
			if (p[0] == '^')
			{  /* negative discriminative attribute */
					strcpy(name_attribute, p + 2); /* trim also '@' */
					discrimination = negative_discriminative;
			}
			else
			{  /* positive discriminative attribute */
					strcpy(name_attribute, p + 1); /* trim also '@' */
					discrimination = positive_discriminative;
			}
			attribute_nr = add_to_tree(name_attribute, symbol_is_attribute);
			help_p = my_malloc(sizeof(discriminating_attributes_of_element_type));
			help_p->attribute = attribute_nr;
			help_p->discrimination = discrimination;
			help_p->next = discriminating_attributes_of_element_p[element_nr];
			discriminating_attributes_of_element_p[element_nr] = help_p;
			p = strtok(NULL, "} \t|\n");
		}
	}
  fclose(infile);
} 

void revxml_common2()
/* common to functions revxml and revxml_p 
Purpose:	to calculate the differences between two xml/sgml files  
Parameters:       
  input:
	output:
Global variables: 
Algorithm:
    - read def-file
		- if switch preproc is on
			- keep_atomic_unit_of_text = atomic_unit_of_text
			- atomic_unit_of_text = 4 (pcdata)
    - read file with names of empty elements and discriminating attributes and store them in a btree
    - initialize variables
    - read text from buffers and create text trees
    - call node_diff to construct compressed matrices d for the
      differences between nodes
    - call reconstruct_diffs to create difference markers for the
      path which traces the minimal differences betwen nodes
		- if switch preproc is on
			- follow the existing logic of visiting adjacent nodes and checking if difference tags
				have to be written; but do not actually write anything:
				- call create_all_shadow_nodes (logic of write_to_outputbuffer)
					- calls function follow_diffs_and_create_shadow_nodes (logic of write_for_node)
						- calls process_diff_markers (logic of write_diff_tags)
							- calls check_for_rev_sequence (logic of write_diff_tag)
			- atomic_unit_of_text = keep_atomic_unit_of_text
			- do for all created shadow_node_pairs
				- for the shadow nodes: call recalculate_atomics
				- call node_diff
				- call reconstruct_diffs
				- replace the corresponding diff-markers of the original nodes with the markers
					of the corresponding shadow nodes
    - call write_to_outputbuffer to create text with textual difference
      markers in the output buffer
    - write statistics to the logfile

------- update 5-12-2001 -----------------------------------

New syntax in .emp file

				//EL2[@*][@CH]	; the element EL2 with pos. discr. attribute CH
				//EL2[@*][@*]		; all attrs. of el. EL2 are pos. discr.
				//*[@*][@TEST]	; all elements with pos. discr. attr. TEST
				//*[@*][@*]			; all attrs. of all el. are pos. discr.
				//EL2\					; the element EL2 is empty

Examples:
				//EL1\
				//EL2 [@*][ @CH ]
				//EL2A [@*][ @CH ]
				//EL3 [@*][ @VERSION ]
				//EL3A [@*][ @VERSION ]
				//EL4 [@*][ @CH | ^@VERSION ]
				//EL4A [@*][ @CH | ^@VERSION ]
				//EL4B [@*][ @CH | ^@VERSION ]
				//EL4C [@*][ @CH | ^@VERSION ]
				//EL4D [@*][ @CH | ^@VERSION1 | ^@VERSION2 ]

The definition of discriminating attributes (and empty SGML elements) has to be made in the file Revxml.emp. The syntax is akin to the syntax of Xpath:
Emp_line			::= '//' Element '[@ Wildcard]' (S '{' S Discr_attr (S '|' S Discr_attr)* S '}')?
Element				::= Name (Empty_indicator)? | Wildcard
Empty_indicator ::= '\'
Discr_attr		::= ('^')? ( Name | Wildcard )
S							::= (space | tab| newline)*
Wildcard			::= '*'

In a future version of RevXml the syntax for the selection of elements and for the attributes within '[' and ']' will be extended, using the syntax of XPath. 
Currently, elements with discriminating attributes have to be specified each on a separate line, where a wildcard stands for all elements that are not specified on other lines. Between square brackets all attributes have to be selected for processing by a wildcard. Between the braces the positive and negative discriminating attributes are specified. A wildcard is allowed. The wildcard stands for all attributes that are not otherwise specified between the same braces.

------------------------------------------------------------

====================================================================*/
{
	
	if (preproc)
	{ /* first prepare for a rough diff in mode rev_pcdata */
		keep_atomic_unit_of_text = atomic_unit_of_text;
		keep_combtags = combtags;
		atomic_unit_of_text = rev_pcdata;
		combtags = false;
	}

  switch (atomic_unit_of_text)
    /* determine separator character in between atomic units of text */
    { case rev_chars:
        strcpy(char_end, "\xff");
        break;
      case rev_words:
        strcpy(char_end, " \t<>?!");
        break;
      case rev_sentences:
        strcpy(char_end, ".<>?!");
        break;
      case rev_pcdata:
        strcpy(char_end, "\xff");
        break;
     }; /* switch */

  alloc_ptr = 0;  /* stackpointer for simulated alloc's */
  position_of_last_close_revision_tag = 0;
	free_d = 0;

  d_value = my_malloc(d_size*sizeof(long));
  d_comes_from = my_malloc(d_size*sizeof(my_byte));

  /* maintenance of btrees for names of empty elements, elements, attributes and attribute values */
	discriminating_attributes_of_element_p = my_calloc(1, max_nr_of_names_of_elements * 
																						sizeof(discriminating_attributes_of_element_type));
	empty_element_p = my_calloc(1, max_nr_of_names_of_elements * sizeof(boolean));
  names_of_elements = my_malloc(max_nr_of_names_of_elements * sizeof(long));
  names_of_attributes = my_malloc(max_nr_of_names_of_attributes * sizeof(long));
  names_of_attribute_values = my_malloc(max_nr_of_names_of_attribute_values * sizeof(long));
  root_of_element_names = NULL;
	root_of_attribute_names = NULL;
	root_of_attribute_values = NULL;
  last_code_of_element_names = 0;
	last_code_of_attribute_names = 0;
	last_code_of_attribute_values = 0;

	/* reserve numbers voor arbs */
	code_of_arb_element = add_to_tree("*", symbol_is_element);
	code_of_arb_attribute = add_to_tree("*", symbol_is_attribute);
	
	read_elem_atts_specs();

  /* initialize counters for the space that will be allocated */
  total_d = 0;  /* the maximum of d that is used */


  top_node_a = my_calloc(1, sizeof(node_type));
	strcpy(top_node_a->tag, "top");
	top_node_a->node_has_to_be_compared = true;

  current_node = top_node_a;

  readfile_from_buffer((char*) data_a, false, len_a, "A");
    /* no eol processing */

  bufferp_char_total_a = bufferp;

  top_node_b = my_calloc(1, sizeof(node_type));
  strcpy(top_node_b->tag, "top");
	top_node_b->node_has_to_be_compared = true;

  current_node = top_node_b;

  readfile_from_buffer((char*) data_b, eol_proc, len_b, "B");

  bufferp_char_total_b = bufferp;

  barposition[1] = 0;
  barsigns[1] = ' ';

  if (trace )
  {
    fprintf(tracefile, "The tree for text data_a is :\n");
    print_tree(top_node_a, 1, (char*) data_a);
    fprintf(tracefile, "\nThe tree for text data_b is :\n");
    print_tree(top_node_b, 1, (char*) data_b);
  };

	all_shadow_nodes_pair_p = NULL;
	active_shadow_nodes_p = NULL;

  distance = node_diff(top_node_a, top_node_b);

  closed_for_b = true;
  closed_for_a = true;

  reconstruct_diffs(top_node_a, top_node_b, true);

  if (trace)
  {
    fprintf(tracefile, "\n");
    fprintf(tracefile,
					"The internal structure for text data_a is :\n");
    print_internal_structure(top_node_a);
    fprintf(tracefile,
          "The internal structure for text data_b is :\n");
    print_internal_structure(top_node_b);
    fprintf(tracefile,
					"The length of the longest common distance is %d\n", distance);
 		fflush(tracefile);
 };

if (preproc)
{
		atomic_unit_of_text = keep_atomic_unit_of_text;
		combtags = keep_combtags;

/* #define print_intermediate_C_file */
#ifdef print_intermediate_C_file
		len_c = 0;
		write_to_outputbuffer();
		add_to_buffer_c(&eol_char, 1);
		add_to_buffer_c(&eol_char, 1);
#endif

		create_all_shadow_nodes();

#define no_intermediate_delete_cm_file
#ifdef intermediate_delete_cm_file
		/* close and delete Cache Memory file */	
		fprintf(logfile, "1st size of revxml.cm was %ld\n", getFileSize(cm_cache_p->fil));
		exit_store(TRUE);
		/* open it again */
		open_store(APS, ABS, return_code);
		if (*return_code)
		{
			/* give message */
		}
#endif

		switch (atomic_unit_of_text)
			/* determine separator character in between atomic units of text */
    { case rev_chars:
        strcpy(char_end, "\xff");
        break;
      case rev_words:
        strcpy(char_end, " \t<>?!");
        break;
      case rev_sentences:
        strcpy(char_end, ".<>?!");
        break;
      case rev_pcdata:
        strcpy(char_end, "\xff");
        break;
     }; /* switch */

		active_shadow_nodes_p = all_shadow_nodes_pair_p;
		while (active_shadow_nodes_p != NULL)
		{
			node_type_p shadow_node_a, shadow_node_b;
			diff_marker_type_p *start_of_diff_markers_to_be_replaced_p_p,
												 *end_of_diff_markers_to_be_replaced_p_p,
												 diff_marker_p;

			shadow_node_a = active_shadow_nodes_p->node_a;
			shadow_node_b = active_shadow_nodes_p->node_b;
			recalculate_atomics((char*) data_a, shadow_node_a);
			recalculate_atomics((char*) data_b, shadow_node_b);

			distance = node_diff(shadow_node_a, shadow_node_b);

			closed_for_b = true;
			closed_for_a = true;
			reconstruct_diffs(shadow_node_a, shadow_node_b, true);

			if (trace)
			{
				fprintf(tracefile, "\n");
				fprintf(tracefile,
							"The internal structure for text data_b is, after 2nd reconstruct_diffs:\n");
				print_internal_structure(top_node_b);
				fprintf(tracefile,
							"The length of the longest common distance is %d\n", distance);
				fflush(tracefile);
			};

			/* replace the corresponding diff-markers of the original nodes with the markers
					of the corresponding shadow nodes */

			diff_marker_p = shadow_node_b->start_diff_marker;
			start_of_diff_markers_to_be_replaced_p_p = active_shadow_nodes_p->start_of_diff_markers_to_be_replaced_p_p;
			end_of_diff_markers_to_be_replaced_p_p = active_shadow_nodes_p->end_of_diff_markers_to_be_replaced_p_p;

			/* replace last pointer in the list of new diff markers */
			if (diff_marker_p != NULL)
			{
				*start_of_diff_markers_to_be_replaced_p_p	= diff_marker_p;
				while (diff_marker_p->next != NULL)
				{
					diff_marker_p = diff_marker_p->next;
				}
				diff_marker_p->next = *end_of_diff_markers_to_be_replaced_p_p;
			}
			else
			{	/* the list of new diff markers is empty */
				*start_of_diff_markers_to_be_replaced_p_p = *end_of_diff_markers_to_be_replaced_p_p;
			}
 			active_shadow_nodes_p = active_shadow_nodes_p->next;
		}

		if (trace)
		{
			fprintf(tracefile, "\n");
			fprintf(tracefile,
						"The internal structure for text data_b is, after diff marker merging:\n");
			print_internal_structure(top_node_b);
			fprintf(tracefile,
						"The length of the longest common distance is %d\n", distance);
			fflush(tracefile);
		}
	}	/* end "if (preproc)" */

  write_to_outputbuffer();

	if (trace)
	{
		fclose(tracefile);
	}

  result = 0;  /* no error for status */

#ifndef DEMO_limited_size
  fprintf(logfile, "Resultcode = %ld\n", result);
	fprintf(logfile,
		"Bytes specified in def-file: data_a and data_b each: %ld d_space: %ld  heap: %ld\n",
     textsize_global, d_size, alloc_size);
  fprintf(logfile,
		"Bytes used for: data_a: %ld  data_b: %ld  d_space: %ld  heap: %ld  text+heap: %ld\n",
     bufferp_char_total_a, bufferp_char_total_b, total_d, alloc_ptr,
     bufferp_char_total_a + bufferp_char_total_b + total_d + alloc_ptr);
#endif

	fprintf(logfile, "Size of revxml.cm was %ld\n", getFileSize(cm_cache_p->fil));
  fclose(logfile);

	/* close and delete Cache Memory file */	
	exit_store(true);

} /* end revxml_common2 */


DllExport(void) revxml_on_buffers( 
												 long in_data_a,					/* pointer to old text */
												 long in_len_a,						/* length of old text */
												 long in_data_b,					/* pointer to new text */
												 long in_len_b,						/* length of new text */
												 long in_data_c,					/* pointer to text with revision markers */
												 long *return_len_c,			/* length of text with revision markers */

												 long *in_cr_data,						/* array which contains the positions of the cr's in data_b */
												 long in_cr_data_len,				/* length of the array cr_data */
												 long *eol_char_param,			/* end of line character on current platform */
												 char *filename_def_param, 		/* filename of file with settings, by default "revxml.def" */
												 char *filename_emp_param, 		/* filename of file with empty elements and discriminative 
																										attributes, by default "revxml.emp" */
												 char *filename_log_param, 		/* filename of logfile */
												 char *filename_lic_param, 					/* filename of license file */
												 char *return_message,		/* string with message after abnormal termination */
												 long eol_proc_param,					/* boolean: true if end-of-lines in file B should be kept in file C */
												 long	*return_error_code	/* error code after abnormal termination */
												 )
{
	if (setjmp(exception_status) != 0)
		 /* a call "longjmp(exception_status, n)" with n != 0
		 will return processing to this position; we transfer
		 here the control back to the calling program; before
		 the call to "longjmp" a status code should have been
		 entered into *result */
	{
		strcpy(return_message, buffer_sprintf);
		if (logfile)
		{
			fprintf(logfile, "Resultcode = %ld\n", result);
			fprintf(logfile, "%s\n", buffer_sprintf);
			fclose(logfile);
		}
		*return_error_code = result;
		return;
	}

	strcpy(def_filename, filename_def_param);
	strcpy(emp_filename, filename_emp_param);
	strcpy(log_filename, filename_log_param);
	strcpy(lic_filename, filename_lic_param);
	eol_proc = eol_proc_param;


	data_a = (char*) in_data_a;
	data_b = (char*) in_data_b;
	data_c = (char*) in_data_c;
	len_a = in_len_a;
	len_b = in_len_b;
	len_c = 0;
	cr_data = in_cr_data;
	cr_data_len = in_cr_data_len;
	eol_char = *((char*) eol_char_param);
	result = 0;
	*return_error_code = 0;

	revxml_common1();

	revxml_common2();

#ifdef mallocs_from_array
	free(alloc_base);
#endif

	*return_len_c = len_c;
	*return_error_code = return_code;

} /* end revxml_on_buffers */ 


/* =====================  revxml_on_files  =================================
Purpose:		to separate the reading and writing of text files from the
						calculation	of differences, done by the function revxmlp
Parameters:
  input:    filename_a	: filename of old text
            filename_b	: filename of new text
            filename_c	: filename of text with rev markers
            filename_def	: filename of file with settings, by default "revxml.def"
            filename_emp	: filename of file with empty elements and discriminative attributes, by default "revxml.emp"
						return_message:	string with message after abnormal termination
						return_code		: error code after abnormal termination
  output:
		- return_code
		- return_message         
  input/output:   
Global variables: 
Algorithm:
		- allocate memory
		- read in files A and B
		- compare A and B
		- write out file C        
----------------------------------------------------------------------*/
DllExport(void) revxml_on_files(	 
												 char *filename_a,			/* filename of old text */
												 char *filename_b,			/* filename of new text */
												 char *filename_c,			/* filename of text with rev markers */
												 char *filename_def_param, 		/* filename of file with settings, by default "revxml.def" */
												 char *filename_emp_param, 		/* filename of file with empty elements and discriminative 
																										attributes, by default "revxml.emp" */
												 char *filename_log_param, 		/* filename of logfile */
												 char *filename_lic_param, 		/* filename of license file */
												 char *return_message,	/* string with message after abnormal termination */
												 long eol_proc_param,					/* boolean: true if end-of-lines in file B should be kept in file C */
												 long	*return_error_code			/* error code after abnormal termination */
											 ) 
{		if (setjmp(exception_status) != 0)
		 /* a call "longjmp(exception_status, n)" with n != 0
		 will return processing to this position; we transfer
		 here the control back to the calling program; before
		 the call to "longjmp" a status code should have been
		 entered into *result */
	{
		strcpy(return_message, buffer_sprintf);
		if (logfile)
		{
			fprintf(logfile, "Resultcode = %ld\n", result);
			fprintf(logfile, "%s\n", buffer_sprintf);
			fclose(logfile);
		}
		*return_error_code = result;
		return;
	}
	
	strcpy(def_filename, filename_def_param);
	strcpy(emp_filename, filename_emp_param);
	strcpy(log_filename, filename_log_param);
	strcpy(lic_filename, filename_lic_param);
	eol_proc = eol_proc_param;

	revxml_common1();

/* allocate memory */
		data_a = malloc(textsize);
		data_b = malloc(textsize);
		data_c = malloc(2*textsize);
		cr_data  = malloc(sizeof(long)*crsize); 
		eol_char = '\n';

/* read files A and B*/
		
    infile = fopen(filename_a, "r");
		error_file_open(infile, filename_a, &result);
    readfile_in_buffer(data_a, false, &len_a);
			/* do not retain positions of CR in data_a */
    fclose(infile);

    infile = fopen(filename_b, "r");
		error_file_open(infile, filename_b, &result);
    readfile_in_buffer(data_b, eol_proc, &len_b);
			/* if eol_proc: retain positions of CR in data_a */
    fclose(infile);
		

	revxml_common2();

/* write file C */
	strcpy(return_message, buffer_sprintf);
	if (*return_error_code == 0)
	{
		outfile = fopen(filename_c, "w");
		error_file_open(outfile, filename_c, &result);
		writefile_from_buffer();
		fclose(outfile);
	}

	free(data_a);
	free(data_b);
	free(data_c);
	free(cr_data); 
#ifdef mallocs_from_array
	free(alloc_base);
#endif

	*return_error_code = return_code;

} /* end revxml_on_files */ 

